/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Nov 30, 2009
 * Author: T. Rekawek <newton@mat.uni.torun.pl>
 */

package pl.edu.icm.unicore.uvos.db;

import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.apache.ibatis.exceptions.PersistenceException;
import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;

import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.UnsupportedTypeException;
import pl.edu.icm.unicore.uvos.model.AttributeBean;
import pl.edu.icm.unicore.uvos.model.AttributeTypeBean;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;

/**
 * DB operations related to attribute types
 * @author golbi
 * @author T. Rekawek
 */
public class DBAttributeTypes
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, DBAttributeTypes.class);
	private Hashtable<String, AttributeTypeBean> attributeTypes;
	private Hashtable<Integer, String> attributeTypes2;
	private Vector<AttributeType> attributeTypesL;
	private DBLock lock;
	
	public DBAttributeTypes(DB db, DBLock lock) throws InternalException
	{
		this.lock = lock;
		attributeTypes = new Hashtable<String, AttributeTypeBean>();
		attributeTypes2 = new Hashtable<Integer, String>();
		attributeTypesL = new Vector<AttributeType>();
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			List<?> attrTypes = sqlMap.selectList("getAllAttributeTypes");
			for (Object o: attrTypes)
			{
				AttributeTypeBean ab = (AttributeTypeBean)o;
				attributeTypes.put(ab.getType(), ab);
				attributeTypesL.add(new AttributeType(ab.getType(), ab.getShortDesc(),
						ab.getDescription()));
				attributeTypes2.put(ab.getId(), ab.getType());
			}
		} catch (PersistenceException e)
		{
			log.error("DB error: getting types. SQL error is: " + e);
			throw new InternalException("Can't load attribute types.");
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
		
	}

	public int getAttrTypeId(AttributeType at, boolean create, SqlSession sqlMap) 
		throws UnsupportedTypeException, InternalException
	{
		AttributeTypeBean b = attributeTypes.get(at.getName());
		if (b == null && !create)
			throw new UnsupportedTypeException();
		if (b != null)
			return b.getId();
		lock.writeLock();
		try
		{
			AttributeTypeBean atb = new AttributeTypeBean();
			atb.setType(at.getName());
			atb.setShortDesc(at.getShortDescription());
			atb.setDescription(at.getShortDescription());
			sqlMap.insert("addAttributeType", atb);
			sqlMap.clearCache();
			Integer newId = (Integer) sqlMap.selectOne(
				"getAttributeTypeId", atb.getType());
			atb.setId(newId);
			attributeTypes.put(atb.getType(), atb);
			attributeTypes2.put(newId, atb.getType());
			attributeTypesL.add(at);
			return newId;
		} catch (PersistenceException e)
		{
			log.error("Can't create a new attribute type", e);
			throw new InternalException("Can't create a new attribute type");
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public int getAttrTypeId(String name) throws UnsupportedTypeException,
		InternalException
	{
		AttributeTypeBean b = attributeTypes.get(name);
		if (b == null)
			throw new UnsupportedTypeException();
		return b.getId();
	}

	public void addAttrType(AttributeType at, SqlSession sqlMap) throws InternalException
	{
		AttributeTypeBean b = new AttributeTypeBean();
		b.setDescription(at.getDescription());
		b.setShortDesc(at.getShortDescription());
		b.setType(at.getName());
		lock.writeLock();
		try
		{
			sqlMap.insert("addAttributeType", b);
			Integer newId = (Integer) sqlMap.selectOne(
					"getAttributeTypeId", at.getName());
			b.setId(newId);
			attributeTypes.put(at.getName(), b);
			attributeTypes2.put(newId, at.getName());
			attributeTypesL.add(at);
		} catch (PersistenceException e)
		{
			log.error("Can't create a new attribute type", e);
			throw new InternalException("Can't create a new attribute type");
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public void updateAttrType(AttributeType at, SqlSession sqlMap) throws InternalException
	{
		AttributeTypeBean b = attributeTypes.get(at.getName());
		b.setDescription(at.getDescription());
		b.setShortDesc(at.getShortDescription());
		b.setType(at.getName());
		lock.writeLock();
		try
		{
			sqlMap.update("updateAttributeType", b);
			for (int i = 0; i<attributeTypesL.size(); i++)
				if (attributeTypesL.get(i).getName().equals(at.getName()))
				{
					attributeTypesL.set(i, at);
					break;
				}
		} catch (PersistenceException e)
		{
			log.error("Can't create a new attribute type", e);
			throw new InternalException("Can't create a new attribute type");
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public void deleteAttrType(AttributeType at, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			Integer newId = (Integer) sqlMap.selectOne(
					"getAttributeTypeId", at.getName());
			sqlMap.delete("deleteAttributeType", at.getName());
			attributeTypes.remove(at.getName());
			attributeTypes2.remove(newId);
			for (int i = 0; i<attributeTypesL.size(); i++)
				if (attributeTypesL.get(i).getName().equals(at.getName()))
				{
					attributeTypesL.remove(i);
					break;
				}
		} catch (PersistenceException e)
		{
			log.info("Can't remove attribute type " +
				"(can be OK, if it has a corresponding attribute): " + 
				e.getMessage());
			//throw new InternalException("Can't create a new attribute type");
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public List<AttributeType> getAttributeTypes()
	{
		return attributeTypesL;
	}
	
	public String getAttrType4Id(int id)
	{
		return attributeTypes2.get(id);
	}

	List<Attribute> prepareAttributeList(List<?> rawL, String scope) 
		throws InternalException
	{
		HashMap<String, Attribute> hash = new HashMap<String, Attribute>();  
		for (Object obj: rawL)
		{
			AttributeBean ab = (AttributeBean)obj;
			AttributeTypeBean atb = attributeTypes.get(getAttrType4Id(ab.getTypeId()));
			String typeName = atb.getType();
			if (ab.getGroupId() != -1 && scope == null)
			{
				log.error("BUG!!! att without requested scope has scope set: " +
						ab.getId());
			}
			typeName = (scope == null ? "" : scope) + "//" + typeName;
			Attribute at = hash.get(typeName);
			if (at == null)
			{
				at = new Attribute(new AttributeType(atb.getType(), atb.getShortDesc(), 
						atb.getShortDesc()),
						new Vector<String>(), scope, 
						ab.getCreationTS(), null);
				hash.put(typeName, at);
			}
			List<String> curVals = at.getValues();
			if (ab.getVal() != null && !curVals.contains(ab.getVal()))
				at.addValue(ab.getVal());
		}
		Vector<Attribute> ret = new Vector<Attribute>();
		Iterator<Attribute> hashVals = hash.values().iterator();
		while (hashVals.hasNext())
			ret.add(hashVals.next());
		return ret;		
	}
	
}
