/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 20, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.db;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.apache.ibatis.exceptions.PersistenceException;
import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;

import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.PermissionDesignator;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.api.AttributePermission;
import pl.edu.icm.unicore.uvos.api.PermissionDesignator.SpecialDesignator;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.UnsupportedTypeException;
import pl.edu.icm.unicore.uvos.model.AuthzBean;
import pl.edu.icm.unicore.uvos.model.GroupBean;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;

/**
 * DB operations related to authorization.
 * @author K. Benedyczak
 */
public class DBAuthz
{
	private static final String OWNER_CODE = "O";
	private static final String MEMBER_CODE = "M";

	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, DBAuthz.class);
	private DBAttributeTypes attributeTypes;
	private DBGroupResolver groupResolver;
	private DBLock lock;
	
	public DBAuthz(DBAttributeTypes attributeTypes, DBGroupResolver groupResolver,
			DBLock lock)
	{
		this.attributeTypes = attributeTypes;
		this.groupResolver = groupResolver;
		this.lock = lock;
	}
	
	public void addAuthZAttr(Attribute attr, int group, Permissions perms, SqlSession sqlMap) 
		throws InternalException, UnsupportedTypeException
	{
		lock.writeLock();
		try
		{
			AuthzBean param = new AuthzBean();
			int attrType = attributeTypes.getAttrTypeId(attr.getType(), true, sqlMap);
			param.setAttrType(attrType);
			param.setGroupId(group);
			param.setPermission(perms.getValue());
			
			List<String> vals = attr.getValues();
			if (vals.size() == 0)
			{
				param.setAttrValue(null);
				try
				{
					sqlMap.insert("insertGroupAuthZ", param);
				} catch (PersistenceException e)
				{
					log.error("DB error: Can't add attr authZ for group " + group + 
						". SQL exception is: " + e);
					throw new InternalException("DB error: authZ for group: " + 
							group);  
				}
			} else
			{
				Set<String> uniqueVals = new HashSet<String>();
				uniqueVals.addAll(vals);
				for (String val: uniqueVals)
				{
					param.setAttrValue(val);
					try
					{
						sqlMap.insert("insertGroupAuthZ", param);
					} catch (PersistenceException e)
					{
						log.error("DB error: Can't add attr authZ for group " + group + 
								". SQL exception is: " + e);
						throw new InternalException("DB error: change " +
								"attr authZ for group: " + group);  
					}
				}
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public void addAuthZSpecial(SpecialDesignator spec, int group, Permissions perms, SqlSession sqlMap) 
		throws InternalException
	{
		lock.writeLock();
		try
		{
			AuthzBean param = new AuthzBean();
			param.setGroupId(group);
			param.setPermission(perms.getValue());
			param.setSpecial(OWNER_CODE);
			if (spec.equals(SpecialDesignator.MEMBER))
				param.setSpecial(MEMBER_CODE);
			try
			{
				sqlMap.insert("insertSpecialAuthZ", param);
			} catch (PersistenceException e)
			{
				log.error("DB error: Can't add special authZ for group " + group + 
						". SQL exception is: " + e);
				throw new InternalException("DB error: change special " +
						"authZ for group: " + group);  
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public void removeAuthZSpecial(SpecialDesignator spec, int group, SqlSession sqlMap) 
		throws InternalException
	{
		lock.writeLock();
		try
		{
			AuthzBean param = new AuthzBean();
			param.setGroupId(group);
			param.setSpecial(OWNER_CODE);
			if (spec.equals(SpecialDesignator.MEMBER))
				param.setSpecial(MEMBER_CODE);
			try
			{
				if (group == -1)
					sqlMap.delete("deleteGlobalAuthZSpecial", 
							param.getSpecial());
				else
					sqlMap.delete("deleteGroupAuthZSpecial", param);
			} catch (PersistenceException e)
			{
				log.error("DB error: Can't delete special authZ for group " + group + 
						". SQL exception is: " + e);
				throw new InternalException("DB error: change special " +
						"authZ for group: " + group);  
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public void removeAuthZAttr(Attribute attr, int group, SqlSession sqlMap) 
		throws InternalException, UnsupportedTypeException
	{
		lock.writeLock();
		try
		{
			AuthzBean param = new AuthzBean();
			param.setGroupId(group);
			int attrType = attributeTypes.getAttrTypeId(attr.getName());
			param.setAttrType(attrType);
			
			try
			{
				if (attr.getValues().size() == 0)
				{
					if (group == -1)
						sqlMap.delete("deleteGlobalAuthZAttr", param);
					else
						sqlMap.delete("deleteGroupAuthZAttr", param);
				} else
				{
					for (String val: attr.getValues())
					{
						param.setAttrValue(val);
						if (group == -1)
							sqlMap.delete("deleteGlobalAuthZAttrVal", param);
						else
							sqlMap.delete("deleteGroupAuthZAttrVal", param);					
					}
				}
			} catch (PersistenceException e)
			{
				log.error("DB error: Can't delete attr authZ for group " + group + 
						". SQL exception is: " + e);
				throw new InternalException("DB error: change attr " +
						"authZ for group: " + group);  
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	private List<AuthzBean> getGroupAuthZRaw(int id, SqlSession sqlMap) throws InternalException
	{
		List<?> res1, res2;
		try
		{
			if (id != -1)
			{
				res1 = sqlMap.selectList("getAllGroupAuthZAttr", 
						Integer.valueOf(id));
				res2 = sqlMap.selectList("getAllGroupAuthZSpecial", 
						Integer.valueOf(id));
			} else
			{
				res1 = sqlMap.selectList("getAllGlobalAuthZAttr");
				res2 = sqlMap.selectList("getAllGlobalAuthZSpecial");
			}
		} catch (PersistenceException e)
		{
			log.error("DB error: Can't get authZ for group " + id + 
					". SQL exception is: " + e);
			throw new InternalException("DB error: authZ for group: " + id);  
		}
		List<AuthzBean> ret = new Vector<AuthzBean>();
		for (Object o: res1)
			ret.add((AuthzBean)o);
		for (Object o: res2)
			ret.add((AuthzBean)o);
		return ret;
	}
	
	private List<AttributePermission> convertResult(List<AuthzBean> raw)
	{
		List<AttributePermission> permissions = new Vector<AttributePermission>();
		Hashtable<String, AttributePermission> attrs = 
			new Hashtable<String, AttributePermission>();
		for (AuthzBean ab: raw)
		{
			PermissionDesignator designator;
			if (ab.getSpecial() != null)
			{
				designator = (ab.getSpecial().equals(OWNER_CODE)) ? 
					new PermissionDesignator(SpecialDesignator.OWNER) :
					new PermissionDesignator(SpecialDesignator.MEMBER);

				permissions.add(new AttributePermission(designator, 
							new Permissions(ab.getPermission())));
			} else
			{
				String typeName = attributeTypes.getAttrType4Id(ab.getAttrType());
				String key = ab.getPermission() + typeName;
				key = (ab.getAttrValue() != null ? "V" : "N") + key;
				
				AttributePermission at = attrs.get(key);
				if (at == null)
				{
					Attribute a = new Attribute(
						new AttributeType(typeName),
						new Vector<String>(), null, null, null);
					
					at = new AttributePermission(new PermissionDesignator(
							a), new Permissions(ab.getPermission()));
					attrs.put(key, at);
				}
				if (ab.getAttrValue() != null)
					at.getDesignator().getAttribute().
						addValue(ab.getAttrValue());
			}
		}
		Iterator<AttributePermission> hashVals = attrs.values().iterator();
		while (hashVals.hasNext())
			permissions.add(hashVals.next());
		return permissions;
	}

	public List<AuthzBean> getGroupAuthZRaw(GroupBean groupB, boolean effective, SqlSession sqlMap) 
		throws InternalException, GroupNotKnownException
	{
		lock.writeLock();
		try
		{
			List<AuthzBean> raw = getGroupAuthZRaw(groupB.getId(), sqlMap);
			if (effective && groupB.getId() != -1)
			{
				//if no policy set find first parent's policy
				while (raw.size() == 0 && groupB.getParent() != -1)
				{
					groupB = groupResolver.resolveGroup4Id(groupB.getParent(), sqlMap);
					raw = getGroupAuthZRaw(groupB.getId(), sqlMap);
				}
				//if still no policy set for any parent use global one
				if (raw.size() == 0)
					raw = getGroupAuthZRaw(-1, sqlMap);
			}
			return raw;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public List<AttributePermission> getGroupAuthZ(GroupBean groupB, boolean effective, SqlSession sqlMap) 
		throws InternalException, GroupNotKnownException
	{
		lock.writeLock();
		try
		{
			return convertResult(getGroupAuthZRaw(groupB, effective, sqlMap));
		}
		finally
		{
			lock.writeUnlock();
		}
	}
}

	
	
	
	
	
	
	
	
	
	
	