/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 20, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.db;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.security.cert.X509Certificate;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Vector;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.store.MemoryStoreEvictionPolicy;

import org.apache.ibatis.exceptions.PersistenceException;
import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;

import eu.emi.security.authn.x509.impl.X500NameUtils;
import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.ModificationEvent;
import pl.edu.icm.unicore.uvos.api.ModificationEvent.TYPE;
import pl.edu.icm.unicore.uvos.api.exceptions.AttributeNotFoundException;
import pl.edu.icm.unicore.uvos.api.exceptions.AuthenticationException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementAlreadyExistsException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.LabelAlreadyUsedException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.UnsupportedTypeException;
import pl.edu.icm.unicore.uvos.engine.InternalAuthorization;
import pl.edu.icm.unicore.uvos.engine.PersistentIdManager;
import pl.edu.icm.unicore.uvos.idhelpers.IdentityHelperFactory;
import pl.edu.icm.unicore.uvos.idhelpers.IdentityTypeHelper;
import pl.edu.icm.unicore.uvos.idhelpers.X509IdentityHelper;
import pl.edu.icm.unicore.uvos.model.AttributeBean;
import pl.edu.icm.unicore.uvos.model.AuthnBean;
import pl.edu.icm.unicore.uvos.model.EntityBean;
import pl.edu.icm.unicore.uvos.model.GenericBean;
import pl.edu.icm.unicore.uvos.model.GroupBean;
import pl.edu.icm.unicore.uvos.model.IdentityBean;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;

/**
 * DB operations related to identities and entities
 * @author K. Benedyczak
 */
public class DBEntities
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, DBEntities.class);
	private Cache certSubjectsCache;
	private static final String cacheName = "CertSubjectsCache";
	private DBAttributeTypes attributeTypes;
	private DBIdentityTypes identityTypes;
	private DBGroupResolver groupResolver;
	private DBGroups groups;
	private DBEntitiesCache entitiesCache;
	private PersistentIdManager idManager;
	private DBLock lock;
	
	public DBEntities(DBAttributeTypes attributeTypes, DBIdentityTypes identityTypes,
			DBGroupResolver groupResolver, PersistentIdManager idManager, DBGroups groups,
			DBEntitiesCache entitiesCache, DBLock lock)
	{
		this.attributeTypes = attributeTypes;
		this.identityTypes = identityTypes;
		this.groupResolver = groupResolver;
		this.idManager = idManager;
		this.groups = groups;
		this.entitiesCache = entitiesCache;
		this.lock = lock;
		prepareCache();
	}
	
	private void prepareCache()
	{
		ByteArrayInputStream bis = new ByteArrayInputStream(
				entitiesCache.getCacheConfig().getBytes());
		CacheManager man = CacheManager.create(bis);
		if(man.cacheExists(cacheName))
			certSubjectsCache = man.getCache(cacheName);
		else
		{
			certSubjectsCache = new Cache("CertSubjectsCache", 
					entitiesCache.getCacheMaxEntities(), MemoryStoreEvictionPolicy.LFU,
					false, null, true, 3600, 3600, false, 240, null);
			man.addCache(certSubjectsCache);
		}
	}
	
	//************************* outside accessible methods ***********************
	
	/*IdentityBean resolveIdentity(Identity identity, Date when) 
		throws IdentityNotKnownException, InternalException, InvalidValueException
	{
		return entitiesCache.resolveIdentity(identity, when);
	}*/

	public Identity getX509BySubject(String canonSubject, SqlSession sqlMap) throws InternalException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			net.sf.ehcache.Element cached = certSubjectsCache.get(canonSubject);
			if (cached != null)
			{
				Integer cachedId = (Integer)cached.getObjectValue();
				try
				{
					IdentityBean cachedIB = entitiesCache.getIdentity4Id(cachedId, sqlMap);
					return resolveIdentityBean(cachedIB, sqlMap);
				} catch(InvalidValueException e)
				{
				//	OK - we had stale entry in cache. Let's remove it and
				//	continue normally
					certSubjectsCache.remove(canonSubject);
				}
			}
			log.debug("Cache miss when trying to get certificate identity by subject. Iterating.");
			List<?> raw;
			try
			{
				raw = sqlMap.selectList("getAllIdentities");
			} catch (PersistenceException e1)
			{
				String msg = "DB error: Can't retrieve identities"; 
				log.warn(msg +". SQL exception is: " + e1);
				throw new InternalException(msg);
			}
			for (Object idRaw: raw)
			{
				IdentityBean ib = (IdentityBean) idRaw;
				IdentityType type = identityTypes.getIdentityType4Id(ib.getTypeId());
				if (!type.equals(IdentityType.X509CERT))
					continue;
				Identity id = resolveIdentityBean(ib, sqlMap);
				X509Certificate cert = new X509IdentityHelper(id).getCertificate();
				String n = X500NameUtils.getComparableForm(
						cert.getSubjectX500Principal().getName());
				certSubjectsCache.put(new net.sf.ehcache.Element(canonSubject, ib.getId()));
				if (n.equals(canonSubject))
				{
					return id;
				}
			}
			return null;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	/**
	 * Adds a new identity.
	 * @param toAdd
	 * @param ctx
	 * @throws InternalException
	 * @throws ElementAlreadyExistsException
	 * @throws LabelAlreadyUsedException 
	 * @throws UnsupportedTypeException 
	 * @throws InvalidValueException 
	 * @throws AuthenticationException 
	 */
	public void addIdentity(Identity toAdd, boolean authTokenEncoded, SqlSession sqlMap) 
		throws InternalException, ElementAlreadyExistsException, 
		UnsupportedTypeException, LabelAlreadyUsedException, 
		InvalidValueException, AuthenticationException
	{
		lock.writeLock();
		try
		{
			if (toAdd.getValue() == null || toAdd.getType() == null)
				throw new InternalException("Wrong (null) attributes of " +
						"identity to create");
			IdentityTypeHelper helper = IdentityHelperFactory.getHelper(toAdd);
			helper.setAuthNTokenEncode(!authTokenEncoded);
			if (toAdd.getLabel() == null)
				toAdd.setLabel(helper.generateUniqueLabel());
			try
			{
				resolveEntity(toAdd.getLabel(), sqlMap);
				throw new LabelAlreadyUsedException();
			} catch (IdentityNotKnownException e) {/*ok*/}
			try
			{
				entitiesCache.resolveIdentity(toAdd, null, sqlMap);
				throw new ElementAlreadyExistsException(new Element(null, toAdd));
			} catch (IdentityNotKnownException e) {/*ok*/}
			
			EntityBean eb = createEntity(toAdd.getLabel(), sqlMap);
			createIdentity(toAdd, eb, helper, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	/**
	 * Adds a new identity, which represents the same entity as one already existing.
	 * @param toAdd
	 * @param equivalentIdentity
	 * @param ctx
	 * @throws InternalException
	 * @throws ElementAlreadyExistsException
	 * @throws IdentityNotKnownException
	 * @throws UnsupportedTypeException 
	 * @throws InvalidValueException 
	 * @throws AuthenticationException 
	 */
	public void addIdentity(Identity toAdd, Identity equivalentIdentity, 
			boolean authTokenEncoded, SqlSession sqlMap) 
		throws InternalException, ElementAlreadyExistsException,
		IdentityNotKnownException, UnsupportedTypeException, 
		InvalidValueException, AuthenticationException
	{
		lock.writeLock();
		try
		{
			if (toAdd.getValue() == null ||	toAdd.getType() == null)
				throw new InternalException("Wrong (null) attributes of " +
						"identity to create");
			if (equivalentIdentity.getValue() == null || 
					equivalentIdentity.getType() == null)
				throw new InternalException("Wrong (null) attributes of " +
						"eqivalent identity");
			
			IdentityTypeHelper helper = IdentityHelperFactory.getHelper(toAdd);
			helper.setAuthNTokenEncode(!authTokenEncoded);
			
			EntityBean eb = entitiesCache.getEntity4Identity(equivalentIdentity, null, sqlMap);
			toAdd.setLabel(eb.getName());
			equivalentIdentity.setLabel(eb.getName());
			try
			{
				entitiesCache.resolveIdentity(toAdd, null, sqlMap);
				throw new ElementAlreadyExistsException(new Element(null, toAdd));
			} catch (IdentityNotKnownException e) {/*ok*/}		
			
			createIdentity(toAdd, eb, helper, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	/**
	 * Deletes identity. If it is the last identity of an entity, 
	 * the corresponding entity is also deleted. If entity is deleted then it's 
	 * attributes are also removed and it is removed from all groups it was a member of.
	 * @param toRemove
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws InvalidValueException 
	 */
	public void removeIdentity(Identity toRemove, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			IdentityBean ib = entitiesCache.resolveIdentity(toRemove, null, sqlMap);
			List<IdentityBean> eqiv = getEqivalents(ib.getEntityId(), sqlMap);
			GenericBean param = new GenericBean();
			param.setP1(new Date());

			if (eqiv.size() == 1)
			{
				param.setP2(ib.getEntityId());
				try
				{
					sqlMap.update("removeAllEntityAttributes", param);
				} catch (PersistenceException e)
				{
					String msg = "DB error: Can't remove entity attributes of " +
						"entity with id " + ib.getEntityId(); 
					log.warn(msg +". SQL exception is: " + e);
					throw new InternalException(msg);
				}
				
				groups.removeFromAllGroups(ib.getEntityId(), sqlMap);
				
				try
				{
					sqlMap.update("removeEntity", param);
				} catch (PersistenceException e)
				{
					String msg = "DB error: Can't remove entity with id " + 
						ib.getEntityId(); 
					log.warn(msg +". SQL exception is: " + e);
					throw new InternalException(msg);
				}
				
			}
			
			entitiesCache.remove(Integer.valueOf(ib.getId()));
			param.setP2(ib.getId());
			try
			{
				sqlMap.update("removeIdentity", param);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't remove identity " + toRemove; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
			entitiesCache.remove(Integer.valueOf(ib.getId()));
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/**
	 * Changes the label of the identity. Label is common for all equivalent identities.
	 * @param toChange
	 * @param label
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws LabelAlreadyUsedException 
	 * @throws InvalidValueException 
	 */
	public void setIdentityLabel(Identity toChange, String label, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, 
		LabelAlreadyUsedException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			IdentityBean ib = entitiesCache.resolveIdentity(toChange, null, sqlMap);
			EntityBean eb =  entitiesCache.getEntity4Identity(ib.getId(), sqlMap);
			try
			{
				resolveEntity(label, sqlMap);
				throw new LabelAlreadyUsedException();
			} catch (IdentityNotKnownException e) {}
			eb.setName(label);
			try
			{
				sqlMap.update("changeEntityName", eb);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't change identity " +
					toChange + " name to " + label; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
			//clean all as there can be multiple entries (if id has equivalents) 
			//to the changed entity.
			try
			{
				entitiesCache.purgeCache();
			} catch (IOException e) {}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/**
	 * Changes the enabled status of the identity. 
	 * @param toChange
	 * @param how true - enable, false - disable
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws InvalidValueException 
	 */
	public void setIdentityStatus(Identity toChange, boolean how, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, 
		InvalidValueException
	{
		lock.writeLock();
		try
		{
			IdentityBean ib = entitiesCache.resolveIdentity(toChange, null, sqlMap);
			ib.setInactive(how ? 0 : 1);
			try
			{
				sqlMap.update("setIdentityStatus", ib);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't set identity " +
					toChange + " active status to " + how; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	
	/**
	 * Changes the entity's persistent id. It is only a maintnance method
	 * used during DB upgrade or initialization but not during normal server's
	 * operation.
	 * 
	 * @param toChange
	 * @param label
	 * @throws InternalException
	 */
	public void updateEntityPersistentId(int toChange, String newId, SqlSession sqlMap) 
		throws InternalException
	{
		lock.writeLock();
		try
		{
			EntityBean eb = new EntityBean();
			eb.setName(newId);
			eb.setId(toChange);
			try
			{
				sqlMap.update("changeEntityPersistentId", eb);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't change entity's " +
					toChange + " persistentId to " + newId; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
			//clean all as there can be multiple entries (if id has equivalents) 
			//to the changed entity.
			try
			{
				entitiesCache.purgeCache();
			} catch (IOException e) {}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public void changeAuthnToken(Identity toChange, String newToken, SqlSession sqlMap) 
		throws IdentityNotKnownException, InternalException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			IdentityBean ib = entitiesCache.resolveIdentity(toChange, null, sqlMap);
			
			if (newToken == null)
			{
				try
				{
					sqlMap.delete("removeIdentityAuthn", ib.getId());
				} catch (PersistenceException e)
				{
					String msg = "DB error: Can't change identity's " +
						toChange + " password."; 
					log.warn(msg +". SQL exception is: " + e);
					throw new InternalException(msg);
				}
			} else
			{
				AuthnBean param = new AuthnBean();
				param.setIdentityId(ib.getId());
				param.setVal(newToken);
				try
				{
					sqlMap.update("changeIdentityAuthn", param);
				} catch (PersistenceException e)
				{
					String msg = "DB error: Can't clear identity's " +
						toChange + " password."; 
					log.warn(msg +". SQL exception is: " + e);
					throw new InternalException(msg);
				}
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	/**
	 * Adds a new entity attribute. 
	 * @param whom Who is the owner of the attribute.
	 * @param toAdd Attribute to add
	 * @param update If true and attribute with the same name exists it will be updated. If false
	 * and attribute with the same name exists there will be exception 
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws ElementAlreadyExistsException if update is false and attribute with the same 
	 * name exists
	 * @throws UnsupportedTypeException 
	 * @throws GroupNotKnownException 
	 * @throws IdentityNotKnownException 
	 * @throws InvalidValueException 
	 */
	public void setAttribute(Element whom, Attribute toAdd, boolean update, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException,
		ElementAlreadyExistsException, 
		GroupNotKnownException, IdentityNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			AttributeBean ab = new AttributeBean();
			AttributeType type = toAdd.getType();
			int typeId;
			try
			{
				typeId = attributeTypes.getAttrTypeId(type, true, sqlMap);
			} catch (UnsupportedTypeException e1)
			{
				log.fatal("BUG: Unsupported type thrown when type should be created.");
				throw new InternalException("BUG in set attribute");
			}
			EntityBean eb = entitiesCache.getEntity4Identity(whom.getIdentity(), null, sqlMap);
			
			if (whom.getGroup() != null)
			{
				Group g = whom.getGroup();
				GroupBean gb = groupResolver.resolveGroup(g, null, sqlMap);
				ab.setGroupId(gb.getId());
			} else
				ab.setGroupId(-1);		
			ab.setCreationTS(new Timestamp(new Date().getTime()));
			ab.setTypeId(typeId);
			ab.setEntityId(eb.getId());
			List<AttributeBean> rawAttrs = getAttributes(eb.getId(), sqlMap);
	
			for (AttributeBean ab2: rawAttrs)
			{
				if (ab2.getTypeId() == typeId && ab2.getGroupId() == ab.getGroupId())
				{
					if (!update)
					{
						throw new ElementAlreadyExistsException(
								new Element(null, null));
					}
					ab2.setRemovalTS(new Timestamp(new Date().getTime()));
					try
					{
						sqlMap.update("removeEntityAttribute", ab2);
					} catch (PersistenceException e)
					{
						String msg = "DB error: Can't remove (to update) entity" +
							" attributes of entity" + eb.getName(); 
						log.warn(msg +". SQL exception is: " + e);
						throw new InternalException(msg);
					}
				}
			}
			
			List<String> vals = toAdd.getValues();
			ab.setVal(null);
			if (vals == null || vals.size() == 0)
			{
				try
				{
					sqlMap.insert("addEntityAttribute", ab);
				} catch (PersistenceException e)
				{
					String msg = "DB error: Can't insert entity attributes for entity " + 
						eb.getName(); 
					log.warn(msg +". SQL exception is: " + e);
					throw new InternalException(msg);
				}
			} else
			{
				for (String val: vals)
				{
					ab.setVal(val);
					try
					{
						sqlMap.insert("addEntityAttribute", ab);
					} catch (PersistenceException e)
					{
						String msg = "DB error: Can't insert entity attributes for entity " + 
							eb.getName(); 
						log.warn(msg +". SQL exception is: " + e);
						throw new InternalException(msg);
					}
				}
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	/**
	 * Removes the entity attribute. 
	 * @param whom
	 * @param toRemove
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws UnsupportedTypeException 
	 * @throws InvalidValueException 
	 */
	public void removeAttribute(Element whom, String toRemove, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, 
		UnsupportedTypeException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			AttributeType aType = new AttributeType(toRemove);
			int typeId = attributeTypes.getAttrTypeId(aType.getName());
			EntityBean eb = entitiesCache.getEntity4Identity(whom.getIdentity(), null, sqlMap);
	
			int gid = -1;
			if (whom.getGroup() != null)
			{
				Group g = whom.getGroup();
				GroupBean gb = groupResolver.resolveGroup(g, null, sqlMap);
				gid = gb.getId();
			}
			
			List<AttributeBean> rawAttrs = getAttributes(eb.getId(), sqlMap);
	
			for (AttributeBean ab2: rawAttrs)
			{
				if (ab2.getTypeId() == typeId && ab2.getGroupId() == gid)
				{
					ab2.setRemovalTS(new Timestamp(new Date().getTime()));
					try
					{
						sqlMap.update("removeEntityAttribute", ab2);
					} catch (PersistenceException e)
					{
						String msg = "DB error: Can't remove entity" +
							" attributes of entity " + eb.getName(); 
						log.warn(msg +". SQL exception is: " + e);
						throw new InternalException(msg);
					}
				}
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	/**
	 * Disables the entity attribute. 
	 * @param whom
	 * @param toDisable
	 * @param value
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws AttributeNotFoundException 
	 * @throws InvalidValueException 
	 * @throws UnsupportedTypeException 
	 */
	public void disableEnableAttribute(Element whom, String toDisable, String value, int how, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, 
		AttributeNotFoundException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			AttributeType aType = new AttributeType(toDisable);
			int typeId;
			try
			{
				typeId = attributeTypes.getAttrTypeId(aType.getName());
			} catch (UnsupportedTypeException e1)
			{
				throw new AttributeNotFoundException(null);
			}
			EntityBean eb = entitiesCache.getEntity4Identity(whom.getIdentity(), null, sqlMap);
	
			int gid = -1;
			if (whom.getGroup() != null)
			{
				Group g = whom.getGroup();
				GroupBean gb = groupResolver.resolveGroup(g, null, sqlMap);
				gid = gb.getId();
			}
			
			AttributeBean ab = new AttributeBean();
			ab.setEntityId(eb.getId());
			ab.setGroupId(gid);
			ab.setTypeId(typeId);
			ab.setVal(value);
			ab.setInactive(how);
			
			String sql;
			if (value == null && gid == -1)
				sql = "disableEnableEntityAttributes1";
			else if (value == null && gid != -1)
				sql = "disableEnableGEntityAttributes1";
			else if (value != null && gid == -1)
				sql = "disableEnableEntityAttributes2";
			else
				sql = "disableEnableGEntityAttributes2";
			
			try
			{
				if (sqlMap.update(sql, ab) == 0)
					throw new AttributeNotFoundException(toDisable);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't (en|dis)able entity" +
					" attributes of entity " + eb.getName(); 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/**
	 * Returns a list with attributes that have disabled values. Only those disabled
	 * values are in attributes returned. 
	 * @param whose
	 * @param ctx
	 * @return
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws InvalidValueException 
	 */
	public List<Attribute> getDisabledAttributes(Element whose, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			int gid = -1;
			String scope = null;
			if (whose.getGroup() != null)
			{
				gid = groupResolver.resolveGroup(whose.getGroup(), null, sqlMap).getId();
				scope = whose.getGroup().toString();
			}
			EntityBean eb = entitiesCache.getEntity4Identity(whose.getIdentity(), null, sqlMap);
			GenericBean param = new GenericBean();
			param.setP1(eb.getId());
			param.setP2(gid);
			List<?> rawL = null;
			try
			{
				if (gid == -1)
					rawL = sqlMap.selectList("listDisabledEntityAttributes", 
						eb.getId());
				else
					rawL = sqlMap.selectList("listDisabledGEntityAttributes", 
						param);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't list disabled entity" +
					" attributes of the entity " + eb.getId(); 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
			return attributeTypes.prepareAttributeList(rawL, scope);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public List<Attribute> getAllAttributes(Element owner, boolean effective, Date when, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			GroupBean groupBean = null;
			String scope = null;
			if (owner.getGroup() != null)
			{
				groupBean = groupResolver.resolveGroup(owner.getGroup(), when, sqlMap);
				scope = owner.getGroup().toString();
			}
				
			EntityBean eb = entitiesCache.getEntity4Identity(owner.getIdentity(), when, sqlMap);
			List<AttributeBean> attrBs;
			if (!effective || groupBean == null)
				attrBs = getAttributes(eb.getId(), 
					groupBean == null ? -1 : groupBean.getId(), when, sqlMap);
			else
				attrBs = getEffectiveAttributes(eb.getId(), groupBean, when, sqlMap);
			return attributeTypes.prepareAttributeList(attrBs, scope);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public boolean areEquivalent(Identity i1, Identity i2, Date when, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			EntityBean eb1 = entitiesCache.getEntity4Identity(i1, when, sqlMap);
			EntityBean eb2 = entitiesCache.getEntity4Identity(i2, when, sqlMap);
			return eb1.getId() == eb2.getId();
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public List<Identity> getAllEquivalents(Identity who, Date when, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			IdentityBean ibBase = entitiesCache.resolveIdentity(who, when, sqlMap);
			return getAllEquivalents(ibBase.getEntityId(), when, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public List<Identity> getAllEquivalents(int entityId, Date when, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException
	{
		lock.writeLock();
		try
		{
			List<?> raw;
			try
			{
				if (when == null)
					raw = sqlMap.selectList("getEquivalentIdentities", 
							entityId);
				else
					raw = sqlMap.selectList("getEquivalentIdentitiesH", 
							new GenericBean(entityId, when));
			} catch (PersistenceException e1)
			{
				String msg = "DB error: Can't retrieve equivalent identities of " + 
				entityId; 
				log.warn(msg +". SQL exception is: " + e1);
				throw new InternalException(msg);
			}
			return convertIdentitiesList(raw, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public Attribute getAttribute(Element owner, String attribute, boolean effective, Date when, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, 
		AttributeNotFoundException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			GroupBean groupB = null;
			String scope = null;
			if (owner.getGroup() != null)
			{
				groupB = groupResolver.resolveGroup(owner.getGroup(), when, sqlMap);
				scope = owner.getGroup().toString();
			}
			EntityBean eb = entitiesCache.getEntity4Identity(owner.getIdentity(), when, sqlMap);
			int typeId;
			try
			{
				typeId = attributeTypes.getAttrTypeId(attribute);
			} catch (UnsupportedTypeException e)
			{
				throw new AttributeNotFoundException(null);
			}
			List <AttributeBean> attrB;
			if (effective && groupB != null)
			{
				List<AttributeBean> allAttrs =
					getEffectiveAttributes(eb.getId(), groupB, when, sqlMap);
				attrB = new Vector<AttributeBean>();
				for (AttributeBean tmp: allAttrs)
					if (tmp.getTypeId() == typeId)
						attrB.add(tmp);
			} else
				attrB =	getAttribute(eb.getId(), groupB == null ? -1 : groupB.getId(), 
						typeId, when, sqlMap);
			if (attrB.size() == 0)
				throw new AttributeNotFoundException(null);
			return attributeTypes.prepareAttributeList(attrB, scope).get(0);
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/*List<EntityBean> getAllEntities() throws InternalException
	{
		return entitiesCache.getAllEntities();
	}*/

	public List<Identity> getAllIdentities(Date when, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List<?> raw;
			try
			{
				if (when == null)
					raw = sqlMap.selectList("getAllIdentities");
				else
					raw = sqlMap.selectList("getAllIdentitiesH", 
							new GenericBean(when));
			} catch (PersistenceException e1)
			{
				String msg = "DB error: Can't retrieve identities"; 
				log.warn(msg +". SQL exception is: " + e1);
				throw new InternalException(msg);
			}
			return convertIdentitiesList(raw, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	/**
	 * Checks if there is authn data for the given identity set, and if yes this data is
	 * returned.
	 * 
	 * @param id id of the identity
	 * @return
	 * @throws InternalException 
	 */
	public String getAuthnData(int id, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			AuthnBean res = (AuthnBean) sqlMap.selectOne("getIdentityAuth", id);
			if (res != null)
				return res.getVal();
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't check identity authn data for: " + id; 
			log.warn(msg +". SQL exception is: " + e);
			throw new InternalException(msg);
		}
		finally
		{
			lock.writeUnlock();
		}
		return null;
	}
	
	
	public void getIdentityEvents(List<ModificationEvent> list, Date from, Date to, 
			int typeFilter, boolean fillDetails, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List<Object> res1 = new ArrayList<Object>(), res2 = new ArrayList<Object>();
			groups.getEventsGeneric(res1, res2, from, to, typeFilter, "getEventsId", sqlMap);
			convertEventsId(list, res1, fillDetails, TYPE.ADD, sqlMap);
			convertEventsId(list, res2, fillDetails, TYPE.REMOVE, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public void getEntityEvents(List<ModificationEvent> list, Date from, Date to, 
			int typeFilter, boolean fillDetails, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List<Object> res1 = new ArrayList<Object>(), res2 = new ArrayList<Object>();
			groups.getEventsGeneric(res1, res2, from, to, typeFilter, "getEventsEnt", sqlMap);
			convertEventsEnt(list, res1, fillDetails, TYPE.ADD);
			convertEventsEnt(list, res2, fillDetails, TYPE.REMOVE);
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public void getAttributeEvents(List<ModificationEvent> list, Date from, Date to, 
			int typeFilter, boolean fillDetails, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List<Object> res1 = new ArrayList<Object>(), res2 = new ArrayList<Object>();
			groups.getEventsGeneric(res1, res2, from, to, typeFilter, "getEventsAttrGlobal", sqlMap);
			groups.convertEventsAttr(list, res1, fillDetails, TYPE.ADD, sqlMap);
			groups.convertEventsAttr(list, res2, fillDetails, TYPE.REMOVE, sqlMap);
			
			res1 = new ArrayList<Object>();
			res2 = new ArrayList<Object>();
			groups.getEventsGeneric(res1, res2, from, to, typeFilter, "getEventsAttrScoped", sqlMap);
			groups.convertEventsAttr(list, res1, fillDetails, TYPE.ADD, sqlMap);
			groups.convertEventsAttr(list, res2, fillDetails, TYPE.REMOVE, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	//TODO this method should be moved to QueryModule when synchro will be converted to locks.
	public List<Attribute> getIdentityAttributes(Element owner, String attribute, boolean includeScoped, 
		boolean globalReadPerm, boolean effective, boolean includeImplied, boolean self,
		Date when, InternalAuthorization authz, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, InvalidValueException, SecurityException
	{
		lock.writeLock();
		try
		{
			//simple variant
			if (!includeScoped)
				return getGlobalAttrs(owner, attribute, effective, when, sqlMap);
			
			List<Attribute> ret;
			if (!globalReadPerm)
				ret = new ArrayList<Attribute>();
			else
				ret = getGlobalAttrs(owner, attribute, effective, when, sqlMap);
			
			//get all groups honoring includeImplied. Permissions are anyway checked 
			//in getAllGroups, as only groups with read access are returned.
			List<Group> groupsAll = groups.getAllReadableGroups(globalReadPerm, 
					owner.getIdentity(), includeImplied, authz, sqlMap);
			Element grUser = new Element(null, owner.getIdentity());
			for (Group group: groupsAll)
			{
				grUser.setGroup(group);
				if (attribute == null)
					ret.addAll(getAllAttributes(grUser, effective, when, sqlMap));
				else
				{
					try
					{
						ret.add(getAttribute(grUser, attribute, 
							effective, when, sqlMap));
					} catch (AttributeNotFoundException e) {}
				}
			}
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	
	//*************************** internal methods *********************************** 
	
	private List<Attribute> getGlobalAttrs(Element owner, String attribute, 
		boolean effective, Date when, SqlSession sqlMap) throws InternalException,
		ElementNotKnownException, InvalidValueException
	{
		if (attribute == null)
			return getAllAttributes(owner, effective, when, sqlMap);
		ArrayList<Attribute> ret = new ArrayList<Attribute>();
		try
		{
			ret.add(getAttribute(owner, attribute, effective, when, sqlMap));
		} catch (AttributeNotFoundException e) {}
		return ret;
	}


	
	private void convertEventsId(List<ModificationEvent> list, List<Object> res, 
			boolean fillDetails, TYPE type, SqlSession sqlMap) throws InternalException
	{
		boolean add = type.equals(TYPE.ADD);
		for (Object o: res)
		{
			IdentityBean id = (IdentityBean)o;
			ModificationEvent e = new ModificationEvent(
					add ? id.getCreationTS() : id.getRemovalTS(), 
					ModificationEvent.CATEGORY.IDENTITY,
					type);
			
			if (fillDetails)
				e.setDescription(resolveIdentityBean(id, sqlMap).toString());
			list.add(e);
		}		
	}

	private void convertEventsEnt(List<ModificationEvent> list, List<Object> res, 
			boolean fillDetails, TYPE type) throws InternalException
	{
		boolean add = type.equals(TYPE.ADD);
		for (Object o: res)
		{
			EntityBean id = (EntityBean)o;
			ModificationEvent e = new ModificationEvent(
					add ? id.getCreationTS() : id.getRemovalTS(), 
					ModificationEvent.CATEGORY.ENTITY,
					type);
			
			if (fillDetails)
				e.setDescription(id.getName() + " (persistent id: " +
						id.getPersistentId() + ")");
			list.add(e);
		}		
	}
	
	
	private Identity resolveIdentityBean(IdentityBean ib, SqlSession sqlMap) throws InternalException
	{
		return resolveIdentityBeanCore(ib, sqlMap);
	}

	private Identity resolveIdentityBeanCore(IdentityBean ib, SqlSession sqlMap) throws InternalException
	{
		IdentityType it = identityTypes.getIdentityType4Id(ib.getTypeId());
		EntityBean e;
		try
		{
			e = entitiesCache.getEntity4Identity(ib.getId(), sqlMap);
		} catch (IdentityNotKnownException e1)
		{
			throw new InternalException("Can't resolve identity bean", e1);
		}
		Identity ret = new Identity(it, ib.getValue(), e.getName(), 
				e.getPersistentId(),
				ib.getCreationTS(), ib.getRemovalTS());
		ret.setInactive(ib.getInactive() != 0);
		return ret;
	}

	private List<Identity> convertIdentitiesList(List<?> raw, SqlSession sqlMap) throws InternalException
	{
		Vector<Identity> ret = new Vector<Identity>();
		for (Object o: raw)
		{
			Identity i = resolveIdentityBeanCore((IdentityBean)o, sqlMap);
			ret.add(i);
		}
		return ret;
	}
	
	private EntityBean resolveEntity(String name, SqlSession sqlMap) 
		throws IdentityNotKnownException, InternalException
	{
		EntityBean res;

		try
		{
			res = (EntityBean) sqlMap.selectOne("resolveEntity", name);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't resolve entity " + name; 
			log.warn(msg +". SQL exception is: " + e);
			throw new InternalException(msg);
		}
		if (res == null)
			throw new IdentityNotKnownException(null);
		return res;
	}
/*
	EntityBean getEntity4Identity(int id) 
		throws InternalException, IdentityNotKnownException
	{
		net.sf.ehcache.Element cached = entitiesCache.get(Integer.valueOf(id));
		if (cached != null)
			return (EntityBean) cached.getObjectValue();
		EntityBean res;
		try
		{
			res = (EntityBean) 
				sqlMap.queryForObject("getEntity4Identity", Integer.valueOf(id));
		} catch (PersistenceException e)
		{
			log.error("DB error: Can't get entity for identity id " + id + 
					". SQL exception is: " + e);
			throw new InternalException("DB error: No entity for identity: " + id);  
		}
		if (res == null)
			throw new IdentityNotKnownException(null);
		entitiesCache.put(new net.sf.ehcache.Element(Integer.valueOf(id), res));
		return res;
	}*/
		
	/**
	 * It should be ensured earlier that label does not exist
	 * @param label
	 * @return
	 * @throws InternalException
	 */
	private EntityBean createEntity(String label, SqlSession sqlMap) throws InternalException
	{
		GenericBean gb = new GenericBean();
		EntityBean entityB; 
		gb.setP1(label);
		gb.setP2(new Date());
		String id = idManager.getNewId();
		gb.setP3(id);
		try
		{
			sqlMap.insert("addEntity", gb);
			sqlMap.clearCache();
			entityB = (EntityBean) 
				sqlMap.selectOne("resolveEntity", label);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't add a new entity " + label;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
		return entityB;
	}
	
	/**
	 * It should be ensured earlier that identity does not exist
	 * Also it is NOT atomic so must be called inside transaction scope!!
	 * @param identity
	 * @param entityB
	 * @throws InternalException
	 * @throws UnsupportedTypeException
	 * @throws AuthenticationException 
	 * @throws InvalidValueException 
	 * @throws ElementAlreadyExistsException
	 */
	private void createIdentity(Identity identity, EntityBean entityB, 
			IdentityTypeHelper helper, SqlSession sqlMap) 
		throws InternalException, UnsupportedTypeException, AuthenticationException,
		InvalidValueException
	{
		String type = identity.getType().getName().toString();
		int typeId = identityTypes.getIdentityTypeId(type);
		
		IdentityBean param = new IdentityBean();
		param.setCreationTS(new Timestamp(Calendar.getInstance().getTimeInMillis()));
		param.setEntityId(entityB.getId());
		param.setTypeId(typeId);
		param.setValue(identity.getValue());
		param.setCmp(helper.getComparableValue());
		try
		{
			
			sqlMap.insert("addIdentity", param);
			if (identity.getAuthnData() != null)
			{
				AuthnBean param2 = new AuthnBean();
				param2.setVal(helper.generateAuthNToken());
				IdentityBean idB = null;
				try
				{
					idB = entitiesCache.resolveIdentity(identity, null, sqlMap);
					param2.setIdentityId(idB.getId());
				} catch (IdentityNotKnownException e){}
				sqlMap.update("changeIdentityAuthn", param2);
			}
			sqlMap.clearCache();
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't add a new entity " + identity.getLabel();
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);
		}
	}
	
	private List<IdentityBean> getEqivalents(int entityId, SqlSession sqlMap) throws InternalException
	{
		List<?> eqiv;
		try
		{
			eqiv = sqlMap.selectList("getEquivalentIdentities", entityId);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't retirieve equivalent identities " +
					"with entity id " + entityId; 
			log.warn(msg +". SQL exception is: " + e);
			throw new InternalException(msg);
		}
		Vector<IdentityBean> ret = new Vector<IdentityBean>();
		for (Object o: eqiv)
			ret.add((IdentityBean) o);
		return ret;
	}
	
	private List<AttributeBean> getAttributes(int entityId, SqlSession sqlMap) throws InternalException
	{
		List<?> rawAttrs = null;
		try
		{
			rawAttrs = sqlMap.selectList("getEntityAttributes", entityId);
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve entity attributes for entity " + 
				entityId; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
		Vector<AttributeBean> ret = new Vector<AttributeBean>();
		for (Object o: rawAttrs)
			ret.add((AttributeBean) o);
		return ret;
	}

	private List<AttributeBean> getAttributes(int entityId, int gid, Date when, SqlSession sqlMap) 
		throws InternalException
	{
		List<?> rawAttrs = null;
		GenericBean gb = new GenericBean();
		gb.setP1(entityId);
		gb.setP2(gid);
		try
		{
			if (when == null)
			{
				if (gid == -1)
					rawAttrs = sqlMap.selectList(
						"getEntityGlobalAttributes", entityId);
				else
					rawAttrs = sqlMap.selectList(
						"getEntityGroupAttributes", gb);
			} else
			{
				gb.setP3(when);
				if (gid == -1)
					rawAttrs = sqlMap.selectList(
						"getEntityGlobalAttributesH", 
						new GenericBean(entityId, when));
				else
					rawAttrs = sqlMap.selectList(
						"getEntityGroupAttributesH", gb);				
			}
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve entity attributes for entity " + 
				entityId; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
		Vector<AttributeBean> ret = new Vector<AttributeBean>();
		for (Object o: rawAttrs)
			ret.add((AttributeBean) o);
		return ret;
	}

	public List<AttributeBean> getEffectiveAttributes(int entityId, GroupBean groupB, Date when, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, GroupNotKnownException
	{
		lock.writeLock();
		try
		{
			// global
			List<AttributeBean> ret = getAttributes(entityId, -1, when, sqlMap);
			if (groupB.getId() == -1)
				return ret;
			
			List<GroupBean> allGroups = groups.getAllGroupsRaw(entityId, when, sqlMap);
	
			for (GroupBean gb: allGroups)
			{
				List<GroupBean> path = groups.getGroupPath(groupB, gb, sqlMap);
				for (GroupBean subGroup: path)
				{
					List<AttributeBean> scopedAttrs = getAttributes(entityId, 
							subGroup.getId(), when, sqlMap);
					List<AttributeBean> groupAttrs = groups.getAttributes(
							subGroup.getId(), when, sqlMap);
					//group attributes are added only if there is no scoped attribute
					//with the same name (== type id)
					for (AttributeBean ab: groupAttrs)
					{
						int i;
						for (i=0; i<scopedAttrs.size(); i++)
							if (scopedAttrs.get(i).getTypeId() == 
								ab.getTypeId())
								break;
						if (i == scopedAttrs.size())
							ret.add(ab);
					}
					ret.addAll(scopedAttrs);
				}
			}
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	private List<AttributeBean> getAttribute(int entityId, int gid, int typeId, Date when, SqlSession sqlMap) 
		throws InternalException
	{
		List<?> rawAttrs = null;
		AttributeBean param = new AttributeBean();
		param.setEntityId(entityId);
		param.setGroupId(gid);
		param.setTypeId(typeId);
		try
		{
			if (when == null)
			{
				if (gid == -1)
					rawAttrs = sqlMap.selectList(
						"getEntityGlobalAttribute", param);
				else
					rawAttrs = sqlMap.selectList(
						"getEntityGroupAttribute", param);
			} else
			{
				param.setRemovalTS(new Timestamp(when.getTime()));
				if (gid == -1)
					rawAttrs = sqlMap.selectList(
						"getEntityGlobalAttributeH", param);
				else
					rawAttrs = sqlMap.selectList(
						"getEntityGroupAttributeH", param);
			}
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve entity attributes for entity " + 
			entityId; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
		Vector<AttributeBean> ret = new Vector<AttributeBean>();
		for (Object o: rawAttrs)
			ret.add((AttributeBean) o);
		return ret;
	}
}
