/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Nov 30, 2009
 * Author: T. Rekawek <newton@mat.uni.torun.pl>
 */

package pl.edu.icm.unicore.uvos.db;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Date;
import java.util.List;

import net.sf.ehcache.Cache;
import net.sf.ehcache.CacheException;
import net.sf.ehcache.CacheManager;
import net.sf.ehcache.store.MemoryStoreEvictionPolicy;

import org.apache.ibatis.exceptions.PersistenceException;
import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;

import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.idhelpers.IdentityHelperFactory;
import pl.edu.icm.unicore.uvos.idhelpers.IdentityTypeHelper;
import pl.edu.icm.unicore.uvos.model.EntityBean;
import pl.edu.icm.unicore.uvos.model.GenericBean;
import pl.edu.icm.unicore.uvos.model.IdentityBean;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;

/**
 * DB operations related to entities involving cache. 
 * @author T. Rekawek
 */
public class DBEntitiesCache
{
	private static final String cacheName = "EntitiesCache";
	private static final int CACHE_MAX_ENTITIES = 1000;
	private static final String cacheConfig = "<ehcache updateCheck=\"false\">" +
		"<defaultCache maxElementsInMemory=\"" + CACHE_MAX_ENTITIES + "\" "+
		"eternal=\"true\" "+
		"timeToIdleSeconds=\"10\" "+
		"timeToLiveSeconds=\"10\" "+
		"overflowToDisk=\"false\" "+
		"diskPersistent=\"false\" "+
		"diskExpiryThreadIntervalSeconds=\"1200\"/>" +
		"</ehcache>";
	
	private DBLock lock;
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, DBEntitiesCache.class);
	private Cache entitiesCache;
	
	public DBEntitiesCache(DBLock lock)
	{
		this.lock = lock;
		prepareCache();
	}
	
	private void prepareCache()
	{
		ByteArrayInputStream bis = new ByteArrayInputStream(
				cacheConfig.getBytes());
		CacheManager man = CacheManager.create(bis);
		if(man.cacheExists(cacheName))
			entitiesCache = man.getCache(cacheName);
		else
		{
			entitiesCache = new Cache(cacheName, 
					CACHE_MAX_ENTITIES, MemoryStoreEvictionPolicy.LFU,
					false, null, true, 3600, 3600, false, 240, null);
			man.addCache(entitiesCache);
		}
	}
	
	public EntityBean getEntity4Identity(int id, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException
	{
		lock.writeLock();
		try
		{
			net.sf.ehcache.Element cached = entitiesCache.get(Integer.valueOf(id));
			if (cached != null)
				return (EntityBean) cached.getObjectValue();
			EntityBean res;
			try
			{
				res = (EntityBean) 
					sqlMap.selectOne("getEntity4Identity", Integer.valueOf(id));
				if(res != null)
					entitiesCache.put(new net.sf.ehcache.Element(Integer.valueOf(id), res));
			} catch (PersistenceException e)
			{
				log.error("DB error: Can't get entity for identity id " + id + 
						". SQL exception is: " + e);
				throw new InternalException("DB error: No entity for identity: " + id);  
			}
			if (res == null)
				throw new IdentityNotKnownException(null);
			return res;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public EntityBean getEntity4Identity(Identity identity, Date when, SqlSession sqlMap) 
	throws IdentityNotKnownException, InternalException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			IdentityBean ib = resolveIdentity(identity, when, sqlMap);
			return getEntity4Identity(ib.getId(), sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public IdentityBean resolveIdentity(Identity identity, Date when, SqlSession sqlMap) 
		throws IdentityNotKnownException, InternalException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			IdentityBean res;
			try
			{
				IdentityTypeHelper helper;
				helper = IdentityHelperFactory.getHelper(identity);
				GenericBean param = new GenericBean(helper.getComparableValue(),
						identity.getType().toString());
				if (when == null)
					res = (IdentityBean) sqlMap.selectOne("resolveIdentity", param);
				else
				{
					param.setP3(when);
					res = (IdentityBean) sqlMap.selectOne("resolveIdentityH", param);
				}
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't resolve identity " + identity; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
			if (res == null)
				throw new IdentityNotKnownException(identity);
			return res;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	
	/**
	 * Changes the entity's persistent id. It is only a maintnance method
	 * used during DB upgrade or initialization but not during normal server's
	 * operation.
	 * 
	 * @param toChange
	 * @param label
	 * @throws InternalException
	 */
	public void updateEntityPersistentId(int toChange, String newId, SqlSession sqlMap) 
		throws InternalException
	{
		lock.writeLock();
		try
		{
			EntityBean eb = new EntityBean();
			eb.setName(newId);
			eb.setId(toChange);
			try
			{
				sqlMap.update("changeEntityPersistentId", eb);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't change entity's " +
					toChange + " persistentId to " + newId; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
			//clean all as there can be multiple entries (if id has equivalents) 
			//to the changed entity.
			entitiesCache.removeAll();
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public void purgeCache() throws IllegalStateException, CacheException, IOException
	{
		entitiesCache.removeAll();
	}
	
	public void remove(Integer id)
	{
		entitiesCache.remove(id);
	}
	
	public int getCacheMaxEntities()
	{
		return DBEntitiesCache.CACHE_MAX_ENTITIES;
	}
	
	public String getCacheConfig()
	{
		return DBEntitiesCache.cacheConfig;
	}
	
	@SuppressWarnings("unchecked")
	public List<EntityBean> getAllEntities(SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			return sqlMap.selectList("getAllEntities");
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve entities"; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	IdentityBean getIdentity4Id(int id, SqlSession sqlMap) throws InternalException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			IdentityBean res;
			try
			{
				res = (IdentityBean) sqlMap.selectOne("getIdentity4Id", 
						Integer.valueOf(id));
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't get identity for id" + id; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
			if (res == null)
				throw new InvalidValueException("No such identity.");
			return res;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
}