/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Nov 30, 2009
 * Author: T. Rekawek <newton@mat.uni.torun.pl>
 */

package pl.edu.icm.unicore.uvos.db;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.ibatis.exceptions.PersistenceException;
import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;

import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.model.GenericBean;
import pl.edu.icm.unicore.uvos.model.GroupBean;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;

/**
 * DB operations related to resolving groups
 * @author T. Rekawek
 */
public class DBGroupResolver
{
	private DBLock lock;
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, DBGroupResolver.class);
	
	public DBGroupResolver(DBLock lock)
	{
		this.lock = lock;
	}
	
	GroupBean resolveGroup4Id(int id, SqlSession sqlMap) 
		throws InternalException, GroupNotKnownException
	{
		lock.writeLock();
		try
		{
			GroupBean res = null;
			try
			{
				res = (GroupBean)sqlMap.selectOne("resolveGroup4Id", id);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't resolve group name for id " + id;
				log.error(msg + ". SQL exception is: " + e);
				throw new InternalException(msg);  
			}
			if (res == null)
			{
				log.info("Can't find group with id " + id);
				throw new GroupNotKnownException(null);
			}
			return res;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public GroupBean resolveGroup(String name, int parentId, Date when, SqlSession sqlMap) 
		throws InternalException, GroupNotKnownException
	{
		GroupBean res = null;
		GroupBean param = new GroupBean();
		param.setName(name);
		param.setParent(parentId);
		if (when != null)
			param.setRemovalTS(new Timestamp(when.getTime()));
		try
		{
			if (when == null)
			{
				if (parentId == -1)
					res = (GroupBean)sqlMap.selectOne("resolveTopGroup", name);
				else
					res = (GroupBean)sqlMap.selectOne("resolveGroup", param);
			} else
			{
				if (parentId == -1)
					res = (GroupBean)sqlMap.selectOne("resolveTopGroupH", new GenericBean(name, when));
				else
					res = (GroupBean)sqlMap.selectOne("resolveGroupH", param);
			}
			
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't resolve group " + name + 
				" parent " + parentId;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
		if (res == null)
			throw new GroupNotKnownException(null);
		return res;
	}

	public GroupBean resolveGroup(Group g, Date when, SqlSession sqlMap) throws InternalException, GroupNotKnownException 
	{
		lock.writeLock();
		try
		{
			String path[] = (g != null) ? g.getPath() : new String[0];
			int p = -1;
			GroupBean b = new GroupBean();
			b.setId(-1);
			b.setParent(-1);
			for (int i=0; i<path.length; i++)
			{
				b = resolveGroup(path[i], p, when, sqlMap);
				p = b.getId();
			}
			return b;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public Group resolveGroupBean(GroupBean g, SqlSession sqlMap) throws InternalException, GroupNotKnownException 
	{
		lock.writeLock();
		try
		{
			String path = g.getName();
			while (g.getParent() != -1)
			{
				g = resolveGroup4Id(g.getParent(), sqlMap);
				path = g.getName() + "/" + path;
			}
			path = "/" + path;
			return new Group(path);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public List<GroupBean> getParentGroupBeans(GroupBean g, SqlSession sqlMap) throws InternalException, GroupNotKnownException 
	{
		List<GroupBean> ret = new ArrayList<GroupBean>();
		lock.writeLock();
		try
		{
			while (g.getParent() != -1)
			{
				g = resolveGroup4Id(g.getParent(), sqlMap);
				ret.add(g);
			}
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
}
