/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 20, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.db;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.apache.ibatis.exceptions.PersistenceException;
import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;

import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.ModificationEvent;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.api.ModificationEvent.TYPE;
import pl.edu.icm.unicore.uvos.api.exceptions.AttributeNotFoundException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementAlreadyExistsException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotEmptyException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.UnsupportedTypeException;
import pl.edu.icm.unicore.uvos.engine.InternalAuthorization;
import pl.edu.icm.unicore.uvos.model.AttributeBean;
import pl.edu.icm.unicore.uvos.model.AttributeWithOwnerBean;
import pl.edu.icm.unicore.uvos.model.AuthzBean;
import pl.edu.icm.unicore.uvos.model.ElementBean;
import pl.edu.icm.unicore.uvos.model.EntityBean;
import pl.edu.icm.unicore.uvos.model.GenericBean;
import pl.edu.icm.unicore.uvos.model.GroupBean;
import pl.edu.icm.unicore.uvos.model.IdentityBean;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;

/**
 * DB operations related to groups. Those methods are available through DB class. 
 * Synchronization is also done there.
 * TODO add extensive caching to eliminate resolve operation.
 * @author K. Benedyczak
 */
public class DBGroups
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, DBGroups.class);
	private DBAttributeTypes attributeTypes;
	private DBIdentityTypes identityTypes;
	private DBAuthz authz;
	private DBEntitiesCache entitiesCache;
	private DBApps dbApps;
	private DBGroupResolver groupResolver;
	private DBLock lock;
	
	private DB db;

	public DBGroups(DBAttributeTypes attributeTypes, DBIdentityTypes identityTypes,
			DBApps dbApps, DBAuthz authz, DBGroupResolver groupResolver, DBEntitiesCache entitiesCache,
			DBLock lock, DB db)
	{
		this.db = db;
		this.attributeTypes = attributeTypes;
		this.identityTypes = identityTypes;
		this.authz = authz;
		this.dbApps = dbApps;
		this.groupResolver = groupResolver;
		this.entitiesCache = entitiesCache;
		this.lock = lock;
	}


	//****************** External interface (used by DB class) *************************
	
	/**
	 * Adds a new group. Pass null parent to create top-level group (i.e. VO)
	 * @param parent
	 * @param name
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws ElementAlreadyExistsException
	 */
	public Group addGroup(Group parent, String name, SqlSession sqlMap) 
		throws InternalException, GroupNotKnownException, 
		ElementAlreadyExistsException
	{
		lock.writeLock();
		try
		{
			int parentId = -1;
			if (parent != null)
			{
				GroupBean pb = groupResolver.resolveGroup(parent, null, sqlMap);
				parentId = pb.getId();
			}
			
			Group result = new Group(parent == null ? new String[0] : parent.getPath(),
					name, null, null); 
			try
			{
				groupResolver.resolveGroup(result, null, sqlMap);
				throw new ElementAlreadyExistsException(new Element(result, null));
			} catch(GroupNotKnownException e){/*ok*/}
			createGroup(parentId, name, sqlMap);
			
			GroupBean res = groupResolver.resolveGroup(name, parentId, null, sqlMap); 
			result.setCreationTS(res.getCreationTS());
			return result;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	/**
	 * Removes the given group. All group attributes of this group and entity attributes 
	 * bound to this group are also removed. All authz (both owner and normal) permissions
	 * bound to this group are also removed. All application forms bound to this group
	 * are also removed.
	 * <p>
	 * If group have members (either subgroups or entities) those members
	 * are removed if and only if recursive is true. Otherwise exception is thrown.
	 * @param toRemove
	 * @param recursive
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws GroupNotEmptyException
	 */
	public void removeGroup(Group toRemove, boolean recursive, SqlSession sqlMap) 
		throws InternalException, GroupNotKnownException, GroupNotEmptyException
	{
		lock.writeLock();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(toRemove, null, sqlMap);
			if (gb.getId() == -1)
				throw new GroupNotKnownException(toRemove);
			Timestamp ts = new Timestamp(new Date().getTime());
			removeGroupRecursive(gb.getId(), recursive, ts, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	/**
	 * Disallow: newParent+newName == srcGB
	 * srcGB == /
	 * newParent == srcGb
	 * @param srcGb
	 * @param newParent
	 * @param newName
	 * @param removeOriginal
	 * @return
	 * @throws GroupNotKnownException
	 * @throws InternalException
	 * @throws ElementAlreadyExistsException
	 * @throws InvalidValueException 
	 */
	public Group copyGroup(GroupBean srcGb, Group newParent, String newName, 
		boolean removeOriginal, SqlSession sqlMap) throws GroupNotKnownException, 
		InternalException, ElementAlreadyExistsException, InvalidValueException 
	{
		lock.writeLock();
		try
		{
			if (srcGb.getId() == -1)
				throw new InvalidValueException("Can't copy root");
			GroupBean newPGB = groupResolver.resolveGroup(newParent, null, sqlMap);
			if (newPGB.getId() == srcGb.getId())
				throw new InvalidValueException("Can't copy group to itself");
			if (srcGb.getParent() == newPGB.getId() && newName.equals(srcGb.getName()))
				throw new InvalidValueException("Copy destination is equal to source");
			Group res;
			try
			{
				res = copyGroupInt(srcGb, newParent, newName, removeOriginal, sqlMap);
			} catch (GroupNotEmptyException e)
			{
				log.fatal("BUG: GroupNotEmptyExc during recursive remove", e);
				throw new InternalException("BUG: " +
					"GroupNotEmptyExc during recursive remove");
			} 
			return res;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	private Group copyGroupInt(GroupBean srcGb, Group newParent, String newName, 
			boolean removeOriginal, SqlSession sqlMap) 
		throws GroupNotKnownException, InternalException, 
		ElementAlreadyExistsException, GroupNotEmptyException
	{
		//create target
		Group result = addGroup(newParent, newName, sqlMap);
		GroupBean dstGb = groupResolver.resolveGroup(result, null, sqlMap);
		//copy of everything except subgroups
		copyMembers(srcGb.getId(), dstGb.getId(), sqlMap);
		copyGroupAttributes(srcGb.getId(), dstGb.getId(), sqlMap);
		copyAuthz(srcGb, dstGb.getId(), sqlMap);
		//do recursive copy of the subgroups (without delete!)
		List subGroups = getGroupSubgroupsFullRaw(srcGb.getId(), null, sqlMap);
		for (Object o: subGroups)
		{
			GroupBean gb = (GroupBean)o;
			copyGroupInt(gb, result, gb.getName(), false, sqlMap);
		}
		//delete recursively orginal if needed
		if (removeOriginal)
		{
			Timestamp ts = new Timestamp(new Date().getTime());
			removeGroupRecursive(srcGb.getId(), true, ts, sqlMap);
		}
		return result;
	}
	
	private void copyAuthz(GroupBean srcGb, int dstG, SqlSession sqlMap) throws InternalException,
		GroupNotKnownException
	{
		List<AuthzBean> authzList = authz.getGroupAuthZRaw(srcGb, false, sqlMap);
		for (AuthzBean ab: authzList)
		{
			ab.setGroupId(dstG);
			try
			{
				if (ab.getSpecial() != null)
					sqlMap.insert("insertGroupAuthZ", ab);
				else
					sqlMap.insert("insertSpecialAuthZ", ab);
			} catch (PersistenceException e)
			{
				log.error("DB error: Can't add authZ data for group id " + 
					dstG + ". SQL exception is: " + e);
				throw new InternalException("DB error: adding " +
					"authZ for group id " + dstG);  
			}
		}
	}
	
	private void copyMembers(int srcG, int dstG, SqlSession sqlMap) throws InternalException
	{
		List ids = getGroupIdentitiesFullRaw(srcG, null, sqlMap);
		GroupBean idInsParam = new GroupBean();
		idInsParam.setId(dstG);
		for (Object id: ids)
		{
			IdentityBean ib = (IdentityBean) id;
			EntityBean eb = null;
			try
			{
				eb = entitiesCache.getEntity4Identity(ib.getEntityId(), sqlMap);
				idInsParam.setParent(eb.getId());
			} catch (IdentityNotKnownException e){}
			idInsParam.setCreationTS(new Timestamp((new Date()).getTime()));
			try
			{
				sqlMap.insert("addToGroup", idInsParam);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't add to group id " + 
					dstG + " entity with id " + eb.getId();
				log.error(msg + ". SQL exception is: " + e);
				throw new InternalException(msg);  
			}
		}
	}
	
	private void copyGroupAttributes(int srcG, int dstG, SqlSession sqlMap) throws InternalException
	{
		List rawAttrs = null;
		try
		{
			rawAttrs = sqlMap.selectList("getAllGroupAttributes", srcG);
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve group attributes for group" + 
				srcG; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
		for (Object rAttr: rawAttrs)
		{
			AttributeBean attr = (AttributeBean) rAttr;
			attr.setGroupId(dstG);
			attr.setCreationTS(new Timestamp(new Date().getTime()));			
			try
			{
				sqlMap.insert("addGroupAttribute", attr);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't insert group attribute " +
					"for group no " + dstG; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
		}		
	}
	
	/**
	 * Assigns the given identity to a group.
	 * @param toAdd
	 * @param group
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 * @throws InvalidValueException 
	 */
	public void addToGroup(Identity toAdd, Group group, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, 
		GroupNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(group, null, sqlMap);
			if (gb.getId() == -1)
				throw new GroupNotKnownException(group);
			EntityBean eb = entitiesCache.getEntity4Identity(toAdd, null, sqlMap);
			gb.setParent(eb.getId());
			gb.setCreationTS(new Timestamp((new Date()).getTime()));
			try
			{
				if (isMember(eb.getId(), gb.getId(), null, sqlMap))
					return;
				sqlMap.insert("addToGroup", gb);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't add to group " + 
					group + " identity " + toAdd;
				log.error(msg + ". SQL exception is: " + e);
				throw new InternalException(msg);  
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/**
	 * Removes the given identity from the given group.
	 * @param toRemove
	 * @param group
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 * @throws InvalidValueException 
	 */
	public void removeFromGroup(Identity toRemove, Group group, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException,
		GroupNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(group, null, sqlMap);
			EntityBean eb = entitiesCache.getEntity4Identity(toRemove, null, sqlMap);
			if (!isMember(eb.getId(), gb.getId(), null, sqlMap))
				return;
			removeGroupMemberEntity(eb.getId(), gb.getId(), null, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/**
	 * Removes the given entity from all groups.
	 * @param toRemove
	 * @param group
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 */
	void removeFromAllGroups(int entityId, SqlSession sqlMap) 
		throws InternalException
	{
		GroupBean gb = new GroupBean();
		gb.setRemovalTS(new Timestamp(new Date().getTime()));
		gb.setId(entityId);
		try
		{
			sqlMap.update("removeFromAllGroups", gb);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't remove from all groups " + 
				" entity " + entityId;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
	}

	
	/**
	 * Adds a new group attribute. 
	 * @param whom Who is the owner of the attribute.
	 * @param toAdd Attribute to add
	 * @param update If true and attribute with the same name exists it will be updated. If false
	 * and attribute with the same name exists there will be exception 
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws ElementAlreadyExistsException if update is false and attribute with the same 
	 * name exists
	 * @throws UnsupportedTypeException 
	 * @throws GroupNotKnownException 
	 * @throws IdentityNotKnownException 
	 */
	public void setAttribute(Group whom, Attribute toAdd, boolean update, SqlSession sqlMap) 
		throws InternalException, ElementAlreadyExistsException, 
		GroupNotKnownException
	{
		lock.writeLock();
		try
		{
			AttributeBean ab = new AttributeBean();
			AttributeType type = toAdd.getType();
			int typeId;
			try
			{
				typeId = attributeTypes.getAttrTypeId(type, true, sqlMap);
			} catch (UnsupportedTypeException e1)
			{
				log.fatal("BUG: Unsupported type thrown when type should be created.");
				throw new InternalException("BUG in set attribute");
			}
	
			GroupBean gb = groupResolver.resolveGroup(whom, null, sqlMap);
			if (gb.getId() == -1)
				throw new GroupNotKnownException(whom);
	
			ab.setGroupId(gb.getId());
			ab.setCreationTS(new Timestamp(new Date().getTime()));
			ab.setTypeId(typeId);
			ab.setInactive(0);
			
			List<AttributeBean> rawAttrs = getAttributes(gb.getId(), null, sqlMap);
	
			for (AttributeBean ab2: rawAttrs)
			{
				if (ab2.getTypeId() == typeId && ab2.getGroupId() == ab.getGroupId())
				{
					if (!update)
					{
						throw new ElementAlreadyExistsException(
								new Element(null, null));
					}
					ab2.setRemovalTS(new Timestamp(new Date().getTime()));
					try
					{
						sqlMap.update("removeGroupAttribute", ab2);
					} catch (PersistenceException e)
					{
						String msg = "DB error: Can't remove (to update) group" +
							" attributes of group" + whom; 
						log.warn(msg +". SQL exception is: " + e);
						throw new InternalException(msg);
					}
				}
			}
			
			List<String> vals = toAdd.getValues();
			if (vals == null)
				vals = new ArrayList<String>();
			ab.setVal(null);
			if (vals.size() == 0)
				try
				{
					sqlMap.insert("addGroupAttribute", ab);
				} catch (PersistenceException e)
				{
					String msg = "DB error: Can't insert " +
						"group attribute for group " + whom; 
					log.warn(msg +". SQL exception is: " + e);
					throw new InternalException(msg);
				}
				
			for (String val: vals)
			{
				ab.setVal(val);
				try
				{
					sqlMap.insert("addGroupAttribute", ab);
				} catch (PersistenceException e)
				{
					String msg = "DB error: Can't insert group attribute for group " + 
						whom; 
					log.warn(msg +". SQL exception is: " + e);
					throw new InternalException(msg);
				}
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/**
	 * Removes the group attribute. 
	 * @param whom
	 * @param toRemove
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws UnsupportedTypeException 
	 */
	public void removeAttribute(Group whom, String toRemove, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, UnsupportedTypeException
	{
		lock.writeLock();
		try
		{
			int typeId = attributeTypes.getAttrTypeId(toRemove);
			GroupBean gb = groupResolver.resolveGroup(whom, null, sqlMap);
			if (gb.getId() == -1)
				throw new GroupNotKnownException(whom);
			
			List<AttributeBean> rawAttrs = getAttributes(gb.getId(), null, sqlMap);
	
			for (AttributeBean ab2: rawAttrs)
			{
				if (ab2.getTypeId() == typeId)
				{
					ab2.setRemovalTS(new Timestamp(new Date().getTime()));
					try
					{
						sqlMap.update("removeGroupAttribute", ab2);
					} catch (PersistenceException e)
					{
						String msg = "DB error: Can't remove group" +
							" attributes of group" + whom; 
						log.warn(msg +". SQL exception is: " + e);
						throw new InternalException(msg);
					}
				}
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/**
	 * Disables the group attribute. 
	 * @param whom
	 * @param toDisable
	 * @param value
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws AttributeNotFoundException 
	 * @throws UnsupportedTypeException 
	 */
	public void disableEnableAttribute(Group whom, String toDisable, String value, int how, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, AttributeNotFoundException
	{
		lock.writeLock();
		try
		{
			int typeId;
			try
			{
				typeId = attributeTypes.getAttrTypeId(toDisable);
			} catch (UnsupportedTypeException e1)
			{
				throw new AttributeNotFoundException(null);
			}
			GroupBean gb = groupResolver.resolveGroup(whom, null, sqlMap);
			if (gb.getId() == -1)
				throw new GroupNotKnownException(whom);
			
			AttributeBean ab = new AttributeBean();
			ab.setGroupId(gb.getId());
			ab.setTypeId(typeId);
			ab.setVal(value);
			ab.setInactive(how);
			
			try
			{
				int ret;
				if (value == null)
					ret = sqlMap.update("disableEnableGroupAttributes1", ab);
				else
					ret = sqlMap.update("disableEnableGroupAttributes2", ab);
				if (ret == 0)
					throw new AttributeNotFoundException(toDisable);
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't disable/enable group " +
					" attributes of group " + whom; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	
	/**
	 * Returns a list with attributes that have disabled values. Only those disabled
	 * values are in attributes returned. 
	 * @param whose
	 * @param ctx
	 * @return
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 */
	public List<Attribute> getDisabledAttributes(Group whose, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException
	{
		lock.writeLock();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(whose, null, sqlMap);
	
			List rawL = null;
			try
			{
				rawL = sqlMap.selectList("listDisabledGroupAttributes", 
						gb.getId());
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't list disabled group" +
				" attributes of a group" + whose; 
				log.warn(msg +". SQL exception is: " + e);
				throw new InternalException(msg);
			}
			return attributeTypes.prepareAttributeList(rawL, whose.toString());
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	/**
	 * Checks if given identity is a memeber of the given group.
	 * <p>
	 * Requires read perm group/global.
	 * @param who
	 * @param group 
	 * @return
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 * @throws InvalidValueException 
	 */
	public boolean isMember(Identity who, boolean effective, Group group, Date when, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, 
		GroupNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(group, when, sqlMap);
			EntityBean eb = entitiesCache.getEntity4Identity(who, when, sqlMap);
			if (effective)
				return isEffectiveMember(eb.getId(), gb, when, sqlMap);
			else
				return isMember(eb.getId(), gb.getId(), when, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/**
	 * Retrieves content of the group. Elements are either other groups or identities.
	 * <p>
	 * Requires global or group read perm.
	 * @param group
	 * @return
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 */
	public List<Element> getGroupContent(Group group, Date when, SqlSession sqlMap)
		throws InternalException, GroupNotKnownException
	{
		lock.writeLock();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(group, when, sqlMap);
			List<Element> ret = getGroupIdentitiesFull(group, gb.getId(), when, sqlMap);
			ret.addAll(getGroupSubgroupsFull(group, gb.getId(), when, sqlMap));
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public List<Attribute> getAllAttributes(Group owner, Date when, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException
	{
		lock.writeLock();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(owner, when, sqlMap);
			List<AttributeBean> attrBs = getAttributes(gb.getId(), when, sqlMap);
			return attributeTypes.prepareAttributeList(attrBs, owner.toString());
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	/*
	 * Return path from bottom to top group inclusive, or empty list if there is no path.
	 */
	List<GroupBean> getGroupPath(GroupBean topGroup, GroupBean bottomGroup, SqlSession sqlMap) 
		throws GroupNotKnownException, InternalException
	{
		lock.writeLock();
		try
		{
			Vector<GroupBean> ret = new Vector<GroupBean>();
			GroupBean tmp  = new GroupBean();
			tmp.setId(bottomGroup.getId());
			tmp.setParent(bottomGroup.getParent());
			tmp.setCreationTS(bottomGroup.getCreationTS());
			tmp.setName(bottomGroup.getName());
			while (tmp.getId() != topGroup.getId() && tmp.getParent() != -1)
			{
				ret.add(tmp);
				tmp = groupResolver.resolveGroup4Id(tmp.getParent(), sqlMap);
			}
			if (tmp.getId() != topGroup.getId())
				ret.clear();
			else
				ret.add(tmp);
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	public List<GroupBean> getAllGroupsRaw(int who, Date when, SqlSession sqlMap)
		throws InternalException, IdentityNotKnownException
	{
		lock.writeLock();
		try
		{
			List res;
			try
			{
				if (when == null)
					res = sqlMap.selectList("getEntityGroups", who);
				else
					res = sqlMap.selectList("getEntityGroupsH", 
							new GenericBean(who, when));
			} catch (PersistenceException e)
			{
				String msg = "DB error: Can't get groups the entity " + who + 
					" is member of.";
				log.error(msg + ". SQL exception is: " + e);
				throw new InternalException(msg);  
			}
			List<GroupBean> ret = new Vector<GroupBean>();
			for (Object o: res)
			{
				ret.add((GroupBean)o);
			}
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public List<Group> getAllGroups(Identity who, boolean implied, Date when, SqlSession sqlMap)
		throws InternalException, IdentityNotKnownException, InvalidValueException
	{
		lock.writeLock();
		try
		{
			EntityBean eb = entitiesCache.getEntity4Identity(who, when, sqlMap);
			List<GroupBean> rawGroups = getAllGroupsRaw(eb.getId(), when, sqlMap); 
			List<Group> ret = new Vector<Group>();
			for (GroupBean gb: rawGroups)
			{
				Group g;
				try
				{
					g = groupResolver.resolveGroupBean(gb, sqlMap);
					ret.add(g);
				} catch (GroupNotKnownException e){} 
			}
			if (implied)
				return getImpliedGroups(ret);
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	//FIXME - this method should be moved to the query module, when synchronization will be updated to locks.
	public List<Group> getAllReadableGroups(boolean hasGlobalReadPerm, Identity who, 
		boolean implied, InternalAuthorization authzInternal, SqlSession sqlMap) 
		throws IdentityNotKnownException, InternalException, InvalidValueException, SecurityException
	{
		lock.writeLock();
		try
		{
			EntityBean eb = entitiesCache.getEntity4Identity(who, null, sqlMap);
			List<GroupBean> rawGroups = getAllGroupsRaw(eb.getId(), null, sqlMap); 
			Set<GroupBean> fullRawGroups = new HashSet<GroupBean>(rawGroups);
			
			if (implied)
			{
				List<GroupBean> impliedGroups;
				for (GroupBean gb: rawGroups)
				{
					try
					{
						impliedGroups = groupResolver.getParentGroupBeans(gb, sqlMap);
						fullRawGroups.addAll(impliedGroups);
					} catch (GroupNotKnownException cantHappen) {
						log.fatal("Group not known in getAllReadableGroups - BUG", cantHappen);
					}
				}
			}
			
			boolean self = authzInternal.isSelf(who, sqlMap);
			List<Group> ret = new ArrayList<Group>();
			boolean groupReadable;
			for (GroupBean gb: fullRawGroups)
			{
				Group g;
				groupReadable = hasGlobalReadPerm;
				try
				{
					if (!groupReadable)
					{
						Permissions localP = authzInternal.assemblePermissions(gb, self, sqlMap);
						groupReadable = localP.isRead();
					}
					if (groupReadable)
					{
						g = groupResolver.resolveGroupBean(gb, sqlMap);
						ret.add(g);
					}
				} catch (GroupNotKnownException cantHappen) {
					log.fatal("Group not known in getAllReadableGroups - BUG", cantHappen);
				}
			}

			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public Attribute getGroupAttribute(Group owner, String attribute, Date when, SqlSession sqlMap) 
		throws InternalException, ElementNotKnownException, 
		AttributeNotFoundException
	{
		lock.writeLock();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(owner, when, sqlMap);
			int attrType;
			try
			{
				attrType = attributeTypes.getAttrTypeId(attribute);
			} catch (UnsupportedTypeException e)
			{
				throw new AttributeNotFoundException(attribute);
			}
			List<AttributeBean> attrBs = getAttribute(gb.getId(), attrType, when, sqlMap);
			if (attrBs.size() == 0)
				throw new AttributeNotFoundException(attribute);
			return attributeTypes.prepareAttributeList(attrBs, owner.toString()).get(0);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	List<AttributeBean> getAttributes(int groupId, Date when, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List rawAttrs = null;
			try
			{
				if (when == null)
					rawAttrs = sqlMap.selectList("getGroupAttributes", groupId);
				else
					rawAttrs = sqlMap.selectList("getGroupAttributesH", 
							new GenericBean(groupId, when));
			} catch (PersistenceException e1)
			{
				String msg = "DB error: Can't retrieve group attributes for group" + 
					groupId; 
				log.warn(msg +". SQL exception is: " + e1);
				throw new InternalException(msg);
			}
			Vector<AttributeBean> ret = new Vector<AttributeBean>();
			for (Object o: rawAttrs)
				ret.add((AttributeBean) o);
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public void getGroupEvents(List<ModificationEvent> list, Date from, Date to, 
			int typeFilter, boolean fillDetails, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List<Object> res1 = new ArrayList<Object>(), res2 = new ArrayList<Object>();
			getEventsGeneric(res1, res2, from, to, typeFilter, "getEventsGrp", sqlMap);
			convertEventsGrp(list, res1, fillDetails, TYPE.ADD, sqlMap);
			convertEventsGrp(list, res2, fillDetails, TYPE.REMOVE, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	public void getMembershipEvents(List<ModificationEvent> list, Date from, Date to, 
			int typeFilter, boolean fillDetails, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List<Object> res1 = new ArrayList<Object>(), res2 = new ArrayList<Object>();
			getEventsGeneric(res1, res2, from, to, typeFilter, "getEventsMem", sqlMap);
			convertEventsMem(list, res1, fillDetails, TYPE.ADD, sqlMap);
			convertEventsMem(list, res2, fillDetails, TYPE.REMOVE, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	void getAttributeEvents(List<ModificationEvent> list, Date from, Date to, 
			int typeFilter, boolean fillDetails, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List<Object> res1 = new ArrayList<Object>(), res2 = new ArrayList<Object>();
			getEventsGeneric(res1, res2, from, to, typeFilter, "getEventsAttrGrp", sqlMap);
			convertEventsAttr(list, res1, fillDetails, TYPE.ADD, sqlMap);
			convertEventsAttr(list, res2, fillDetails, TYPE.REMOVE, sqlMap);
		}
		finally
		{
			lock.writeUnlock();
		}
	}
	
	
	
	
	//************************** private auxiliary methods ******************************
	private void convertEventsGrp(List<ModificationEvent> list, List<Object> res, 
			boolean fillDetails, TYPE type, SqlSession sqlMap) throws InternalException
	{
		boolean add = type.equals(TYPE.ADD);
		for (Object o: res)
		{
			GroupBean g = (GroupBean)o;
			ModificationEvent e = new ModificationEvent(
					add ? g.getCreationTS() : g.getRemovalTS(), 
					ModificationEvent.CATEGORY.GROUP,
					type);
			
			if (fillDetails)
				try
				{
					e.setDescription(groupResolver.resolveGroupBean(g, sqlMap).toString());
				} catch (GroupNotKnownException e1)
				{
					throw new InternalException("Problem resolving " +
							"group bean", e1);
				}
			list.add(e);
		}		
	}

	private void convertEventsMem(List<ModificationEvent> list, List<Object> res, 
			boolean fillDetails, TYPE type, SqlSession sqlMap) throws InternalException
	{
		boolean add = type.equals(TYPE.ADD);
		for (Object o: res)
		{
			ElementBean g = (ElementBean)o;
			ModificationEvent e = new ModificationEvent(
					add ? g.getCreationTS() : g.getRemovalTS(), 
					ModificationEvent.CATEGORY.MEMBERSHIP,
					type);
			
			if (fillDetails)
			{
				String grp;
				try
				{
					GroupBean gb = new GroupBean();
					gb.setId(g.getGroupId());
					gb.setName(g.getGroupName());
					gb.setParent(g.getParent());
					grp = groupResolver.resolveGroupBean(gb, sqlMap).toString();
				} catch (GroupNotKnownException e1)
				{
					throw new InternalException("Problem resolving " +
							"group bean", e1);
				}
				e.setDescription("Entity: " + g.getEntityName() + 
						" group: " + grp);
			}
			list.add(e);
		}		
	}

	
	private void createGroup(int parent, String name, SqlSession sqlMap) throws InternalException
	{
		GroupBean param = new GroupBean();
		param.setName(name);
		param.setParent(parent);
		try
		{
			sqlMap.insert("insertGroup", param);
			try
			{
				param = groupResolver.resolveGroup(name, parent, null, sqlMap);
			} catch(GroupNotKnownException e) {}
			if (parent != -1)
				sqlMap.insert("insertGroupChild", param);
			sqlMap.clearCache();
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't add a new group " + name + 
				" parent " + parent;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
	}
	
	private void removeGroup(int id, Timestamp ts, SqlSession sqlMap) throws InternalException
	{
		GroupBean param = new GroupBean();
		param.setId(id);
		if (ts == null)
			ts = new Timestamp(new Date().getTime());
		param.setRemovalTS(ts);
		try
		{
			sqlMap.update("deleteGroup", param);
			sqlMap.update("deleteFromSubgroups", param);
			sqlMap.clearCache();
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't delete group " + id;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
		
	}

	public boolean isEffectiveMember(int entityId, GroupBean groupB, Date when, SqlSession sqlMap) 
		throws InternalException, IdentityNotKnownException, GroupNotKnownException
	{
		lock.writeLock();
		try
		{
			List<GroupBean> entityGroups = getAllGroupsRaw(entityId, when, sqlMap);
			for (GroupBean gb: entityGroups)
			{
				List<GroupBean> ret = getGroupPath(groupB, gb, sqlMap);
				if (ret.size() > 0)
					return true;
			}
			return false;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	
	private boolean isMember(int entityId, int groupId, Date when, SqlSession sqlMap) 
		throws InternalException
	{
		GroupBean gb = new GroupBean();
		gb.setId(groupId);
		gb.setParent(entityId);
		Object ret;
		try
		{
			if (when == null)
				ret = sqlMap.selectOne("isMember", gb);
			else
			{
				gb.setRemovalTS(new Timestamp(when.getTime()));
				ret = sqlMap.selectOne("isMemberH", gb);
			}
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't check memebership of " + entityId +
				" in group " + groupId;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
		if (ret == null)
			return false;
		return true;
	}

	private void removeGroupRecursive(int groupId, boolean recursive, Timestamp ts, SqlSession sqlMap) 
		throws InternalException, GroupNotEmptyException
	{
		List<Integer> subgroups = getGroupSubgroups(groupId, sqlMap);
		List<Integer> entities = getGroupEntities(groupId, sqlMap);
		boolean isEmpty = true;
		if (subgroups.size() > 0 || entities.size() > 0)
			isEmpty = false;
		
		if (!isEmpty)
		{
			if (!recursive)
				throw new GroupNotEmptyException();
			
			//remove all subgroups
			for (Integer subId: subgroups)
				removeGroupRecursive(subId, recursive, ts, sqlMap);
			
			//remove all member entities
			for (Integer entityId: entities)
				removeGroupMemberEntity(entityId, groupId, ts, sqlMap);
		}
		
		//remove group attributes
		AttributeBean ab = new AttributeBean();
		ab.setRemovalTS(ts);
		ab.setId(groupId);
		try
		{
			sqlMap.update("removeAllGroupAttributes", ab);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't remove all group attributes for group" + 
				groupId; 
			log.warn(msg +". SQL exception is: " + e);
			throw new InternalException(msg);
		}
		
		//remove entity attributes bound to this group
		GroupBean gb = new GroupBean();
		gb.setId(groupId);
		gb.setRemovalTS(ts);
		try
		{
			sqlMap.update("removeGroupBoundEntityAttributes", gb);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't remove group bound entity attributes" +
					" for group " + groupId;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
		
		//remove authZ data bound to this group
		removeGroupAuthZ(groupId, sqlMap);
		
		//remove application forms bound to this group
		dbApps.removeAppFormsByGroup(groupId, sqlMap);
		
		//remove the group
		removeGroup(groupId, ts, sqlMap);
	}
	
	private void removeGroupAuthZ(int groupId, SqlSession sqlMap) throws InternalException
	{
		try
		{
			sqlMap.delete("removeGroupAuthZAttr", groupId);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't remove group authZ data for group " + 
				groupId;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
		try
		{
			sqlMap.delete("removeGroupAuthZSpecial", groupId);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't remove group authZ Owner data for group " + 
				groupId;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
	}
	
	
	private void removeGroupMemberEntity(int entityId, int groupId, Timestamp ts, SqlSession sqlMap)
		throws InternalException
	{
		GroupBean gb = new GroupBean();
		gb.setId(groupId);
		if (ts == null)
			ts = new Timestamp(new Date().getTime());
		gb.setRemovalTS(ts);
		gb.setParent(entityId);
		try
		{
			sqlMap.update("removeFromGroup", gb);
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't remove from group " + 
				groupId + " identity " + 
				entityId;
			log.error(msg + ". SQL exception is: " + e);
			throw new InternalException(msg);  
		}
	}
	
	private List<AttributeBean> getAttribute(int groupId, int typeId, Date when, SqlSession sqlMap) 
		throws InternalException
	{
		List rawAttrs = null;
		AttributeBean param = new AttributeBean();
		param.setGroupId(groupId);
		param.setTypeId(typeId);
		try
		{
			if (when == null)
				rawAttrs = sqlMap.selectList("getGroupAttribute", param);
			else
			{
				param.setRemovalTS(new Timestamp(when.getTime()));
				rawAttrs = sqlMap.selectList("getGroupAttributeH", param);
			}
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve group attributes for group" + 
				groupId; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
		Vector<AttributeBean> ret = new Vector<AttributeBean>();
		for (Object o: rawAttrs)
			ret.add((AttributeBean) o);
		return ret;
	}
	
	private List<Integer> getGroupEntities(int groupId, SqlSession sqlMap) throws InternalException
	{
		List rawAttrs = null;
		try
		{
			rawAttrs = sqlMap.selectList("getGroupEntities", groupId);
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve group entities for group" + 
				groupId; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
		Vector<Integer> ret = new Vector<Integer>();
		for (Object o: rawAttrs)
			ret.add((Integer) o);
		return ret;		
	}
	
	public List<Integer> getGroupSubgroups(int groupId, SqlSession sqlMap) throws InternalException
	{
		lock.writeLock();
		try
		{
			List rawAttrs = null;
			try
			{
				rawAttrs = sqlMap.selectList("getGroupSubgroups", groupId);
			} catch (PersistenceException e1)
			{
				String msg = "DB error: Can't retrieve group subgroups for group" + 
					groupId; 
				log.warn(msg +". SQL exception is: " + e1);
				throw new InternalException(msg);
			}
			Vector<Integer> ret = new Vector<Integer>();
			for (Object o: rawAttrs)
				ret.add((Integer) o);
			return ret;
		}
		finally
		{
			lock.writeUnlock();
		}
	}

	private List<Element> getGroupIdentitiesFull(Group g, int groupId, Date when, SqlSession sqlMap) throws InternalException
	{
		List raw = getGroupIdentitiesFullRaw(groupId, when, sqlMap);
		Vector<Element> ret = new Vector<Element>();
		for (Object o: raw)
		{
			IdentityBean ib = (IdentityBean) o;
			EntityBean eb = null;
			try
			{
				eb = entitiesCache.getEntity4Identity(ib.getId(), sqlMap);
			} catch (IdentityNotKnownException e){throw new InternalException("BUG: Identity not known", e);}
			Identity i = new Identity(identityTypes.getIdentityType4Id(ib.getTypeId()), 
				ib.getValue(), eb.getName(), ib.getCreationTS(), null);
			ret.add(new Element(null, i));
		}
		return ret;		
	}

	private List getGroupIdentitiesFullRaw(int groupId, Date when, SqlSession sqlMap) throws InternalException
	{
		List raw = null;
		try
		{
			if (when == null)
				raw = sqlMap.selectList("getGroupIdentities", groupId);
			else
				raw = sqlMap.selectList("getGroupIdentitiesH", 
						new GenericBean(groupId, when));
			return raw;
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve group identities for group id " + 
				groupId; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
	}
	
	private List<Element> getGroupSubgroupsFull(Group parent, int groupId, Date when, SqlSession sqlMap) 
		throws InternalException
	{
		List raw = getGroupSubgroupsFullRaw(groupId, when, sqlMap);
		Vector<Element> ret = new Vector<Element>();
		if (parent == null)
			parent = new Group(new String[0]);
		for (Object o: raw)
		{
			GroupBean gb = (GroupBean)o;
			Group g = new Group((parent.isTopLevel() ? "" : parent.toString()) 
					+ "/" + gb.getName());
			ret.add(new Element(g, null));
		}
		return ret;		
	}
	
	private List getGroupSubgroupsFullRaw(int groupId, Date when, SqlSession sqlMap) 
		throws InternalException
	{
		List raw = null;
		try
		{
			if (when == null)
			{
				if (groupId != -1)
					raw = sqlMap.selectList("getGroupSubgroupsFull", groupId);
				else
					raw = sqlMap.selectList("getTopGroupsFull");
			} else
			{
				if (groupId != -1)
					raw = sqlMap.selectList("getGroupSubgroupsFullH", 
							new GenericBean(groupId, when));
				else
					raw = sqlMap.selectList("getTopGroupsFullH",
							new GenericBean(when));
			}
			return raw;
		} catch (PersistenceException e1)
		{
			String msg = "DB error: Can't retrieve group subgroups for group" + 
			groupId; 
			log.warn(msg +". SQL exception is: " + e1);
			throw new InternalException(msg);
		}
	}
	
	private List<Group> getImpliedGroups(List<Group> base)
	{
		Set<String> allG = new HashSet<String>();
		for (Group g: base)
		{
			String name = g.toString();
			allG.add(name);
			int n = name.lastIndexOf('/');
			while (n > 0)
			{
				name = name.substring(0, n);
				if (!allG.contains(name))
					allG.add(name);
				n = name.lastIndexOf('/');
			}
		}
		Iterator<String> it = allG.iterator();
		base.clear();
		while (it.hasNext())
			base.add(new Group(it.next()));
		return base;
	}
	

	@SuppressWarnings("unchecked")
	void getEventsGeneric(List<Object> res1, List<Object> res2, Date from, Date to, 
			int typeFilter, String cmd, SqlSession sqlMap) throws InternalException
	{
		GenericBean param = new GenericBean(from, to);
		try
		{
			if ((typeFilter & (1 << TYPE.ADD.ordinal())) != 0)
				res1.addAll(sqlMap.selectList(cmd + "Add",
					param));
			if ((typeFilter & (1 << TYPE.REMOVE.ordinal())) != 0)
				res2.addAll(sqlMap.selectList(cmd + "Del",
					param));
		} catch (PersistenceException e)
		{
			String msg = "DB error: Can't get list of history events."; 
			log.warn(msg + " Operation: " + cmd + 
					". SQL exception is: " + e);
			throw new InternalException(msg);
		}
	}

	void convertEventsAttr(List<ModificationEvent> list, List<Object> res, 
			boolean fillDetails, TYPE type, SqlSession sqlMap) throws InternalException
	{
		boolean add = type.equals(TYPE.ADD);
		for (Object o: res)
		{
			AttributeWithOwnerBean g = (AttributeWithOwnerBean)o;
			ModificationEvent e = new ModificationEvent(
					add ? g.getCreationTS() : g.getRemovalTS(), 
					ModificationEvent.CATEGORY.ATTRIBUTE,
					type);
			
			if (fillDetails)
			{
				String grp = null;
				String entity = null;
				if (g.getGroupId() != -1)
					try
					{
						GroupBean gb = new GroupBean();
						gb.setId(g.getGroupId());
						gb.setName(g.getGroupName());
						gb.setParent(g.getParent());
						grp = groupResolver.resolveGroupBean(gb, sqlMap).toString();
					} catch (GroupNotKnownException e1)
					{
						throw new InternalException(
							"Problem resolving " +
							"group bean", e1);
					}
				if (g.getEntityId() != -1)
					entity = g.getEntityName();

				String a = attributeTypes.getAttrType4Id(g.getAttrTypeId());
				
				String owner = "";
				if (grp != null && entity == null)
					owner = "group: " + grp;
				else if (grp != null && entity != null)
					owner = "entity: " + entity + " in scope of group: "
						+ grp;
				else
					owner = "entity: " + entity;
					
				e.setDescription("Attribute: " + a + " value: " + 
					g.getAttrVal() + " owned by: " + owner);
			}
			list.add(e);
		}		
	}
}























