/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on May 13, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.engine;

import java.util.Date;

import org.apache.ibatis.session.SqlSession;

import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.api.exceptions.AuthenticationException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.model.GroupBean;

/**
 * Methods used internally to perform authorization (do not confuse
 * with external authorization API which is used to manage rules
 * basing on which this methods works).
 *  
 * @author K. Benedyczak
 */
public interface InternalAuthorization
{
	/**
	 * Creates security context for the given caller. It can be used later 
	 * in all other operations of this interface. Authentication is performed
	 * if there is authentication data in argument or in database for the argument. 
	 * Otherwise it is assumed that identity was verified earlier by other means.
	 * @param caller identity of the caller 
	 * @return a new context
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws AuthenticationException
	 */
	public SecurityContext establishSecurityContext(Identity caller) 
		throws IdentityNotKnownException, InternalException, 
		AuthenticationException, InvalidValueException;
	
	/**
	 * Authorize in scope of a group. 
	 * 
	 * @param permMask what permissions are required
	 * @param group group for which permissions are checked 
	 * @param self if true then allow to get also permissions which are 
	 * 	designated for owner (self) access 
	 * @throws SecurityException if authorization failed
	 * @throws InternalException if error occured
	 * @throws GroupNotKnownException if group isn't known
	 */
	public void authorize(Permissions permMask, GroupBean group, boolean self, SqlSession sqlMap) 
		throws SecurityException, InternalException, GroupNotKnownException;

	/**
	 * Authorize in scope of a group and all it's subgroups. 
	 * 
	 * @param permMask what permissions are required
	 * @param group group for which permissions are checked 
	 * @throws SecurityException if authorization failed
	 * @throws InternalException if error occured
	 * @throws GroupNotKnownException if group isn't known
	 */
	public void authorizeRecursive(Permissions permMask, GroupBean group, SqlSession sqlMap) 
		throws SecurityException, InternalException, GroupNotKnownException;

	
	/**
	 * Authorize in global scope.
	 * 
	 * @param permMask what permissions are required
	 * @param self if true then allow to get also permissions which are 
	 * 	designated for owner (self) access
	 * @throws SecurityException if authorization failed
	 * @throws InternalException if error occured
	 */
	public void authorize(Permissions permMask, boolean self, SqlSession sqlMap) 
		throws SecurityException, InternalException;
	
	/**
	 * Authorize identity control operation. It is equivalent to check 
	 * global write or global identityCtl.
	 * @param id
	 * @param checkGlobal
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public void authorizeIdentityCtl(SqlSession sqlMap)
		throws SecurityException, InternalException;	

	/**
	 * Authorize identity control operation. It is equivalent to global write or 
	 * (global identityCtl + write for every group id is a memeber of + 
	 * check if caller has at least the same global permissions as id).   
	 * @param id
	 * @param checkGlobal
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws InvalidValueException 
	 * @throws IdentityNotKnownException 
	 */
	public void authorizeIdentityCtl(Identity id, SqlSession sqlMap)
		throws SecurityException, InternalException, IdentityNotKnownException,
		InvalidValueException;	

	
	/**
	 * Checks if the given identity is the same as owner of the security context.
	 * @param id
	 * @return
	 * @throws IdentityNotKnownException
	 * @throws InternalException 
	 * @throws SecurityException 
	 * @throws InvalidValueException 
	 */
	public boolean isSelf(Identity id, SqlSession sqlMap)
		throws IdentityNotKnownException, InternalException, SecurityException, InvalidValueException;
	
	/**
	 * Checks if the given identity was the same as owner of the security context 
	 * at the specified time.
	 * @param id
	 * @return
	 * @throws IdentityNotKnownException
	 * @throws InternalException 
	 * @throws SecurityException 
	 * @throws InvalidValueException 
	 */
	public boolean isSelf(Identity id, Date when, SqlSession sqlMap) 
		throws IdentityNotKnownException, InternalException, SecurityException,
		InvalidValueException;

	
	/**
	 * Retrieves permissions for the given entity in the group.
	 * @param group
	 * @param entityId
	 * @param owner
	 * @return
	 * @throws GroupNotKnownException
	 * @throws InternalException
	 */
	public Permissions assemblePermissions(GroupBean group, boolean owner, SqlSession sqlMap) 
		throws GroupNotKnownException, InternalException, SecurityException;
	
	/**
	 * Checks if the identity is active.
	 * @param toCheck identity to be checked
	 * @return true if it is active, false otherwise.
	 * @throws IdentityNotKnownException
	 * @throws InternalException
	 * @throws InvalidValueException 
	 */
	public boolean isActive(Identity toCheck) 
		throws IdentityNotKnownException, InternalException, InvalidValueException;
	
	/**
	 * If the argument is of DN type and cert to DN mapping is enabled and if there is no 
	 * such identity in the DB then this method will try to find a X509 type identity with 
	 * the subject as the arguments identity value. The mapped identity will be returned.
	 * In any other case the argument is returned.
	 * @param toCheck
	 * @return
	 * @throws InvalidValueException 
	 * @throws InternalException 
	 * @throws IdentityNotKnownException 
	 */
	public Identity mapIfNeededToX509(Identity toCheck) 
		throws InternalException, InvalidValueException, IdentityNotKnownException;
}
