/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 18, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.engine;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.ibatis.session.SqlSession;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.Notification;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.api.SupportedNotificationActions;
import pl.edu.icm.unicore.uvos.api.exceptions.AttributeNotFoundException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementAlreadyExistsException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotEmptyException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.LabelAlreadyUsedException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.TimeOutOfRangeException;
import pl.edu.icm.unicore.uvos.api.exceptions.UnsupportedTypeException;
import pl.edu.icm.unicore.uvos.db.DB;
import pl.edu.icm.unicore.uvos.db.DBAttributeTypes;
import pl.edu.icm.unicore.uvos.db.DBEntities;
import pl.edu.icm.unicore.uvos.db.DBGroupResolver;
import pl.edu.icm.unicore.uvos.db.DBGroups;
import pl.edu.icm.unicore.uvos.db.DBIdentityTypes;
import pl.edu.icm.unicore.uvos.db.DBNotification;
import pl.edu.icm.unicore.uvos.engine.api.UVOSManagementInterface;
import pl.edu.icm.unicore.uvos.idhelpers.IdentityHelperFactory;
import pl.edu.icm.unicore.uvos.idhelpers.IdentityTypeHelper;
import pl.edu.icm.unicore.uvos.mail.MailSender;
import pl.edu.icm.unicore.uvos.mail.VariablesMap;
import pl.edu.icm.unicore.uvos.model.GroupBean;
import pl.edu.icm.unicore.uvos.model.NotifyBean;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;

/**
 * Main class which handles VO content management. 
 * It implements adding/removing/modifying of various elements.
 * @author K. Benedyczak
 */
public class ManagementModule implements UVOSManagementInterface
{
	private DBNotification dbNotify;
	private InternalAuthorization authZ;
	private MailSender mail;
	private DBAttributeTypes attributeTypes;
	private DBIdentityTypes identityTypes;
	private UVOSServerProperties configuration;
	private DBGroups groups;
	private DBGroupResolver groupResolver;
	private DBEntities entities;
	private DB db;
	
	public ManagementModule(DBNotification dbNotify, InternalAuthorization authZ,
			MailSender mail, DBAttributeTypes attributeTypes, DBIdentityTypes identityTypes,
			UVOSServerProperties configuration, DBGroups groups, DBGroupResolver groupResolver, DBEntities entities,
			DB db)
	{
		this.dbNotify = dbNotify;
		this.authZ = authZ;
		this.mail = mail;
		this.attributeTypes = attributeTypes;
		this.identityTypes = identityTypes;
		this.configuration = configuration;
		this.groups = groups;
		this.groupResolver = groupResolver;
		this.entities = entities;
		this.db = db;
	}
	
	private void checkGroupName(String name) throws InternalException
	{
		if (name.contains("/") || name.equals(""))
			throw new InternalException("Group name is invalid " +
				"(can't be empty or contain '/' character)");	
	}
	
	private void checkAttrName(AttributeType attr) throws InvalidValueException
	{
		String name = attr.getName();
		try
		{
			new URI(name);
		} catch (URISyntaxException e)
		{
			throw new InvalidValueException("Attribute name '" + name 
					+ "' is not an URI.");
		}
	}
	
	private void generateAttributes(Identity whom, SqlSession session)
		throws InternalException, InvalidValueException, SecurityException,
		ElementAlreadyExistsException, UnsupportedTypeException, ElementNotKnownException
	{
		IdentityTypeHelper helper = IdentityHelperFactory.getHelper(whom);
		List<Attribute> attributes = helper.generateAttributes();
		for(Attribute a : attributes)
			setAttributeInternal(new Element(null, whom), a, true, session);
	}
	
	//
	// public API implementation ******************************************************
	//
	
	/**
	 * Adds a new group. Pass null parent to create top-level group (i.e. VO)
	 * <p>
	 * Requires write perm for parent group (or global for top-level groups).
	 * @param parent
	 * @param name
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws ElementAlreadyExistsException
	 */
	public Group addGroup(Group parent, String name) 
		throws SecurityException, InternalException, GroupNotKnownException, 
		ElementAlreadyExistsException
	{
		SqlSession session = db.getSqlSession();
		Group ret;
		try
		{
			authZ.authorize(Permissions.WRITE_P, groupResolver.resolveGroup(parent, null, session), 
					false, session);
			checkGroupName(name);
			ret = groups.addGroup(parent, name, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("addGroup").
			setGroup(ret.toString()));
		return ret;
	}

	/**
	 * Removes the given group.
	 * <p>
	 * Requires parent group (global for top-level groups), toRemove group and 
	 * all toRename group subgroups write perm.
	 * @param toRemove
	 * @param recursive
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws GroupNotEmptyException
	 */
	public void removeGroup(Group toRemove, boolean recursive) 
		throws SecurityException, InternalException, GroupNotKnownException, 
		GroupNotEmptyException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorizeRecursive(Permissions.WRITE_P, 
					groupResolver.resolveGroup(toRemove.getParent(), null, session),
					session);
			groups.removeGroup(toRemove, recursive, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		
		mail.sendNotification(new VariablesMap("removeGroup").
			setGroup(toRemove.toString()));
	}
	
	/**
	 * {@inheritDoc}
	 * @throws InvalidValueException 
	 */
	public Group copyGroup(Group toCopy, Group newParent, String newName, 
			boolean deleteOriginal) 
		throws SecurityException, InternalException, 
		GroupNotKnownException, ElementAlreadyExistsException, InvalidValueException
	{
		Group ret;
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.WRITE_P, groupResolver.resolveGroup(newParent, null, session), 
					false, session);
			authZ.authorizeRecursive(Permissions.WRITE_P, 
					groupResolver.resolveGroup(toCopy.getParent(), null, session), session);

			GroupBean srcGb = groupResolver.resolveGroup(toCopy, null, session);
			if (newName == null)
				newName = srcGb.getName();
			ret = groups.copyGroup(srcGb, newParent, newName, deleteOriginal, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("copyGroup").
			setGroup(toCopy.toString()).
			set(VariablesMap.VAR_TARGET, newParent.toString()).
			set(VariablesMap.VAR_MOVED, deleteOriginal ? "moved" : "copied"));
		return ret;
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void addIdentity(Identity toAdd) 
		throws SecurityException, InternalException, ElementAlreadyExistsException, 
		UnsupportedTypeException, LabelAlreadyUsedException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorizeIdentityCtl(session);
			entities.addIdentity(toAdd, false, session);
			try
			{
				if(configuration.getBooleanValue(UVOSServerProperties.GENERATE_ATTRIBUTES_FROM_DN))
					generateAttributes(toAdd, session);
			}
			catch(ElementNotKnownException e)
			{
				throw new InternalException("ElementNotKnown", e);
			}
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("addIdentity").
			setSubject(toAdd.prettyString()));
	}
	
	/**
	 * {@inheritDoc} 
	 */
	public void addIdentity(Identity toAdd, Identity equivalentIdentity) 
		throws SecurityException, InternalException, ElementAlreadyExistsException,
		IdentityNotKnownException, UnsupportedTypeException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorizeIdentityCtl(equivalentIdentity, session);
			entities.addIdentity(toAdd, equivalentIdentity, false, session);
			try
			{
				if(configuration.getBooleanValue(UVOSServerProperties.GENERATE_ATTRIBUTES_FROM_DN))
					generateAttributes(toAdd, session);
			}
			catch(ElementNotKnownException e)
			{
				throw new InternalException("ElementNotKnown", e);
			}
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("addEquivalentIdentity").
			setSubject(toAdd.prettyString()).
			set(VariablesMap.VAR_EQUIVALENT, equivalentIdentity.prettyString()));		
	}

	/**
	 * {@inheritDoc}
	 */
	public void removeIdentity(Identity toRemove) 
		throws SecurityException, InternalException, IdentityNotKnownException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorizeIdentityCtl(toRemove, session);
			entities.removeIdentity(toRemove, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("removeIdentity").
			setSubject(toRemove.prettyString()));
	}
	
	/**
	 * Adds a new attribute. 
	 * <p>
	 * Requires write perm: global for global entity attributes, or specified group
	 * for group and entity@group attributes. 
	 * @param whom Who is the owner of the attribute.
	 * @param toAdd Attribute to add
	 * @param update If true and attribute with the same name exists it will be updated. If false
	 * and attribute with the same name exists there will be exception 
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws ElementAlreadyExistsException if update is false and attribute with the same 
	 * name exists
	 * @throws InvalidValueException 
	 */
	public void setAttribute(Element whom, Attribute toAdd, boolean update) 
		throws SecurityException, InternalException, ElementNotKnownException,
		ElementAlreadyExistsException, UnsupportedTypeException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			setAttributeInternal(whom, toAdd, update, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("setAttribute").
			setSubject(whom.toString()).
			setAttriubute(toAdd.toString()));
	}

	private void setAttributeInternal(Element whom, Attribute toAdd, boolean update, SqlSession session) 
		throws SecurityException, InternalException, ElementNotKnownException,
		ElementAlreadyExistsException, UnsupportedTypeException, InvalidValueException
	{
		if (whom.getGroup() == null && whom.getIdentity() == null)
			throw new InternalException("null element");
		checkAttrName(toAdd.getType());
		if (whom.getIdentity() == null)
		{
			authZ.authorize(Permissions.WRITE_P, groupResolver.resolveGroup(
					whom.getGroup(), null, session), false, session);
			groups.setAttribute(whom.getGroup(), toAdd, update, session);
		} else
		{
			if (whom.getGroup() != null)
			{
				authZ.authorize(Permissions.WRITE_P, groupResolver.resolveGroup(
						whom.getGroup(), null, session), false, session);
				entities.setAttribute(whom, toAdd, update, session);
			} else
			{
				authZ.authorize(Permissions.WRITE_P, false, session);
				entities.setAttribute(whom, toAdd, update, session);
			}
		}
	}

	
	/**
	 * Removes the attribute. 
	 * <p>
	 * Requires write perm: global for global entity attributes, or specified group
	 * for group and entity@group attributes. 
	 * @param whom
	 * @param toRemove
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws InvalidValueException 
	 */
	public void removeAttribute(Element whom, String toRemove) 
		throws SecurityException, InternalException, ElementNotKnownException,
		UnsupportedTypeException, InvalidValueException
	{
		if (whom.getGroup() == null && whom.getIdentity() == null)
			throw new InternalException("null element");
		SqlSession session = db.getSqlSession();
		try
		{
			if (whom.getIdentity() == null)
			{
				authZ.authorize(Permissions.WRITE_P, groupResolver.resolveGroup(
						whom.getGroup(), null, session), false, session);
				groups.removeAttribute(whom.getGroup(), toRemove, session);
			} else
			{
				if (whom.getGroup() != null)
				{
					authZ.authorize(Permissions.WRITE_P, groupResolver.resolveGroup(
							whom.getGroup(), null, session), false, session);
					entities.removeAttribute(whom, toRemove, session);
				} else
				{
					authZ.authorize(Permissions.WRITE_P, false, session);
					entities.removeAttribute(whom, toRemove, session);
				}
			}
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("removeAttribute").
			setSubject(whom.toString()).
			setAttriubute(toRemove.toString()));
	}

	/**
	 * Assigns the given identity to a group.
	 * <p>
	 * Requires write perm of the specified group. 
	 * @param toAdd
	 * @param group
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 * @throws InvalidValueException 
	 */
	public void addToGroup(Identity toAdd, Group group) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		GroupNotKnownException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.WRITE_P, groupResolver.resolveGroup(
					group, null, session), false, session);
			groups.addToGroup(toAdd, group, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("addToGroup").
			setSubject(toAdd.prettyString()).
			setGroup(group.toString()));		
	}
	
	/**
	 * Removes the given identity from the given group.
	 * <p>
	 * Requires write perm of the specified group. 
	 * @param toRemove
	 * @param group
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 * @throws InvalidValueException 
	 */
	public void removeFromGroup(Identity toRemove, Group group) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		GroupNotKnownException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.WRITE_P, groupResolver.resolveGroup(
					group, null, session), false, session);
			groups.removeFromGroup(toRemove, group, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("removeFromGroup").
			setSubject(toRemove.prettyString()).
			setGroup(group.toString()));		
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void setIdentityLabel(Identity toChange, String label) 
		throws SecurityException, InternalException, 
		IdentityNotKnownException, LabelAlreadyUsedException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorizeIdentityCtl(toChange, session);
			entities.setIdentityLabel(toChange, label, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	/**
	 * Returns list of all types of attributes.
	 * <p>
	 * Requires no perm.
	 * @param ctx
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<AttributeType> getAttributeTypes() 
		throws SecurityException, InternalException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(new Permissions(0), false, session);
			return attributeTypes.getAttributeTypes();
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	/**
	 * {@inheritDoc}
	 * @throws InvalidValueException 
	 */
	public void updateAttributeTypes(List<AttributeType> toUpdate, boolean clear) throws SecurityException, InternalException,
			InvalidValueException
	{
		List<AttributeType> current = attributeTypes.getAttributeTypes();
		List<AttributeType> common = new ArrayList<AttributeType>();
		List<AttributeType> newAt = new ArrayList<AttributeType>();
		for (AttributeType a: toUpdate)
		{
			if (containsAtType(current, a.getName()))
				common.add(a);
			else
				newAt.add(a);
		}
		SqlSession session = db.getSqlSession();
		try
		{
			if (common.size() != 0 || clear)
				authZ.authorize(Permissions.WRITE_P, false, session);
			else
				authZ.authorize(new Permissions(0), false, session);

			if (clear)
			{
				for (AttributeType at: common)
					attributeTypes.deleteAttrType(at, session);
			} else
			{
				for (AttributeType at: common)
					attributeTypes.updateAttrType(at, session);
				for (AttributeType at: newAt)
				{
					checkAttrName(at);
					attributeTypes.addAttrType(at, session);
				}
			}
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	private boolean containsAtType(List<AttributeType> list, String name)
	{
		for (AttributeType at: list)
			if (at.getName().equals(name))
				return true;
		return false;
	}
	
	/**
	 * Returns list of all types of identities.
	 * <p>
	 * Requires no perm.
	 * @param ctx
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<IdentityType> getIdentityTypes() 
		throws SecurityException, InternalException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(new Permissions(0), false, session);
			return identityTypes.getIdentityTypes();
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void disableAttribute(Element whose, String toDisable, String valueToDisable) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		AttributeNotFoundException, InvalidValueException
	{
		switchAttribute(whose, toDisable, valueToDisable, 1);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void enableAttribute(Element whose, String toEnable, String valueToEnable) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		AttributeNotFoundException, InvalidValueException
	{
		switchAttribute(whose, toEnable, valueToEnable, 0);
	}
	
	private void switchAttribute(Element whose, String toSwitch, String value, int how)
		throws SecurityException, InternalException, ElementNotKnownException, 
		AttributeNotFoundException, InvalidValueException
	{
		if (whose.getGroup() == null && whose.getIdentity() == null)
			throw new InternalException("null element");
		SqlSession session = db.getSqlSession();
		try
		{
			if (whose.getIdentity() == null)
			{
				authZ.authorize(Permissions.FULL_READ_P, groupResolver.resolveGroup(
						whose.getGroup(), null, session), false, session);
				groups.disableEnableAttribute(whose.getGroup(), toSwitch, value, how, session);
			} else
			{
				boolean self = authZ.isSelf(whose.getIdentity(), session);
				if (whose.getGroup() != null)
				{
					authZ.authorize(Permissions.FULL_READ_P, groupResolver.resolveGroup(
							whose.getGroup(), null, session), self, session);
					entities.disableEnableAttribute(whose, toSwitch, value, how, session);
				} else
				{
					authZ.authorize(Permissions.FULL_READ_P, self, session);
					entities.disableEnableAttribute(whose, toSwitch, value, how, session);
				}
			}		
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public List<Attribute> getDisabledAttributes(Element whose) 
		throws SecurityException, InternalException, ElementNotKnownException,
		InvalidValueException
	{
		if (whose.getGroup() == null && whose.getIdentity() == null)
			throw new InternalException("null element");
		SqlSession session = db.getSqlSession();
		try
		{
			if (whose.getIdentity() == null)
			{
				authZ.authorize(Permissions.FULL_READ_P, groupResolver.resolveGroup(
						whose.getGroup(), null, session), false, session);
				return groups.getDisabledAttributes(whose.getGroup(), session);
			} else
			{
				boolean self = authZ.isSelf(whose.getIdentity(), session);
				if (whose.getGroup() != null)
				{
					authZ.authorize(Permissions.FULL_READ_P, groupResolver.resolveGroup(
									whose.getGroup(), null, session), self, session);
					return entities.getDisabledAttributes(whose, session);
				} else
				{
					authZ.authorize(Permissions.FULL_READ_P, self, session);
					return entities.getDisabledAttributes(whose, session);
				}
			}		
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	/**
	 * Deletes permanently all service data that is older then deleteFrom (exclusive).
	 * <p>
	 * Requires global write perm.
	 * @param deleteFrom
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws TimeOutOfRangeException
	 */
	public void purgeHistoricalData(Date deleteFrom) 
		throws SecurityException, InternalException, TimeOutOfRangeException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.WRITE_P, false, session);
			db.purgeHistory(deleteFrom, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("purgeHistoricalData").
			set(VariablesMap.VAR_DATE, deleteFrom.toString()));		
	}

	public void setIdentityStatus(Identity toChange, boolean enable) throws SecurityException, 
		InternalException, IdentityNotKnownException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorizeIdentityCtl(toChange, session);
			entities.setIdentityStatus(toChange, enable, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
		mail.sendNotification(new VariablesMap("setIdentityStatus").
			setSubject(toChange.prettyString()).
			set(VariablesMap.VAR_STATUS, enable ? "enabled" : "disabled"));
	}

	public void addNotification(String action, String receipents, String groupFilter) 
		throws SecurityException, InternalException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			notifyAuthorize(groupFilter, session);
			if (!SupportedNotificationActions.isSupported(action))
				throw new InternalException("Action " + action + 
				" is unsupported.");
			dbNotify.addNotification(action, groupFilter, receipents, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	public List<Notification> getNotifications(String action, String groupFilter) throws SecurityException, InternalException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			notifyAuthorize(groupFilter, session);
			List<?> all = dbNotify.getNotifications(action, session);
			List<Notification> ret = new ArrayList<Notification>();
			for (Object o: all)
			{
				NotifyBean nb = (NotifyBean)o;
				if (groupFilter != null && nb.getGroupFilter() == null)
					continue;
				if (groupFilter != null && !nb.getGroupFilter().startsWith(groupFilter))
					continue;
				ret.add(new Notification(nb.getId(), nb.getGroupFilter(),
						nb.getRecipients(), nb.getOperation()));
			}
			return ret;
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	public void removeNotification(int id) 
		throws SecurityException, InternalException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			String groupFilter = dbNotify.getNotificationGroup(id, session);
			notifyAuthorize(groupFilter, session);
			dbNotify.removeNotification(id, session);
			session.commit();
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	private void notifyAuthorize(String groupFilter, SqlSession session) 
		throws SecurityException, InternalException
	{
		try
		{
			authZ.authorize(Permissions.WRITE_P, false, session);
		} catch (SecurityException e)
		{
			if (groupFilter != null)
				try
				{
					authZ.authorize(Permissions.WRITE_P, 
						groupResolver.resolveGroup(new Group(groupFilter), null, session),
						false, session);
				} catch (GroupNotKnownException e1)
				{
					throw e;
				}
			else 
				throw e;
		}
	}
}













