/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Jan 15, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.engine;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.ibatis.session.SqlSession;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.ModificationEvent;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.api.ModificationEvent.CATEGORY;
import pl.edu.icm.unicore.uvos.api.exceptions.AttributeNotFoundException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.TimeOutOfRangeException;
import pl.edu.icm.unicore.uvos.db.DB;
import pl.edu.icm.unicore.uvos.db.DBEntities;
import pl.edu.icm.unicore.uvos.db.DBGroupResolver;
import pl.edu.icm.unicore.uvos.db.DBGroups;
import pl.edu.icm.unicore.uvos.engine.api.UVOSQueryHistoryInterface;
import pl.edu.icm.unicore.uvos.model.GroupBean;

/**
 * Class which handles VO history content quering. 
 * @author K. Benedyczak
 */
public class QueryHistoryModule implements UVOSQueryHistoryInterface
{
	private DB db;
	private DBGroups groups;
	private DBGroupResolver groupResolver;
	private DBEntities entities;
	private InternalAuthorization authZ;
	
	public QueryHistoryModule(DB db, DBGroups groups, DBGroupResolver groupResolver, DBEntities entities,
			InternalAuthorization authZ)
	{
		this.db = db;
		this.groups = groups;
		this.groupResolver = groupResolver;
		this.entities = entities;
		this.authZ = authZ;
	}
	
	//
	// public API implementation ******************************************************
	//

	public boolean isMember(Identity who, Group group, boolean effective, Date when) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		GroupNotKnownException, TimeOutOfRangeException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.FULL_READ_P, authZ.isSelf(who, when, session), session);
			return groups.isMember(who, effective, group, when, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	
	
	public List<Element> getGroupContent(Group group, Date when) 
		throws SecurityException, InternalException, GroupNotKnownException, 
		TimeOutOfRangeException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(group, when, session);
			if (gb.getId() != -1)
				authZ.authorize(Permissions.FULL_READ_P, false, session);
			return groups.getGroupContent(group, when, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	public boolean areEquivalent(Identity i1, Identity i2, Date when) 
		throws SecurityException, InternalException, IdentityNotKnownException, 
		TimeOutOfRangeException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(new Permissions(Permissions.IDENTITY_CTL | Permissions.FULL_READ), 
					authZ.isSelf(i1, when, session) || authZ.isSelf(i2, when, session),
					session);
			return entities.areEquivalent(i1, i2, when, session);		
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	public List<Identity> getAllEquivalents(Identity who, Date when) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		TimeOutOfRangeException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(new Permissions(Permissions.IDENTITY_CTL | Permissions.FULL_READ), 
					authZ.isSelf(who, when, session), session);
			return entities.getAllEquivalents(who, when, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	public List<Group> getAllGroups(Identity who, boolean implied, Date when) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		TimeOutOfRangeException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			boolean self = authZ.isSelf(who, when, session);
			authZ.authorize(Permissions.FULL_READ_P, self, session);
			return groups.getAllGroups(who, implied, when, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	

	public List<Identity> getAllIdentities(Date when) 
		throws SecurityException, InternalException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.FULL_READ_P, false, session);
			return entities.getAllIdentities(when, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}


	public List<Attribute> getAttributes(Element owner, String attribute, 
			boolean effective, boolean includeScoped, boolean includeImplied, 
			Date when) 
		throws SecurityException, InternalException, ElementNotKnownException,
		TimeOutOfRangeException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			if (owner.getGroup() == null && owner.getIdentity() == null)
				throw new InternalException("null element");
			if (attribute != null && attribute.equals(""))
				attribute = null;
			ArrayList<Attribute> ret = new ArrayList<Attribute>();

			//group attribute - simple, no flags are in use
			if (owner.getIdentity() == null)
			{
				authZ.authorize(Permissions.FULL_READ_P, false, session);
				if (attribute == null)
					return groups.getAllAttributes(owner.getGroup(), when, session);
				try
				{
					ret.add(groups.getGroupAttribute(owner.getGroup(), attribute, when, session));
				} catch (AttributeNotFoundException e) {}
				return ret;
			} else
			{
				boolean self = authZ.isSelf(owner.getIdentity(), when, session);
				authZ.authorize(Permissions.FULL_READ_P, self, session);

				//id @ group - also simple as only effective flag is used
				if (owner.getGroup() != null)
				{
					if (attribute == null)
						return entities.getAllAttributes(owner, effective, when, session);
					try
					{
						ret.add(entities.getAttribute(owner, attribute, 
								effective, when, session));
					} catch (AttributeNotFoundException e) {}
					return ret;
				} else
				{
					//single id - here is the mess with flags and authZ filtering
					return getIdentityAttributes(owner, attribute, effective, 
							includeScoped, includeImplied, self, when, session);
				}
			}
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	private List<Attribute> getIdentityAttributes(Element owner, String attribute, 
			boolean effective, boolean includeScoped, boolean includeImplied,
			boolean self, Date when, SqlSession session) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		InvalidValueException, TimeOutOfRangeException
	{
		List<Attribute> ret = entities.getIdentityAttributes(owner, attribute, includeScoped, true, 
			effective, includeImplied, self, when, authZ, session);
		return DB.cleanupAttributeList(ret);

	}
	
	public List<ModificationEvent> getHistoryEvents(Date from, Date to, 
		int cf, int typeFilter, boolean fillDetails) 
		throws SecurityException, InternalException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.FULL_READ_P, false, session);
			List<ModificationEvent> ret = new ArrayList<ModificationEvent>();
			ModificationEvent.CATEGORY.IDENTITY.ordinal();
			if ((cf & (1 << CATEGORY.IDENTITY.ordinal())) != 0)
				entities.getIdentityEvents(ret, from, to, typeFilter, fillDetails, session);
			if ((cf & (1 << CATEGORY.ENTITY.ordinal())) != 0)
				entities.getEntityEvents(ret, from, to, typeFilter, fillDetails, session);
			if ((cf & (1 << CATEGORY.GROUP.ordinal())) != 0)
				groups.getGroupEvents(ret, from, to, typeFilter, fillDetails, session);
			if ((cf & (1 << CATEGORY.MEMBERSHIP.ordinal())) != 0)
				groups.getMembershipEvents(ret, from, to, typeFilter, fillDetails, session);
			if ((cf & (1 << CATEGORY.ATTRIBUTE.ordinal())) != 0)
				entities.getAttributeEvents(ret, from, to, typeFilter, fillDetails, session);

			return ret;
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
}










