/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 18, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.engine;

import java.util.ArrayList;
import java.util.List;

import org.apache.ibatis.session.SqlSession;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.api.exceptions.AttributeNotFoundException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.db.DB;
import pl.edu.icm.unicore.uvos.db.DBEntities;
import pl.edu.icm.unicore.uvos.db.DBGroupResolver;
import pl.edu.icm.unicore.uvos.db.DBGroups;
import pl.edu.icm.unicore.uvos.engine.api.UVOSQueryInterface;
import pl.edu.icm.unicore.uvos.model.GroupBean;

/**
 * Class which handles VO content quering. 
 * @author K. Benedyczak
 */
public class QueryModule implements UVOSQueryInterface
{
	private DB db;
	private DBGroups groups;
	private DBGroupResolver groupResolver;
	private DBEntities entities;
	private InternalAuthorization authZ;
	
	public QueryModule(DB db, DBGroups groups, DBGroupResolver groupResolver, DBEntities entities,
			InternalAuthorization authZ)
	{
		this.groups = groups;
		this.groupResolver = groupResolver;
		this.entities = entities;
		this.authZ = authZ;
		this.db = db;
	}

	//
	// public API implementation ******************************************************
	//

	/**
	 * Checks if given identity is a member of the given group.
	 * <p>
	 * Requires read perm in the group. Can be self access.
	 * @param who
	 * @param group 
	 * @param effective if true method will check also membership which is implied by
	 * membership in one of group subgroup otherwise only membership in the group exactely
	 * is checked. 
	 * @param ctx security context to use
	 * @return
	 * @throws InternalException
	 * @throws SecurityException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 * @throws InvalidValueException 
	 */
	public boolean isMember(Identity who, Group group, boolean effective)
		throws SecurityException, InternalException,
		IdentityNotKnownException, GroupNotKnownException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.READ_P, groupResolver.resolveGroup(group, null, session), 
					authZ.isSelf(who, session), session);
			return groups.isMember(who, effective, group, null, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	/**
	 * Retrieves content of the group. Elements are either other groups or identities.
	 * <p>
	 * Requires global or group read perm. Everybody can get '/' group's contents.
	 * @param group
	 * @param ctx
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 */
	public List<Element> getGroupContent(Group group)
		throws SecurityException, InternalException,
		GroupNotKnownException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(group, null, session);
			if (gb.getId() != -1)
				authZ.authorize(Permissions.READ_P, gb, false, session);
			return groups.getGroupContent(group, null, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public boolean areEquivalent(Identity i1, Identity i2) throws SecurityException,
			InternalException, IdentityNotKnownException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.IDENTITY_CTL_P, 
					authZ.isSelf(i1, session) || authZ.isSelf(i2, session), session);
			return entities.areEquivalent(i1, i2, null, session);		
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	public List<Identity> getAllEquivalents(Identity who) throws SecurityException,
			InternalException, IdentityNotKnownException, InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.IDENTITY_CTL_P, authZ.isSelf(who, session), session);
			return entities.getAllEquivalents(who, null, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public List<Group> getAllGroups(Identity who, boolean implied)
		throws SecurityException, InternalException, IdentityNotKnownException, 
		InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			boolean hasGlobalReadPerm = isGlobalRead(authZ.isSelf(who, session), session);
			return groups.getAllReadableGroups(hasGlobalReadPerm, who, implied, authZ, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	/**
	 * Retrieves all identities stored in the database. Also those not assigned
	 * to any group. 
	 * <p>
	 * Requires global read perm.
	 * @param ctx
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<Identity> getAllIdentities() 
		throws SecurityException, InternalException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			authZ.authorize(Permissions.READ_P, false, session);
			return entities.getAllIdentities(null, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public List<Attribute> getAttributes(Element owner, String attribute, boolean effective, 
			boolean includeScoped, boolean includeImplied) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		InvalidValueException
	{
		SqlSession session = db.getSqlSession();
		try
		{
			if (owner.getGroup() == null && owner.getIdentity() == null)
				throw new InternalException("null element");
			if (attribute != null && attribute.equals(""))
				attribute = null;
			ArrayList<Attribute> ret = new ArrayList<Attribute>();

			//group attribute - simple, no flags are in use
			if (owner.getIdentity() == null)
			{
				authZ.authorize(Permissions.READ_P, 
						groupResolver.resolveGroup(owner.getGroup(), null, session), false, session);
				if (attribute == null)
					return groups.getAllAttributes(owner.getGroup(), null, session);
				try
				{
					ret.add(groups.getGroupAttribute(owner.getGroup(), attribute, null, session));
				} catch (AttributeNotFoundException e) {}
				return ret;
			} else
			{
				boolean self = authZ.isSelf(owner.getIdentity(), session);
				//id @ group - also simple as only effective flag is used
				if (owner.getGroup() != null)
				{
					authZ.authorize(Permissions.READ_P, 
							groupResolver.resolveGroup(owner.getGroup(), null, session), 
							self, session);
					if (attribute == null)
						return entities.getAllAttributes(owner, effective, null, session);
					try
					{
						ret.add(entities.getAttribute(owner, attribute, 
								effective, null, session));
					} catch (AttributeNotFoundException e) {}
					return ret;
				} else
				{
					//single id - here is the mess with flags and authZ filtering
					return getIdentityAttributes(owner, attribute, 
							effective, includeScoped, includeImplied, self, session);
				}
			}
		} finally
		{
			db.releaseSqlSession(session);
		}
	}
	
	private List<Attribute> getIdentityAttributes(Element owner, String attribute, 
			boolean effective, boolean includeScoped, boolean includeImplied,
			boolean self, SqlSession session) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		InvalidValueException
	{
		//Only global - need global read perm.
		if (!includeScoped)
			authZ.authorize(Permissions.READ_P, self, session);
		boolean globalRead = isGlobalRead(self, session);
		
		List<Attribute> ret = entities.getIdentityAttributes(owner, attribute, includeScoped, globalRead, 
			effective, includeImplied, self, null, authZ, session);
		return DB.cleanupAttributeList(ret);
	}
	
	
	private boolean isGlobalRead(boolean self, SqlSession session) 
		throws InternalException, SecurityException
	{
		GroupBean root = new GroupBean();
		root.setId(-1);
		try
		{
			Permissions globalP = 
				authZ.assemblePermissions(root, self, session);
			if (!globalP.isRead())
				return false;
		} catch (GroupNotKnownException e) {}
		return true;
	}
}










