/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 18, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.engine;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;

import eu.emi.security.authn.x509.impl.X500NameUtils;
import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.AttributePermission;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.PermissionDesignator;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.api.exceptions.AuthenticationException;
import pl.edu.icm.unicore.uvos.api.exceptions.AuthorizationException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.UnsupportedTypeException;
import pl.edu.icm.unicore.uvos.db.DB;
import pl.edu.icm.unicore.uvos.db.DBAuthz;
import pl.edu.icm.unicore.uvos.db.DBEntities;
import pl.edu.icm.unicore.uvos.db.DBEntitiesCache;
import pl.edu.icm.unicore.uvos.db.DBGroupResolver;
import pl.edu.icm.unicore.uvos.db.DBGroups;
import pl.edu.icm.unicore.uvos.engine.api.UVOSAuthZInterface;
import pl.edu.icm.unicore.uvos.idhelpers.IdentityHelperFactory;
import pl.edu.icm.unicore.uvos.idhelpers.IdentityTypeHelper;
import pl.edu.icm.unicore.uvos.model.AttributeBean;
import pl.edu.icm.unicore.uvos.model.AuthzBean;
import pl.edu.icm.unicore.uvos.model.EntityBean;
import pl.edu.icm.unicore.uvos.model.GroupBean;
import pl.edu.icm.unicore.uvos.model.IdentityBean;
import pl.edu.icm.unicore.uvos.server.idresolver.Authenticator;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;

/**
 * Main class which handles VO access authorization issues. 
 * It implements authZ related public API methods and also internal authorization
 * stack.
 * @author K. Benedyczak
 */
public class VOAuthorizationModule implements UVOSAuthZInterface, InternalAuthorization
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, VOAuthorizationModule.class);
	private boolean allowX509AsDN;
	private UVOSServerProperties conf;
	private DBGroupResolver groupResolver;
	private DBGroups groups;
	private DBEntities entities;
	private DBEntitiesCache entitiesCache;
	private DBAuthz authz;
	private DB db;

	public VOAuthorizationModule(DBGroupResolver groupResolver, DBGroups groups, DBEntities entities,
			DBEntitiesCache entitiesCache, UVOSServerProperties conf, DBAuthz authz, DB db)
	{
		this.db = db;
		this.groupResolver = groupResolver;
		this.groups = groups;
		this.entities = entities;
		this.entitiesCache = entitiesCache;
		this.authz = authz;
		this.conf = conf;
		allowX509AsDN = conf.getBooleanValue(UVOSServerProperties.MAP_CERT_TO_DN);
		if (allowX509AsDN && !isTLSEnabled())
			log.warn("Mapping of X.509 certificates to DNs will be fully " +
				"operable only when TLS " + 
				" authentication method is enabled. It is strongly " +
				"advised to turn on this authN mechanism.");
	}

	private boolean isTLSEnabled()
	{
		String rawOrder = conf.getValue(UVOSServerProperties.AUTHN_ORDER);
		String [] orderS = rawOrder.split(" ");
		for (int i=0; i<orderS.length; i++)
			if("TLS".equals(orderS[i]))
				return true;
		return false;
	}
	
	//
	// internal API implementation ******************************************************
	//	
	@Override
	public SecurityContext establishSecurityContext(Identity caller) 
	throws IdentityNotKnownException, InternalException, AuthenticationException,
		InvalidValueException
	{
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			IdentityBean idB = entitiesCache.resolveIdentity(caller, null, sqlMap);

			if (idB.getInactive() != 0)
			{
				log.debug("Denying authentication attempt " +
						"for disabled identity " + caller);
				throw new IdentityNotKnownException(caller);
			}

			String authn = entities.getAuthnData(idB.getId(), sqlMap);

			if (caller.getAuthnData() != null || authn != null)
			{
				IdentityTypeHelper idHelper;
				try
				{
					idHelper = IdentityHelperFactory.getHelper(caller);
				} catch (InvalidValueException e)
				{
					throw new IdentityNotKnownException(caller);
				}
				idHelper.validateAuthNToken(authn);
			}

			EntityBean eb = entitiesCache.getEntity4Identity(idB.getId(), sqlMap);
			SecurityContextImpl ctx = new SecurityContextImpl(caller, eb);
			return ctx;
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public void authorize(Permissions permMask, GroupBean group, boolean owner, SqlSession sqlMap) 
		throws SecurityException, InternalException, GroupNotKnownException
	{
		SecurityContext ctx = Authenticator.getCurrentContext();
		//1 - build permissions for caller from attributes etc
		Permissions actualPerms;
		if (ctx instanceof VirtualSecurityContextImpl)
			actualPerms = ((VirtualSecurityContextImpl)ctx).getPerms();
		else
		{
			checkCtxBean(ctx);
			actualPerms = assemblePermissions(group, 
				ctx.getEntity().getId(), owner, sqlMap);
		}
		//2 - check if all from permMask are in actualPerms
		if ((permMask.getValue() & actualPerms.getValue()) != permMask.getValue())
			throw new AuthorizationException("Authorization failed");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void authorizeRecursive(Permissions permMask, GroupBean group, SqlSession sqlMap) 
		throws SecurityException, InternalException, GroupNotKnownException
	{
		authorize(permMask, group, false, sqlMap);
		List<Integer> subgroups = groups.getGroupSubgroups(group.getId(), sqlMap);
		GroupBean g = new GroupBean();
		for (Integer i: subgroups)
		{
			g.setId(i);
			g.setParent(group.getId());
			authorizeRecursive(permMask, g, sqlMap);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void authorize(Permissions permMask, boolean owner, SqlSession sqlMap) 
		throws SecurityException, InternalException
	{
		GroupBean nullGroup = new GroupBean();
		nullGroup.setId(-1);
		try
		{
			authorize(permMask, nullGroup, owner, sqlMap);
		} catch (GroupNotKnownException e)
		{
			throw new InternalException("Authorization failed - BUG " +
					"(group not known)");
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void authorizeIdentityCtl(SqlSession sqlMap)
		throws SecurityException, InternalException	
	{
		authorizeIdentityCtlInternal(sqlMap);
	}

	/**
	 * Authorize identity control operation. It is equivalent to check 
	 * global write or global identityCtl.
	 * @param ctx
	 * @return Global permissions for the caller.
	 * @throws SecurityException
	 * @throws InternalException
	 */
	private Permissions authorizeIdentityCtlInternal(SqlSession sqlMap)
		throws SecurityException, InternalException	
	{
		SecurityContext ctx = Authenticator.getCurrentContext();
		Permissions globalPerms;
		if (ctx instanceof VirtualSecurityContextImpl)
			globalPerms = ((VirtualSecurityContextImpl)ctx).getPerms();
		else
		{
			checkCtxBean(ctx);
			//build global permissions for caller from attributes etc
			GroupBean gb = new GroupBean();
			gb.setId(-1);
			try
			{
				globalPerms = assemblePermissions(gb, 
					ctx.getEntity().getId(), false, sqlMap);
			} catch (GroupNotKnownException e)
			{
				throw new InternalException("Authorization failed " +
						"- BUG (root group not known)");
			}
		}
		if (globalPerms.isWrite())
			return globalPerms;
		if (!globalPerms.isIdentityCtl())
			throw new AuthorizationException("Authorization failed");
		return globalPerms;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void authorizeIdentityCtl(Identity id, SqlSession sqlMap)
		throws SecurityException, InternalException, IdentityNotKnownException, InvalidValueException	
	{
		//1 - build global permissions for caller from attributes etc
		Permissions globalPerms = authorizeIdentityCtlInternal(sqlMap);

		//2 - build global permissions for the id
		Permissions idGlobalPerms;
		EntityBean eb;
		try
		{
			GroupBean gb = new GroupBean();
			gb.setId(-1);
			eb = entitiesCache.getEntity4Identity(id, null, sqlMap);			
			idGlobalPerms = assemblePermissions(gb,	eb.getId(), false, sqlMap);
		} catch (GroupNotKnownException e)
		{
			throw new InternalException("Authorization failed - BUG " +
				"(root group not known)");
		}

		//3 - are global perms of caller better then id's?
		if ((idGlobalPerms.getValue() | globalPerms.getValue()) > 
				globalPerms.getValue())
			throw new AuthorizationException("Authorization failed");
		
		//4 - does caller possess write perm for all id's groups? 
		List<GroupBean> groups = this.groups.getAllGroupsRaw(eb.getId(), null, sqlMap);
		for (GroupBean g: groups)
		{
			try
			{
				authorize(Permissions.WRITE_P, g, false, sqlMap);
			} catch (GroupNotKnownException e)
			{
				throw new InternalException("Authorization failed - BUG " +
					"(group not known)");
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isSelf(Identity id, SqlSession sqlMap) 
		throws IdentityNotKnownException, InternalException, SecurityException,
		InvalidValueException
	{
		return isSelf(id, null, sqlMap);
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean isSelf(Identity id, Date when, SqlSession sqlMap) 
		throws IdentityNotKnownException, InternalException, SecurityException,
		InvalidValueException
	{
		SecurityContext ctx = Authenticator.getCurrentContext();
		if (ctx instanceof VirtualSecurityContextImpl)
			return false;
		EntityBean eb = entitiesCache.getEntity4Identity(id, when, sqlMap);
		checkCtxBean(ctx);
		if (ctx.getEntity().getId() == eb.getId())
			return true;
		return false;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Permissions assemblePermissions(GroupBean group, boolean owner, SqlSession sqlMap) 
		throws GroupNotKnownException, InternalException, SecurityException
	{
		SecurityContext ctx = Authenticator.getCurrentContext();
		if (ctx instanceof VirtualSecurityContextImpl)
			return ((VirtualSecurityContextImpl)ctx).getPerms();
		checkCtxBean(ctx);
		return assemblePermissions(group, ctx.getEntity().getId(), owner, sqlMap);
	}
	
	private Permissions assemblePermissions(GroupBean group, int entityId, boolean owner, SqlSession sqlMap) 
		throws GroupNotKnownException, InternalException
	{
		GroupBean gb = new GroupBean();
		gb.setId(-1);
		Permissions pGlobal = assemblePermissionsI(gb, entityId, owner, sqlMap);
		Permissions pGroup = new Permissions(0);
		if (group.getId() != -1)
			pGroup = assemblePermissionsI(group, entityId, owner, sqlMap);
		Permissions pFinal = new Permissions(pGroup.getValue() | pGlobal.getValue());
		return pFinal;
	}	

	private Permissions assemblePermissionsI(GroupBean group, int entityId, boolean owner, SqlSession sqlMap) 
		throws GroupNotKnownException, InternalException
	{
		List<AuthzBean> policy = authz.getGroupAuthZRaw(group, true, sqlMap);
		List<AttributeBean> attrs;
		boolean member = false;
		try
		{
			attrs =	entities.getEffectiveAttributes(entityId, group, null, sqlMap);
			if (group.getId() != -1)
				member = groups.isEffectiveMember(entityId, group, null, sqlMap);
		} catch (IdentityNotKnownException e)
		{
			throw new InternalException("Identity not known in security stack ??");
		}

		Permissions actualPerms = new Permissions(0);
		HashMap<String, List<String>> possesedAttributes = 
			new HashMap<String, List<String>>();
		for (AttributeBean ab: attrs)
		{
			String key = ab.getTypeId() + "";
			List<String> cur = possesedAttributes.get(key);
			if (cur == null)
			{
				cur = new ArrayList<String>();
				possesedAttributes.put(key, cur);
			}
			if (ab.getVal() != null)
				cur.add(ab.getVal());
		}

		for (AuthzBean authzB: policy)
		{
			String special = authzB.getSpecial();
			if (special != null)
			{
				if ((special.equals("O") && owner) || 
						(special.equals("M") && member))
					actualPerms.setValue(
							actualPerms.getValue() | authzB.getPermission());
				continue;
			}
			String key = authzB.getAttrType() + "";
			List<String> matchingA = possesedAttributes.get(key);
			if (matchingA == null)
				continue;
			if (authzB.getAttrValue() == null)
			{
				actualPerms.setValue(actualPerms.getValue() | 
						authzB.getPermission());
				continue;				
			}
			for (String s: matchingA)
				if (s.equals(authzB.getAttrValue()))
				{
					actualPerms.setValue(actualPerms.getValue() 
							| authzB.getPermission());
					break;				
				}
		}
		return actualPerms;
	}
	
	private void checkCtxBean(SecurityContext ctx) throws SecurityException
	{
		if (ctx.getEntity() == null)
		{
			String identity = ctx.getCaller() == null ? "null" : ctx.getCaller().toString();
			log.fatal("Security context for identity " + identity + " is null. It is a bug.");
			Thread.dumpStack();
			throw new SecurityException("Entity is not known");
		}
	}
	
	@Override
	public boolean isActive(Identity toCheck) 
		throws IdentityNotKnownException, InternalException, InvalidValueException
	{
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			IdentityBean ib = entitiesCache.resolveIdentity(toCheck, null, sqlMap);
			return (ib.getInactive() == 0);
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
	}

	@Override
	public Identity mapIfNeededToX509(Identity toCheck) 
		throws InternalException, InvalidValueException, IdentityNotKnownException
	{
		if (!allowX509AsDN)
			return toCheck;
		if (!toCheck.getType().equals(IdentityType.DN))
			return toCheck;
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			entitiesCache.resolveIdentity(toCheck, null, sqlMap);
			return toCheck;
		} catch(IdentityNotKnownException e)
		{
			String key = X500NameUtils.getComparableForm(toCheck.getValue());
			Identity newId = entities.getX509BySubject(key, sqlMap);
			if (newId != null)
			{
				log.debug("Mapped DN " + toCheck.getValue() + 
					" to an X.509 certificate with the same subject.");
				return newId;
			}
			throw e;
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
	}
	
	//
	// public API implementation ******************************************************
	//
	
	/**
	 * Retrieves set of permissions for the given identity in the group. 
	 * <p>
	 * Requires read perm for the group. Can be self access.
	 * @param group
	 * @param whose
	 * @param ctx
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws IdentityNotKnownException
	 * @throws InvalidValueException 
	 */
	public Permissions checkPermissions(Group group, Identity whose) 
		throws SecurityException, InternalException, GroupNotKnownException,
		IdentityNotKnownException, InvalidValueException
	{
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			boolean self = isSelf(whose, sqlMap);
			GroupBean groupB = groupResolver.resolveGroup(group, null, sqlMap);
			authorize(Permissions.READ_P, groupB, self, sqlMap);
			EntityBean eb = entitiesCache.getEntity4Identity(whose, null, sqlMap);
			return assemblePermissions(groupB, eb.getId(), false, sqlMap);
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
	}

	/**
	 * Retrieves specification of authZ settings of the given group.
	 * <p>
	 * Requires write perm for the group.
	 * @param group
	 * @param effective if true then permissions which are effective are returned 
	 * (so those can be permissions inherited from parent group). Otherwise only exact group
	 * configuration is returned.  
	 * @param ctx
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 */
	public List<AttributePermission> getGroupAuthZ(Group group, boolean effective) 
		throws SecurityException, InternalException, GroupNotKnownException
	{
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			GroupBean groupB = groupResolver.resolveGroup(group, null, sqlMap);
			authorize(Permissions.WRITE_P, groupB, false, sqlMap);
			return authz.getGroupAuthZ(groupB, effective, sqlMap);
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
	}

	/**
	 * Modifies permissions of the group. 
	 * <p>
	 * If the attribute name in the designator 
	 * has already assigned permissions in this group then those are modified. In this case
	 * if the already existing attribute designator has other values then those in the
	 * new designator, the values of designator attribute are updated (but not added).
	 * <p>
	 * If there is no such attribute designator for the group the permissions are added. 
	 * <p>
	 * If the permissions attribute is null then the entry for designator is removed.
	 * <p>
	 * Requires write perm for the group.
	 * @param group Pass null to modify global permissions
	 * @param designator owner of what attributes will be granted permissions
	 * @param permissions set of permissions to be granted, or null to remove permissions.
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws UnsupportedTypeException 
	 */
	public void modifyPermissions(Group group, PermissionDesignator designator,
			Permissions permissions) 
		throws SecurityException, InternalException, GroupNotKnownException,
		UnsupportedTypeException
	{
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			GroupBean gb = groupResolver.resolveGroup(group, null, sqlMap);
			int gid = gb.getId();
			authorize(Permissions.WRITE_P, gb, false, sqlMap);

			if (designator.isSpecial())
			{
				authz.removeAuthZSpecial(designator.getSpecial(), gid, sqlMap);
				if (permissions != null)
					authz.addAuthZSpecial(designator.getSpecial(), gid, permissions, sqlMap);
			} else
			{
				authz.removeAuthZAttr(designator.getAttribute(), gid, sqlMap);
				if (permissions != null)
					authz.addAuthZAttr(designator.getAttribute(), gid, permissions, sqlMap);
			}
			sqlMap.commit();
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
	}

	/**
	 * Changes authentication token of the given identity.
	 * <p>
	 * Requires either global write perm or self access (i.e. is self access, 
	 * and also user can modify it's own passwd).
	 * @param id
	 * @param newToken
	 * @param ctx
	 * @throws SecurityException
	 * @throws InvalidValueException 
	 */
	public void modifyAuthenticationData(Identity id, Object newToken) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		InvalidValueException
	{
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			if (!isSelf(id, sqlMap))
				authorize(Permissions.WRITE_P, true, sqlMap);

			id.setAuthnData(newToken);
			String tokenToDB = null;
			if (newToken != null)
			{
				IdentityTypeHelper idHelper;
				try
				{
					idHelper = IdentityHelperFactory.getHelper(id);
				} catch (InvalidValueException e)
				{
					throw new IdentityNotKnownException(id);
				}
				tokenToDB = idHelper.generateAuthNToken();
			}
			entities.changeAuthnToken(id, tokenToDB, sqlMap);
			sqlMap.commit();
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
	}

	public List<Identity> getMyIdentities() 
		throws SecurityException, InternalException, IdentityNotKnownException
	{
		SqlSession sqlMap = db.getSqlSession();
		try
		{
			SecurityContext ctx = Authenticator.getCurrentContext();
			if (ctx instanceof VirtualSecurityContextImpl)
				return Collections.singletonList(ctx.getCaller());
			checkCtxBean(ctx);
			return entities.getAllEquivalents(ctx.getEntity().getId(), null, sqlMap);
		} finally
		{
			db.releaseSqlSession(sqlMap);
		}
	}

}






