/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 14, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.engine.api;

import java.util.Date;
import java.util.List;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.Notification;
import pl.edu.icm.unicore.uvos.api.exceptions.AttributeNotFoundException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotEmptyException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementAlreadyExistsException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.LabelAlreadyUsedException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.TimeOutOfRangeException;
import pl.edu.icm.unicore.uvos.api.exceptions.UnsupportedTypeException;

/**
 * This interface encapsulates all externally available operations 
 * regarding VO service content management that can be used with the UVOS software.
 * 
 * @author K. Benedyczak
 */
public interface UVOSManagementInterface
{
	/**
	 * Adds a new group. Pass null parent to create top-level group (i.e. VO)
	 * <p>
	 * Requires write perm for parent group (or global for top-level groups).
	 * @param parent
	 * @param name
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws ElementAlreadyExistsException
	 */
	public Group addGroup(Group parent, String name) 
		throws SecurityException, InternalException, GroupNotKnownException, 
		ElementAlreadyExistsException;

	/**
	 * Removes the given group.
	 * <p>
	 * Requires parent group (global for top-level groups), toRemove group and 
	 * all toRename group subgroups write perm.
	 * @param toRemove
	 * @param recursive
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws GroupNotEmptyException
	 */
	public void removeGroup(Group toRemove, boolean recursive) 
		throws SecurityException, InternalException, GroupNotKnownException, 
		GroupNotEmptyException;
	
	/**
	 * Copies or moves the given group to the content of a different group.
	 * <p>
 	 * Requires toCopy's parent group (global for top-level groups), toCopy group, 
	 * all toMove group subgroups and newParent write perm.
	 * @param toCopy
	 * @param newParent
	 * @param newName can be null, then existing name is used
	 * @param deleteOriginal if true, original will be removed after copy 
	 * (so it will be moveGroup rather)
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 * @throws ElementAlreadyExistsException
	 */
	public Group copyGroup(Group toCopy, Group newParent, String newName, 
			boolean deleteOriginal) 
		throws SecurityException, InternalException, 
		GroupNotKnownException, ElementAlreadyExistsException, InvalidValueException;
	
	/**
	 * Adds a new identity.
	 * <p>
	 * Requires global identityCtl perm or global write perm
	 * @param toAdd timestamps are ignored.
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementAlreadyExistsException
	 * @throws InvalidValueException
	 * @throws LabelAlreadyUsedException
	 * @throws UnsupportedTypeException 
	 */
	public void addIdentity(Identity toAdd) 
		throws SecurityException, InternalException, ElementAlreadyExistsException,
		UnsupportedTypeException, LabelAlreadyUsedException, InvalidValueException;
	
	/**
	 * Adds a new identity, whch represents the same entity as one already existing.
	 * <p>
	 * Requires either global write perm or (global identityCtl perm + 
	 * write perm for every group equivalentIdentity is a member of + 
	 * the same or better global permissions as equivalentIdentity).
	 * @param toAdd
	 * @param equivalentIdentity
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementAlreadyExistsException
	 * @throws IdentityNotKnownException
	 * @throws InvalidValueException 
	 * @throws {@link UnsupportedTypeException}
	 */
	public void addIdentity(Identity toAdd, Identity equivalentIdentity) 
		throws SecurityException, InternalException, ElementAlreadyExistsException,
		IdentityNotKnownException, UnsupportedTypeException, InvalidValueException;
	
	/**
	 * Deletes identity.
	 * <p>
	 * Requires global write perm or (global identityCtl and write perm for every group
	 * toRemove is a member of + 
	 * the same or better global permissions as equivalentIdentity).
	 * @param toRemove
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws {@link InvalidValueException}
	 */
	public void removeIdentity(Identity toRemove) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		InvalidValueException;
	
	/**
	 * Adds a new attribute. 
	 * <p>
	 * Requires write perm: global for global entity attributes, or specified group
	 * for group and entity@group attributes. 
	 * @param whom Who is the owner of the attribute.
	 * @param toAdd Attribute to add
	 * @param update If true and attribute with the same name exists it will be updated. If false
	 * and attribute with the same name exists there will be exception 
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws ElementAlreadyExistsException if update is false and attribute with the same 
	 * name exists
	 */
	public void setAttribute(Element whom, Attribute toAdd, boolean update) 
		throws SecurityException, InternalException, ElementNotKnownException,
		ElementAlreadyExistsException, UnsupportedTypeException, InvalidValueException;
	
	/**
	 * Removes the attribute. 
	 * <p>
	 * Requires write perm: global for global entity attributes, or specified group
	 * for group and entity@group attributes. 
	 * @param whom
	 * @param toRemove
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 */
	public void removeAttribute(Element whom, String toRemove) 
		throws SecurityException, InternalException, ElementNotKnownException,
		UnsupportedTypeException, InvalidValueException;

	/**
	 * Assigns the given identity to a group.
	 * <p>
	 * Requires write perm of the specified group. 
	 * @param toAdd
	 * @param group
	 * @param ctx
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 */
	public void addToGroup(Identity toAdd, Group group) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		GroupNotKnownException, InvalidValueException;
	
	/**
	 * Removes the given identity from the given group.
	 * <p>
	 * Requires write perm of the specified group. 
	 * @param toRemove
	 * @param group
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 */
	public void removeFromGroup(Identity toRemove, Group group) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		GroupNotKnownException, InvalidValueException;
	
	/**
	 * Changes the label of the identity. Label is common for all equivalent identities.
	 * <p>
	 * Requires global write perm or (global identityCtl perm + write perm for all 
	 * groups toChange is a memeber of + 
	 * the same or better global permissions as equivalentIdentity).
	 * @param toChange
	 * @param label
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws {@link InvalidValueException}
	 * @throws LabelAlreadyUsedException
	 */
	public void setIdentityLabel(Identity toChange, String label) 
		throws SecurityException, InternalException, IdentityNotKnownException, 
		LabelAlreadyUsedException, InvalidValueException;

	/**
	 * Disables or enables the identity. See {@link Identity}.
	 * <p>
	 * Requires global write perm or (global identityCtl perm + write perm for all 
	 * groups toChange is a memeber of + 
	 * the same or better global permissions as equivalentIdentity).
	 * @param toChange
	 * @param label
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws {@link InvalidValueException}
	 */
	public void setIdentityStatus(Identity toChange, boolean enable) 
		throws SecurityException, InternalException, IdentityNotKnownException, 
		InvalidValueException;

	
	/**
	 * Returns list of all types of attributes.
	 * <p>
	 * Requires no perm.
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<AttributeType> getAttributeTypes() 
		throws SecurityException, InternalException;

	/**
	 * Update list of attribute types. If clear is false then types from toUpdate list are
	 * added or modified (if attribute with the same name exists). Otherwise attribute types
	 * with given names in the list are removed but only if are unused (i.e. there is no 
	 * attribute of this type either in current or in historical data). 
	 * <p>
	 * Requires no perm to add a new attribute type. Requires global write perm for all other
	 * operations.
	 * @param toUpdate
	 * @param clear
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public void updateAttributeTypes(List<AttributeType> toUpdate, boolean clear) 
		throws SecurityException, InternalException, InvalidValueException;
	
	/**
	 * Returns list of all types of identities.
	 * <p>
	 * Requires no perm.
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<IdentityType> getIdentityTypes() 
		throws SecurityException, InternalException;
	
	/**
	 * Temporarly disables the given attribute value. Such attribute value is then not
	 * visible in any of query methods of this interface except 
	 * @link{#getDisabledAttributes}. 
	 * <p>
	 * Requires fullRead perm: in the group if group or id@group attributes are 
	 * modified and global otherwise. Can be self access.
	 * @param whose
	 * @param toDisable
	 * @param valueToDisable if null then all values ot he attribute are disabled.
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws AttributeNotFoundException
	 * @throws InvalidValueException 
	 */
	public void disableAttribute(Element whose, String toDisable, String valueToDisable) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		AttributeNotFoundException, InvalidValueException;
	
	/**
	 * Enables disabled earlier value of the given attribute. 
	 * <p>
	 * Requires fullRead perm: in the group if group or id@group attributes are 
	 * modified and global otherwise. Can be self access.
	 * @param whose
	 * @param toEnable
	 * @param valueToEnable if null then all disabled values are enabled.
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws AttributeNotFoundException
	 */
	public void enableAttribute(Element whose, String toEnable, String valueToEnable) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		AttributeNotFoundException, InvalidValueException;
	
	/**
	 * Returns a list with attributes that have disabled values. Only those disabled
	 * values are in attributes returned. It is not recursive, i.e. group A or id@group A
	 * attributes which are disabled in group A subgroups are not returned.   
	 * <p>
	 * Requires fullRead perm: in the group if group or id@group attributes are 
	 * modified and global otherwise. Can be self access.
	 * @param whose
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @throws InvalidValueException 
	 */
	public List<Attribute> getDisabledAttributes(Element whose) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		InvalidValueException;
	
	/**
	 * Deletes permanently all service data that is older then deleteFrom (exclusive).
	 * <p>
	 * Requires global write perm.
	 * @param deleteFrom
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws TimeOutOfRangeException
	 */
	public void purgeHistoricalData(Date deleteFrom) 
		throws SecurityException, InternalException, TimeOutOfRangeException;
	
	/**
	 * Adds a new notification.
	 * <p>
	 * Requires either globar write perm or group write perm for the group groupFilter
	 * (if it is defined).
	 * 
	 * @param action
	 * @param receipents
	 * @param groupFilter can be null when no filtering is requested
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public void addNotification(String action, String receipents, String groupFilter)
		throws SecurityException, InternalException;
	
	/**
	 * Removes a notification
	 * <p>
	 * Requires either globar write perm or group write perm for the group groupFilter
	 * (if it is defined).
	 * @param id
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public void removeNotification(int id)
		throws SecurityException, InternalException;

	/**
	 * Retrieves a list of currently defined notifications which possess the specified 
	 * filter (or any filter which is prefixed by the argument).
	 * <p>
	 * Requires either globar write perm or group write perm for the group groupFilter
	 * (if it is defined).
	 * @param action can be null - then all notifications are returned.
	 * @param groupFilter can be null - then notifications regardless of their 
	 * filter are returned.
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<Notification> getNotifications(String action, String groupFilter)
		throws SecurityException, InternalException;
}
