/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 14, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.engine.api;

import java.util.Date;
import java.util.List;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.ModificationEvent;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.TimeOutOfRangeException;

/**
 * This interface encapsulates all externally available operations 
 * regarding quering a db history that can be used with the UVOS software.
 * 
 * @author K. Benedyczak
 */
public interface UVOSQueryHistoryInterface
{
	/**
	 * Checks if given identity was a memeber of the given group at given time in the past.
	 * <p>
	 * Requires global fullRead perm. Can be self access.
	 * @param who
	 * @param group
	 * @param effective if true method will check also membership which is implied by
	 * membership in one of gorup subgroup otherwise only membership in the group exactely
	 * is checked. 
	 * @param when
	 * @return
	 * @throws InternalException
	 * @throws SecurityException
	 * @throws IdentityNotKnownException
	 * @throws GroupNotKnownException
	 * @throws InvalidValueException
	 * @thorws TimeOutOfRangeException
	 */
	public boolean isMember(Identity who, Group group, boolean effective, 
			Date when) 
		throws SecurityException, InternalException, IdentityNotKnownException, 
		GroupNotKnownException, TimeOutOfRangeException, InvalidValueException;
	
	/**
	 * Retrieves attributes of the given element (i.e. identity, group or 
	 * identity in a group scope). If URN is null or "" then all attributes are returned.
	 * <p>
	 * See description of parameters to get possible versions of invocation.
	 * <p>
	 * Requires global fullRead perm, can be self access.
	 * @param owner
	 * @param attribute the name of interesting attributes, can be null. 
	 * @param effective if true method will get effective attributes e.g. which are implied by
	 * membership in one of group subgroups, otherwise only attributes exactely 
	 * matching are returned.
	 * @param includeScoped important only when owner is an identity 
	 * (not a group or id@group). Then if it is false only global attributes are returned. 
	 * Otherwise attributes global and all group-scoped attributes (for groups where owner 
	 * is a member) are returned.
	 * @param includeImplied Used only when includeScoped is true. Then it governs the set 
	 * of groups which are used to check group-scoped attributes for the owner. If false
	 * only groups where owner is a real memeber are checked. If true then also groups
	 * where membership is implied (by being in group's subgroup) are evaluated. 
	 * @param when
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws ElementNotKnownException
	 * @thorws InvalidValueException
	 * @throws {@link TimeOutOfRangeException}
	 */
	public List<Attribute> getAttributes(Element owner, String attribute, boolean effective,
			boolean includeScoped, boolean includeImplied, Date when) 
		throws SecurityException, InternalException, ElementNotKnownException, 
		TimeOutOfRangeException, InvalidValueException;
	
	/**
	 * Get all groups the given identity was a memeber of in the past.
	 * <p>
	 * Requires global fullRead perm, can be self access.
	 * @param who
	 * @param implied if false then method doesn't return parent 
	 * groups (where membership is implied by membership in subgroup). Otherwise those 
	 * implied groups are returned too. 
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws InvalidValueException
	 * @throws TimeOutOfRangeException
	 */
	public List<Group> getAllGroups(Identity who, boolean implied, Date when) 
		throws SecurityException, InternalException, IdentityNotKnownException, 
		TimeOutOfRangeException, InvalidValueException;
	
	/**
	 * Checks if two identities represented the same entity in the past.
	 * <p>
	 * Requires global fullRead perm and global identityCtl perm. Can be self access.
	 * @param i1
	 * @param i2
	 * @param when
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws InvalidValueException
	 * @throws TimeOutOfRangeException
	 */
	public boolean areEquivalent(Identity i1, Identity i2, Date when) 
		throws SecurityException, InternalException, IdentityNotKnownException, 
		TimeOutOfRangeException, InvalidValueException;

	/**
	 * Retrieves all equivalent identities of the given one in the past.
	 * <p>
	 * Requires global fullRead perm and global identityCtl perm. Can be self access.
	 * @param who
	 * @param when
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws IdentityNotKnownException
	 * @throws InvalidValueException
	 * @throws TimeOutOfRangeException
	 */
	public List<Identity> getAllEquivalents(Identity who, Date when) 
		throws SecurityException, InternalException, IdentityNotKnownException,
		TimeOutOfRangeException, InvalidValueException;
	
	/**
	 * Retrieves past content of the group. 
	 * Elements are either other groups or identities.
	 * <p>
	 * Requires global fullRead perm.
	 * @param group
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws GroupNotKnownException
	 */
	public List<Element> getGroupContent(Group group, Date when)
		throws SecurityException, InternalException, GroupNotKnownException,
		TimeOutOfRangeException;
	
	/**
	 * Retrievies all identities that were stored in the database at some point.
	 * <p>
	 * Requires global fullRead perm.
	 * @param when 
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<Identity> getAllIdentities(Date when)
		throws SecurityException, InternalException;
	
	/**
	 * Retrieves list of changes that occured in the specified period of time.
	 * <p>
	 * Requires global fullRead perm. 
	 * @param from
	 * @param to
	 * @param categoryFilter
	 * @param typeFilter
	 * @param fillDetails
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<ModificationEvent> getHistoryEvents(Date from, Date to, 
			int categoryFilter, int typeFilter, 
			boolean fillDetails)
		throws SecurityException, InternalException;
	
}
