/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Mar 14, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.engine.api;

import java.util.List;

import pl.edu.icm.unicore.uvos.api.ApplicationActions;
import pl.edu.icm.unicore.uvos.api.VOApplication;
import pl.edu.icm.unicore.uvos.api.VOApplicationForm;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.VOApplicationException;

/**
 * This interface encapsulates all engine operations which 
 * regard defining and submitting applications to VOs.
 * 
 * @author K. Benedyczak
 */
public interface VOApplicationInterface
{	
	/**
	 * Adds or updates an application definition. 
	 * <p>
	 * Requires a write perm for the group which is set in applicationDef.
	 * 
	 * @param applicationDef
	 * @param update
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws VOApplicationException
	 */
	public void updateApplicationForm(VOApplicationForm applicationDef,
			boolean update)
		throws SecurityException, InternalException, VOApplicationException;

	/**
	 * Removes an application definition. 
	 * <p>
	 * Requires a write perm for the group which is set in applicationDef.
	 * @param id
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws VOApplicationException
	 */
	public void removeApplicationForm(int id)
		throws SecurityException, InternalException, VOApplicationException;

	/**
	 * Lists all application definitions.
	 * <p>
	 * Requires global no perm.
	 * 
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 */
	public List<VOApplicationForm> getApplicationForms()
		throws SecurityException, InternalException;
	
	/**
	 * Lists all applications of the selected form and/or with selected status.
	 * Both filtering arguments may be null, what eleminates the constraint.
	 * <p>
	 * Requires fullRead perm for the group, which is the application form's base. 
	 * In case of getting applications of any form a global fullRead is required.  
	 * @param formId
	 * @return
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws VOApplicationException 
	 */
	public List<VOApplication> getApplications(Integer formId, String status)
		throws SecurityException, InternalException, VOApplicationException;
	
	/**
	 * Adds a new application.
	 * <p>
	 * Requires no perm.
	 * @param application
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws VOApplicationException
	 */
	public void submitApplication(VOApplication application)
		throws SecurityException, InternalException, VOApplicationException;
	
	/**
	 * Process an application. This operation only marks the application 
	 * accordingly but doesn't add a new identity (it must be performed by client
	 * software manually).
	 * <p>
	 * Requires write perm for the group that is base of application form 
	 * which the selected application is for.
	 * @param id
	 * @param action
	 * @param notes
	 * @param application
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws VOApplicationException
	 */
	public void processApplication(int id, ApplicationActions action, 
			String notes, boolean sendConfirmation, VOApplication application)
		throws SecurityException, InternalException, VOApplicationException;
	
	/**
	 * Used to signal the server that application with CSR only should be updated,
	 * as the contained CSR was processed by a CA.
	 * 
	 * @param csr original CSR
	 * @param accepted wheter the CSR was accepted or not
	 * @param certificate if accepted then a signed certificate in PEM format
	 * ought be put here.
	 * @param sendNotification wheter the server should send a notification 
	 * to the application owner if the CSR was rejected and therefore 
	 * the application is invalidated too. Note that in case of accepted CSR 
	 * the notification won't be send - the application is updated
	 * but still needs to be processed by VO staff.
	 * @throws SecurityException
	 * @throws InternalException
	 * @throws VOApplicationException
	 */
	public void csrProcessedNotification(String csr, boolean accepted, 
			String certificate, boolean sendNotification)
		throws SecurityException, InternalException, VOApplicationException;
}
