/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Feb 1, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.mail;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;

import javax.mail.Authenticator;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.PasswordAuthentication;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.MimeMessage;
import javax.net.ssl.SSLSocketFactory;

import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlException;

import eu.emi.security.authn.x509.impl.SocketFactoryCreator;
import eu.unicore.security.canl.IAuthnAndTrustConfiguration;
import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.VOApplication;
import pl.edu.icm.unicore.uvos.api.VOApplicationForm;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.VOApplicationException;
import pl.edu.icm.unicore.uvos.db.DB;
import pl.edu.icm.unicore.uvos.db.DBNotification;
import pl.edu.icm.unicore.uvos.model.NotifyBean;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;
import xmlbeans.pl.edu.icm.uvos.internal.app.VOApplicationFormConstraintsType;

/**
 * Allows for simple notification support via email.
 * @author K. Benedyczak
 */
public class MailSender
{
	private Properties props;
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, MailSender.class);
	private static final String CFG_ENABLE = "mailx.enable"; 
	private static final String CFG_USER = "mailx.smtp.auth.username"; 
	private static final String CFG_PASSWD = "mailx.smtp.auth.password";
	private static final String CFG_TESTADDR = "mailx.sendTestMessageTo";
	
	private static final String MAIL_TEMPLATE_PRE = "mailtemplate.";
	private static final String ACT_NA = "newApplication"; 
	private static final String ACT_PR = "applicationProcessed"; 
	
	private static final String NEW_APPLICATION_SUBJECT = "A new application was received";
	private static final String APPLICATION_PROCESSED_SUBJECT = "VO application was processed";
	private static final String NEW_APPLICATION_BODY = "A new application" +
			" for the form ${FORM_NAME} (id ${FORM_ID}) was received.\n" +
			"It is for the VO/group: ${FORM_GROUP}.";
	private static final String APPLICATION_PROCESSED_BODY = "Your VO application has been just " +
			"processed by the VO service staff.\n\nThe status is: " +
			"${APP_STATUS}\nAdditional information: ${APP_ADMIN_NOTES}";
	private static final String DEFAULT_TEMPLATE_SUBJECT = "A VO event occured: ${ACTION}";
	private static final String DEFAULT_TEMPLATE_BODY = 
		"Action ${ACTION} was performed by ${CALLER}.\n" +
		"Action parameters are:\n" +
		"Subject: ${SUBJECT}\n" +
		"Group: ${GROUP}\n";
	private DBNotification dbN;
	private DB db;
	boolean enable = false;
	
	public MailSender(DB db, DBNotification dbN, UVOSServerProperties conf) throws FileNotFoundException, IOException
	{
		this.db = db;
		this.dbN = dbN;
		if (conf == null)
			return;
		String configFile = conf.getValue(UVOSServerProperties.MAIL_CONF);
		props = new Properties();
		BufferedInputStream configBis=new BufferedInputStream(new FileInputStream(configFile));
		try{
			props.load(configBis);
		}
		finally{
			configBis.close();
		}
		String templatesFile = conf.getValue(UVOSServerProperties.MAIL_TEMPLATES);
		BufferedInputStream templatesBis = new BufferedInputStream(new FileInputStream(templatesFile));
		try{
			props.load(templatesBis);
		}
		finally {
			templatesBis.close();
		}

		String enableS = props.getProperty(CFG_ENABLE);
		if (enableS != null && enableS.equals("true"))
			enable = true;
		else
			return;
		
		IAuthnAndTrustConfiguration sec = conf.getAuthAndTrust();
		SSLSocketFactory sf = SocketFactoryCreator.getSocketFactory(sec.getCredential(), 
				sec.getValidator());
		props.put("mail.smtp.ssl.socketFactory", sf);
		
		String tmp = props.getProperty(CFG_TESTADDR);
		if (tmp != null)
			try
			{
				sendMessage("VO service test message", 
					"This message was sent upon VO service startup " +
					"to verify mail settings.\n\nYou can disable " +
					"this behaviour by commenting out the " + CFG_TESTADDR 
					+ " property in the mail settings.", tmp);
			} catch (Exception e)
			{
				log.error("Sending a test message failed. Please verify " +
					"mail settings. Please note that the server" +
					"will anyway start up, but mail notification " +
					"will be DISABLED!", e);
				enable = false;
			}
			
		/*
		props.setProperty("mail.host", "leon.mat.uni.torun.pl");
		props.setProperty("mail.transport.protocol", "smtp");
		props.setProperty("mail.debug", "true");
		props.setProperty("mail.smtp.from", "golbi@mat.uni.torun.pl");
		props.setProperty("mail.user", "golbi");
		props.setProperty("mail.smtp.auth", "true");
		props.setProperty("mail.smtp.starttls.enable", "true");
		*/
	}
	
	private class SimpleAuthenticator extends Authenticator
	{
		protected PasswordAuthentication getPasswordAuthentication() 
		{
			return new PasswordAuthentication(props.getProperty(CFG_USER), 
					props.getProperty(CFG_PASSWD));
		}
	}
	
	public void sendNotification(VariablesMap vars) throws InternalException
	{
		if (!enable)
			return;
		
		String addr = getAddressFor(vars.getAction(), vars.getGroup());
		if (addr == null)
			return;
		String subject = getMsgText(vars.getMap(), vars.getAction(), true);
		String body = getMsgText(vars.getMap(), vars.getAction(), false);
		sendMessageBg(subject, body, addr);
	}

	public void sendNewApplicationNotification(VOApplication application, 
			VOApplicationForm form) throws VOApplicationException
	{
		if (!enable)
			return;
		String subject = substituteNewApp(NEW_APPLICATION_SUBJECT, application, form);
		String body = substituteNewApp(NEW_APPLICATION_BODY, application, form);
		VOApplicationFormConstraintsType info;
		try
		{
			info = VOApplicationFormConstraintsType.
				Factory.parse(form.getConstraints());
			String addr = info.getNotifyAddress();
			if (addr != null && !addr.equals(""))
				sendMessageBg(subject, body, addr);
		} catch (XmlException e)
		{
			log.error("BUG! Application form can't be parsed: " + e);
			throw new VOApplicationException("BUG! Application" +
					" form can't be parsed: " + e);
		} 
	}

	public void sendProcessedApplicationNotification(VOApplication application, 
			VOApplicationForm form)
	{
		if (!enable)
			return;
		String subject = substituteProcessedApp(APPLICATION_PROCESSED_SUBJECT, 
				application, form);
		String body = substituteProcessedApp(APPLICATION_PROCESSED_BODY, 
				application, form);
		String addr = application.getEmailAddress();
		if (addr != null && !addr.equals(""))
			sendMessageBg(subject, body, addr);
	}

	
	private void sendMessageBg(final String subject, final String body, 
			final String to) 
	{
		new Thread()
		{
			public void run()
			{
				try
				{
					sendMessage(subject, body, to);
				} catch (Exception e)
				{
					log.warn("Can't send " +
						"email with notification: " + e);
				}
			}
		}.start();
	}	
	
	private void sendMessage(String subject, String body, String to) 
		throws AddressException, MessagingException
	{
		Session session = Session.getInstance(props, new SimpleAuthenticator());
		MimeMessage msg = new MimeMessage(session);
		msg.setFrom();
		msg.setRecipients(Message.RecipientType.TO, to);
		msg.setSubject(subject);
		msg.setSentDate(new Date());
		msg.setText(body);
		Transport.send(msg);
	}
	
	private String substituteNewApp(String text, VOApplication application, 
			VOApplicationForm form)
	{
		text = text.replace(VariablesMap.VAR_FORM_ID, form.getId() + "");
		text = text.replace(VariablesMap.VAR_FORM_GROUP, form.getBaseGroup().toString());
		text = text.replace(VariablesMap.VAR_FORM_NAME, form.getFriendlyName());
		return text;
	}

	private String substituteProcessedApp(String text, VOApplication application, 
			VOApplicationForm form)
	{
		text = text.replace(VariablesMap.VAR_APP_ID, application.getId()+"");
		text = text.replace(VariablesMap.VAR_FORM_GROUP, form.getBaseGroup().toString());
		text = text.replace(VariablesMap.VAR_FORM_NAME, form.getFriendlyName());
		text = text.replace(VariablesMap.VAR_APP_STATUS, application.getStatus().toString());
		String val = application.getAdminNotes();
		if (val == null)
			val = "-";
		text = text.replace(VariablesMap.VAR_APP_ADMIN_NOTES, val);
		return text;
	}
	
	private String getAddressFor(String action, String group) throws InternalException
	{
		SqlSession session = db.getSqlSession();
		List<?> notifications;
		try
		{
			notifications = dbN.getNotifications(action, session);
		} finally
		{
			db.releaseSqlSession(session);
		}
		if (group != null)
		{
			for (int i=notifications.size()-1; i>=0; i--)
			{
				NotifyBean nb = (NotifyBean)notifications.get(i); 
				if (nb.getGroupFilter() != null && 
					!group.equals(nb.getGroupFilter()))
					notifications.remove(i);
			}
		}
		if (notifications.size() == 0)
			return null;
		StringBuffer ret = new StringBuffer();
		for (Object o: notifications)
		{
			ret.append(((NotifyBean)o).getRecipients());
			ret.append(",");
		}
		ret.setLength(ret.length() - 1);
		return ret.toString();
	}
	
	private String getTemplate(String operationKey,	boolean subject)
	{
		String suffix = (subject ? ".subject" : ".body");
		String key = MAIL_TEMPLATE_PRE + operationKey + suffix;
		String template = props.getProperty(key);
		if (template == null)
		{
			if (operationKey.equals(ACT_NA))
				template = subject ? NEW_APPLICATION_SUBJECT :
					NEW_APPLICATION_BODY;
			else if (operationKey.equals(ACT_PR))
				template = subject ? APPLICATION_PROCESSED_SUBJECT :
					APPLICATION_PROCESSED_BODY;
			else
				template = props.getProperty(MAIL_TEMPLATE_PRE + suffix);
		}
		if (template == null)
			template = subject ? DEFAULT_TEMPLATE_SUBJECT :
				DEFAULT_TEMPLATE_BODY;
		return template;
	}
	
	private String getMsgText(Map<String, String> varsMap, String operationKey, 
		boolean subject)
	{
		String template = getTemplate(operationKey, subject);
		Iterator<Entry<String, String>> vars = varsMap.entrySet().iterator();
		while (vars.hasNext())
		{
			Entry<String, String> var = vars.next();
			template = template.replace(var.getKey(), var.getValue());
		}
		return template;
	}
}










