/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Aug 21, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.saml;

import java.security.cert.X509Certificate;
import java.util.Calendar;

import org.apache.log4j.Logger;

import eu.unicore.samly2.SAMLConstants;
import eu.unicore.samly2.elements.NameID;
import eu.unicore.samly2.exceptions.SAMLProtocolException;
import eu.unicore.samly2.exceptions.SAMLRequestException;
import eu.unicore.samly2.proto.AbstractSAMLMessage;
import eu.unicore.security.dsig.DSigException;
import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.exceptions.AuthenticationException;
import pl.edu.icm.unicore.uvos.api.exceptions.IdentityNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.UVOSException;
import pl.edu.icm.unicore.uvos.server.IssuerIdentityProvider;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.UVOSFault;
import xmlbeans.org.oasis.saml2.assertion.NameIDType;
import xmlbeans.org.oasis.saml2.assertion.SubjectType;
import xmlbeans.org.oasis.saml2.protocol.RequestAbstractType;

/**
 * SAML utility base: mostly conversions between SAML and UVOS data types.
 * @author K. Benedyczak
 */
public abstract class SAMLUtil
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, SAMLUtil.class);
	
	public static final String ID_PREFIX = "uvosSaml";
	protected SAMLConfiguration samlConfiguration;
	protected RequestAbstractType request;
	protected IssuerIdentityProvider secProvier;
	protected Calendar issueInstant;

	public SAMLUtil(RequestAbstractType req, IssuerIdentityProvider secProvider, UVOSServerProperties conf, SAMLConfiguration samlConfig)
	{
		samlConfiguration=samlConfig;
		request = req;
		this.secProvier = secProvider;
		issueInstant = req.getIssueInstant();
	}
	
	public SAMLUtil(IssuerIdentityProvider secProvider, UVOSServerProperties conf)
	{
		this.secProvier = secProvider;
	}

	public NameIDType getRequestIssuerRaw()
	{
		return request.getIssuer();
	}
	
	public Identity getRequestIssuer() throws SAMLRequestException
	{
		NameIDType xmlIssuer = request.getIssuer();
		return mapNameID(xmlIssuer);
	}
	
	public NameID getIssuer()
	{
		return getIssuer(secProvier);
	}

	public static NameID getIssuer(IssuerIdentityProvider sec)
	{
		String uri = sec.getIssuerURI();
		return new NameID(uri, SAMLConstants.NFORMAT_ENTITY);
	}

	public void checkRequestValidity(int validityPeriod) 
		throws SAMLProtocolException
	{
		if (issueInstant == null)
			throw new SAMLProtocolException(SAMLConstants.STATUS_REQUESTER,
					SAMLConstants.STATUS2_REQUEST_DENIED, 
					"No issueInstant attribute");
		Calendar cur = Calendar.getInstance();
		cur.setLenient(true);
		cur.add(Calendar.SECOND, -validityPeriod);
		if (issueInstant.before(cur))
			throw new SAMLProtocolException(
					SAMLConstants.STATUS_REQUESTER,
					SAMLConstants.STATUS2_REQUEST_DENIED,
					"Request is too old");
	}
	
	public abstract X509Certificate[] getIssuerIdentityFromSignature() 
		throws DSigException;
	public abstract void parse() throws SAMLProtocolException;

	protected X509Certificate[] getIssuerIdentityFromSignature(
			AbstractSAMLMessage query) 
		throws DSigException
	{
		if (!query.isSigned())
			return null;
		X509Certificate[] issuerCC = query.getIssuerFromSignature();
		if (issuerCC == null)
			return null;
		if (!query.isCorrectlySigned(issuerCC[0].getPublicKey()))
			throw new DSigException("Invalid signature");
		return issuerCC;
	}

	protected SAMLProtocolException convert2SAMLError(UVOSException e, String msg, 
			boolean useExDetail) throws UVOSFault
	{
		String code, subcode = null, message = null;
		if (msg != null)
			message = msg;
		if (msg == null && useExDetail)
		{
			if (e.getMessage() != null)
				message = e.getMessage();
			else
				message = e.getClass().getSimpleName();
		}
			
		if (e instanceof IdentityNotKnownException)
		{
			code = SAMLConstants.STATUS_REQUESTER;
			subcode = SAMLConstants.STATUS2_UNKNOWN_PRINCIPIAL;
		} else if (e instanceof AuthenticationException)
		{
			code = SAMLConstants.STATUS_REQUESTER;
			subcode = SAMLConstants.STATUS2_AUTHN_FAILED;
		} else if (e instanceof InvalidValueException)
		{
			code = SAMLConstants.STATUS_REQUESTER;
			subcode = SAMLConstants.STATUS2_INVALID_ATTR;
		} else if (e instanceof InternalException)
		{
			throw new UVOSFault((InternalException)e);
		}  else if (e instanceof SecurityException)
		{
			code = SAMLConstants.STATUS_REQUESTER;
			subcode = SAMLConstants.STATUS2_REQUEST_DENIED;
		} else
		{
			log.warn("Producing SAML error for unknown exception: " + 
					e.getClass());
			code = SAMLConstants.STATUS_RESPONDER;
		}
		
		SAMLProtocolException ret = new SAMLProtocolException(code, subcode, 
				message);
		return ret;
	}
	
	protected Identity mapSubject(SubjectType subject) throws SAMLRequestException
	{
		NameIDType nameId = subject.getNameID();
		if (nameId == null)
		{
			if (subject.getEncryptedID() != null || subject.getBaseID() != null)
				throw new SAMLRequestException(SAMLConstants.STATUS2_REQUEST_UNSUPP, 
					"Only NameId subjects are supported");
			else
				throw new SAMLRequestException(null, 
					"Empty subject is invalid");
		}
		return mapNameID(nameId);
	}

	protected Identity mapNameID(NameIDType nameId) throws SAMLRequestException
	{
		String format = nameId.getFormat();
		if (format == null)
			format = SAMLConstants.NFORMAT_UNSPEC;
		String nameValue = nameId.getStringValue();
		if (nameValue == null)
			throw new SAMLRequestException(null, 
				"Empty nameId is invalid");
		IdentityType idType = null;
		if (format.equals(IdentityType.DN.toString()))
			idType = IdentityType.DN;
		else if (format.equals(IdentityType.EMAIL.toString()))
			idType = IdentityType.EMAIL;
		else if (format.equals(IdentityType.X509CERT.toString()))
			idType = IdentityType.X509CERT;
		else
		{
			throw new SAMLRequestException(SAMLConstants.STATUS2_UNKNOWN_PRINCIPIAL,
					"Unknown nameId format: " + format);
		}
		Identity ret = new Identity(idType, nameValue, null);
		return ret;
	}

	public Calendar getIssueInstant()
	{
		return issueInstant;
	}
}
