/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Aug 7, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.server;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.log4j.Logger;
import org.codehaus.xfire.XFire;
import org.codehaus.xfire.XFireFactory;
import org.codehaus.xfire.aegis.AegisBindingProvider;
import org.codehaus.xfire.annotations.AnnotationServiceFactory;
import org.codehaus.xfire.service.Service;
import org.codehaus.xfire.service.invoker.BeanInvoker;
import org.codehaus.xfire.xmlbeans.XmlBeansType;
import org.codehaus.xfire.xmlbeans.XmlBeansTypeRegistry;
import org.codehaus.xfire.xmlbeans.XmlBeansWSDLBuilderFactory;
import org.picocontainer.PicoContainer;

import eu.unicore.security.xfireutil.AuthInHandler;
import eu.unicore.security.xfireutil.ETDInHandler;
import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.db.DB;
import pl.edu.icm.unicore.uvos.registration.EMIRHandler;
import pl.edu.icm.unicore.uvos.registration.RegistrationHandler;
import pl.edu.icm.unicore.uvos.util.AttributeTypesLoader;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;
import pl.edu.icm.unicore.uvos.util.LogFileWatcher;
import pl.edu.icm.unicore.uvos.util.ThreadScheduler;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.CISInfoProvider;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.SAMLAuthnInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.SAMLNameIdMappingInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.SAMLQueryInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.WSApplicationsInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.WSAuthZInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.WSManagementInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.WSQueryHistoryInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.WSQueryInterface;

/**
 * WS server impl, based on Jetty/XFire.
 * @author K. Benedyczak
 */
public class VOServer
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, VOServer.class);
	public static final String CONFIGURATION_FILE = "conf/uvosServer.conf";
	
	private ThreadScheduler serverScheduler;
	private UVOSServerProperties c;
	
	private XFire xfire;
	private AnnotationServiceFactory asf;

	private SAMLQueryInterface samlQuery;
	private	SAMLNameIdMappingInterface samlNameIdMapping;
	private SAMLAuthnInterface samlAuthn;
	private WSManagementInterface wsManagement;
	private WSAuthZInterface wsAuthZ;
	private WSQueryInterface wsQuery;
	private WSQueryHistoryInterface wsQueryHistory;
	private WSApplicationsInterface wsApplications;
	private CISInfoProvider wsGLUEInfoProvider;
	private EMIRHandler emirHandler;
	private DB db;
	private JettyServer jetty;
	
	public UVOSServerProperties getConfiguration()
	{
		return c;
	}
	
	public VOServer(UVOSServerProperties config, SAMLQueryInterface samlQuery,
			SAMLNameIdMappingInterface samlNameIdMapping, SAMLAuthnInterface samlAuthn,
			WSManagementInterface wsManagement, WSAuthZInterface wsAuthZ, WSQueryInterface wsQuery,
			WSQueryHistoryInterface wsQueryHistory, WSApplicationsInterface wsApplications,
			CISInfoProvider wsGLUEInfoProvider, EMIRHandler emirHandler,
			DB db, ThreadScheduler timer) throws FileNotFoundException, IOException
	{
		this.c = config;
		this.samlQuery = samlQuery;
		this.samlNameIdMapping = samlNameIdMapping;
		this.samlAuthn = samlAuthn;
		this.wsManagement = wsManagement;
		this.wsAuthZ = wsAuthZ;
		this.wsQuery = wsQuery;
		this.wsQueryHistory = wsQueryHistory;
		this.wsApplications = wsApplications;
		this.wsGLUEInfoProvider = wsGLUEInfoProvider;
		this.emirHandler = emirHandler;
		this.db = db;
		serverScheduler = timer;
	}
	
	protected VOServer(UVOSServerProperties config) throws FileNotFoundException, IOException
	{
		this.c = config;
	}
	
	public void startup(PicoContainer pico) throws Exception
	{
		LogFileWatcher.startLogConfigWatcher(serverScheduler.getScheduledExecutor());
		AttributeTypesLoader attributeLoader = pico.getComponent(AttributeTypesLoader.class);
		attributeLoader.load();

		log.info("Initializing web service engine...");
		initXFire();
		log.info("Configuring services...");
		configureServices();
		log.info("Starting HTTP server...");
		startJetty(pico);
		addToRegistries();
	}
	
	private void configureServices() throws InternalException
	{
		configureService(SAMLQueryInterface.class, samlQuery);
		configureService(SAMLNameIdMappingInterface.class, samlNameIdMapping);
		configureService(SAMLAuthnInterface.class, samlAuthn);
		configureService(WSManagementInterface.class, wsManagement);
		configureService(WSAuthZInterface.class, wsAuthZ);
		configureService(WSQueryInterface.class, wsQuery);
		configureService(WSQueryHistoryInterface.class, wsQueryHistory);
		configureService(WSApplicationsInterface.class, wsApplications);
		configureService(CISInfoProvider.class, wsGLUEInfoProvider);
	}
	
	private void configureService(Class<?> iface, Object impl)
	{
		Service service = asf.create(iface);
		service.setProperty(XmlBeansType.XMLBEANS_NAMESPACE_HACK, "true");
		service.setInvoker(new BeanInvoker(impl));
		xfire.getServiceRegistry().register(service);
		service.addInHandler(new AuthInHandler(false, true, true, null));
		if (c.getBooleanValue(UVOSServerProperties.ETD))
			service.addInHandler(new ETDInHandler(null, c.getAuthAndTrust().getValidator()));
	}
	
	private void initXFire()
	{
		xfire = XFireFactory.newInstance().getXFire();
		
		asf = new AnnotationServiceFactory();
		AegisBindingProvider provider = new AegisBindingProvider(
				new XmlBeansTypeRegistry());
		asf.setBindingProvider(provider);
		asf.setWsdlBuilderFactory(new XmlBeansWSDLBuilderFactory());
	}

	private void startJetty(PicoContainer pico) throws Exception
	{
		jetty = new JettyServer(c, pico);
		jetty.start();
		Runtime.getRuntime().addShutdownHook(new Thread()
		{
			public void run()
			{
				shutdown();
			}
		}	
		);		
	}
	
	public void doShutdown() throws Exception
	{
		db.shutdown();
		jetty.stop();
	}
	
	void shutdown()
	{
		try
		{
			DateFormat df = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.LONG);
			System.out.println(df.format(new Date()) + ": Stopping UVOS Web Server");
			log.info("\n**************************\nStopping UVOS Web Server\n**************************");
			doShutdown();
			log.info("\n**************************\nStopped UVOS Web Server\n**************************");
		} catch (Exception e)
		{
			System.err.println("Severe error when cleaning up" +
				" UVOS Web Server:");
			e.printStackTrace();
		}		
	}

	private void addToRegistries()
	{
		//EMIR handler checks configuration by itself
		emirHandler.start();
		
		boolean doRegister = c.getBooleanValue(UVOSServerProperties.REGISTER);
		if (!doRegister)
			return;
		log.info("Configuring registration in UNICORE registries");
		List<String> registries = new ArrayList<String>();
		String key = UVOSServerProperties.REGISTRY_BASE;
		for (int i=1; ; i++)
		{
			String v =  c.getValue(key + "." + i);
			if (v == null)
				break;
			registries.add(v);
		}
		if (registries.size() == 0)
		{
			log.warn("Registration in external registry is turned ON but no registry " +
					"address is configured. Define '" + 
					UVOSServerProperties.REGISTRY_BASE + ".1' property.");
			return;
		}
		RegistrationHandler regHandler = new RegistrationHandler(registries, c, serverScheduler);
		regHandler.start();
	}
	
	
	static VOServer initializeDI(String[] args) throws Exception
	{
		DateFormat df = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.LONG);
		System.out.println(df.format(new Date()) + ": Starting UVOS Web Server");
		log.info("\n**************************\nStarting UVOS Web Server\n**************************");
		
		String configFile = CONFIGURATION_FILE;
		if (args.length > 0)
			configFile = args[0];
		log.debug("Using configuration file: " + configFile);
		ComponentContainer container = new ComponentContainer(configFile);
		VOServer serverInstance = container.getComponent(VOServer.class);
		serverInstance.startup(container.getContainer());
		log.info("\n**************************\nUVOS Web Server is STARTED\n**************************");
		System.out.println(df.format(new Date()) + ": Started UVOS Web Server");
		return serverInstance;
	}
	
	public static void main(String[] args)
	{
		try
		{
			initializeDI(args);
		} catch (Throwable e)
		{
			System.err.println("Severe error when starting UVOS Web Server:");
			e.printStackTrace();
			log.fatal("Severe error when starting UVOS Web Server", e);
			System.exit(-1);
			throw new RuntimeException("Dummy, can't happen");
		}
	}
}
