/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Aug 7, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.server;

import java.util.Calendar;
import java.util.List;

import javax.jws.WebService;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlException;

import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.UVOSException;
import pl.edu.icm.unicore.uvos.engine.api.UVOSManagementInterface;
import pl.edu.icm.unicore.uvos.server.idresolver.Authenticator;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.UVOSFault;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.WSManagementInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.XMLBeansMapper;
import xmlbeans.pl.edu.icm.uvos.internal.AttributeDocument;
import xmlbeans.pl.edu.icm.uvos.internal.AttributeXType;
import xmlbeans.pl.edu.icm.uvos.internal.AttributesListDocument;
import xmlbeans.pl.edu.icm.uvos.internal.AttributesListType;
import xmlbeans.pl.edu.icm.uvos.internal.ElementDocument;
import xmlbeans.pl.edu.icm.uvos.internal.GroupDocument;
import xmlbeans.pl.edu.icm.uvos.internal.IdentityDocument;
import xmlbeans.pl.edu.icm.uvos.internal.NamesListDocument;
import xmlbeans.pl.edu.icm.uvos.internal.NamesListType;
import xmlbeans.pl.edu.icm.uvos.internal.NotificationListDocument;

/**
 * @author K. Benedyczak
 */
@WebService(endpointInterface="pl.edu.icm.unicore.uvos.wsapi.WSManagementInterface")
public class WSManagementImpl implements WSManagementInterface
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, WSManagementImpl.class);
	private UVOSManagementInterface voMan;
	private Authenticator authenticator;

	public WSManagementImpl(UVOSManagementInterface voMan, Authenticator auth) throws InternalException
	{
		this.voMan = voMan;
		this.authenticator = auth;
	}

	public GroupDocument addGroup(GroupDocument parent, String name) 
	throws UVOSFault
	{
		try
		{
			log.debug("AddGroup invoked");
			authenticator.authenticate();
			Group mparent = XMLBeansMapper.map2APIGroup(parent);
			Group newG = voMan.addGroup(mparent, name);
			if (log.isDebugEnabled())
				log.debug("Added group: " + newG);
			return XMLBeansMapper.map2XMLGroupDoc(newG);
		} catch (UVOSException e)
		{
			log.info("Error invoking addGroup: " + e);
			throw new UVOSFault(e);
		}
	}

	public void removeGroup(GroupDocument toRemove, boolean recursive) 
	throws UVOSFault
	{
		try
		{
			log.debug("RemoveGroup invoked");
			authenticator.authenticate();
			Group g = XMLBeansMapper.map2APIGroup(toRemove);
			voMan.removeGroup(g, recursive);
			if (log.isDebugEnabled())
				log.debug("Removed group " + g);			
		} catch(UVOSException e)
		{
			log.info("Error invoking removeGroup: " + e);
			throw new UVOSFault(e);
		}
	}


	public GroupDocument copyGroup(GroupDocument toCopy, GroupDocument newParent, 
			String newName, boolean deleteOrig) throws UVOSFault
	{
		try
		{
			log.debug("CopyGroup invoked");
			authenticator.authenticate();
			Group src = XMLBeansMapper.map2APIGroup(toCopy);
			Group dst = XMLBeansMapper.map2APIGroup(newParent);
			Group moved = voMan.copyGroup(src, dst,
					newName, deleteOrig);
			if (log.isDebugEnabled())
				log.debug("Copied group " + src + " to " + dst);			
			return XMLBeansMapper.map2XMLGroupDoc(moved);
		} catch (UVOSException e)
		{
			log.info("Error invoking copyGroup: " + e);
			throw new UVOSFault(e);
		}
	}


	public void addIdentity(IdentityDocument toAdd) 
	throws UVOSFault
	{
		try
		{
			log.debug("AddIdentity invoked");
			authenticator.authenticate();
			Identity id = XMLBeansMapper.map2APIIdentity(toAdd);
			voMan.addIdentity(id);
			if (log.isDebugEnabled())
				log.debug("Added identity " + id.prettyString());			
		} catch (UVOSException e)
		{
			log.info("Error invoking addIdentity: " + e);
			throw new UVOSFault(e);
		}
	}

	public void addIdentity(IdentityDocument toAdd, IdentityDocument equivalentIdentity) 
	throws UVOSFault
	{
		try
		{
			log.debug("AddIdentity (with equiv) invoked");
			authenticator.authenticate();
			Identity id = XMLBeansMapper.map2APIIdentity(toAdd);
			Identity eq = XMLBeansMapper.map2APIIdentity(equivalentIdentity);
			voMan.addIdentity(id, eq);
			if (log.isDebugEnabled())
				log.debug("Added identity " + id.prettyString() + 
					" as equivalent to the " + eq.prettyString());			
		} catch (UVOSException e)
		{
			log.info("Error invoking addIdentity: " + e);
			throw new UVOSFault(e);
		}
	}

	public void removeIdentity(IdentityDocument toRemove) 
	throws UVOSFault
	{
		try
		{
			log.debug("RemoveIdentity invoked");
			authenticator.authenticate();
			Identity id = XMLBeansMapper.map2APIIdentity(toRemove);
			voMan.removeIdentity(id);
			if (log.isDebugEnabled())
				log.debug("Removed identity " + id.prettyString());
		} catch (UVOSException e)
		{
			log.info("Error invoking removeIdentity: " + e);
			throw new UVOSFault(e);
		}
	}

	public void setAttribute(ElementDocument whom, AttributeDocument toAdd, boolean update) 
	throws UVOSFault
	{
		try
		{
			log.debug("SetAttribute invoked");
			authenticator.authenticate();
			Element el = XMLBeansMapper.map2APIElement(whom);
			Attribute at = XMLBeansMapper.map2APIAttribute(toAdd);
			voMan.setAttribute(el, at, update);
			if (log.isDebugEnabled())
				log.debug("Set " + el.prettyString() + " attribute " + at);			
		} catch (UVOSException e)
		{
			log.info("Error invoking setAttribute: " + e);
			throw new UVOSFault(e);
		} catch (XmlException e)
		{
			log.info("Error invoking setAttribute: " + e);
			throw new UVOSFault(new InvalidValueException(e.getMessage()));
		}
	}

	public void removeAttribute(ElementDocument whom, String toRemove) 
	throws UVOSFault
	{
		try
		{
			log.debug("RemoveAttribute invoked");
			authenticator.authenticate();
			Element el = XMLBeansMapper.map2APIElement(whom);
			voMan.removeAttribute(el, toRemove);
			if (log.isDebugEnabled())
				log.debug("Removed " + el.prettyString() + " attribute " + toRemove);
		} catch (UVOSException e)
		{
			log.info("Error invoking removeAttribute: " + e);
			throw new UVOSFault(e);
		}
	}

	public void addToGroup(IdentityDocument toAdd, GroupDocument group) 
	throws UVOSFault
	{
		try
		{
			log.debug("AddToGroup invoked");
			authenticator.authenticate();
			Identity id = XMLBeansMapper.map2APIIdentity(toAdd);
			Group g = XMLBeansMapper.map2APIGroup(group);
			voMan.addToGroup(id, g);
			if (log.isDebugEnabled())
				log.debug("Added " + id.prettyString() + " to group " + g);
		} catch (UVOSException e)
		{
			log.info("Error invoking addToGroup: " + e);
			throw new UVOSFault(e);
		}
	}

	public void removeFromGroup(IdentityDocument toRemove, GroupDocument group) 
	throws UVOSFault
	{
		try
		{
			log.debug("RemoveFromGroup invoked");
			authenticator.authenticate();
			Identity id = XMLBeansMapper.map2APIIdentity(toRemove);
			Group g = XMLBeansMapper.map2APIGroup(group);
			voMan.removeFromGroup(id, g);
			if (log.isDebugEnabled())
				log.debug("Removed " + id.prettyString() + " from group " + g);
		} catch (UVOSException e)
		{
			log.info("Error invoking removeFromGroup: " + e);
			throw new UVOSFault(e);
		}
	}

	public void setIdentityLabel(IdentityDocument toChange, String label) 
	throws UVOSFault
	{
		try
		{
			log.debug("SetIdentityLabel invoked");
			authenticator.authenticate();
			Identity id = XMLBeansMapper.map2APIIdentity(toChange);
			voMan.setIdentityLabel(id, label);
			if (log.isDebugEnabled())
				log.debug("Identity " + id.prettyString() + " label was changed to " + label);
		} catch (UVOSException e)
		{
			log.info("Error invoking setIdentityLabel: " + e);
			throw new UVOSFault(e);
		}
	}

	public AttributesListDocument getAttributeTypes() 
		throws UVOSFault
	{
		try
		{
			log.debug("GetAttributeTypes invoked");
			authenticator.authenticate();
			List<AttributeType> rawRet = voMan.getAttributeTypes();
			AttributesListDocument retDoc = AttributesListDocument.Factory.newInstance();
			AttributesListType ret = retDoc.addNewAttributesList();
			for (AttributeType idT: rawRet)
			{
				AttributeXType vv = ret.addNewElement();
				vv.setName(idT.toString());
				vv.setDescription(idT.getDescription());
				vv.setShortDescription(idT.getShortDescription());
			}
			return retDoc;
		} catch (UVOSException e)
		{
			log.info("Error invoking getAttributeTypes: " + e);
			throw new UVOSFault(e);
		}
	}

	public NamesListDocument getIdentityTypes() 
	throws UVOSFault
	{
		try
		{
			log.debug("GetIdentityTypes invoked");
			authenticator.authenticate();
			List<IdentityType> rawRet = voMan.getIdentityTypes();
			NamesListDocument retDoc = NamesListDocument.Factory.newInstance();
			NamesListType ret = retDoc.addNewNamesList();
			for (IdentityType idT: rawRet)
				ret.addElement(idT.toString());
			return retDoc;
		} catch (UVOSException e)
		{
			log.info("Error invoking getIdentityTypes: " + e);
			throw new UVOSFault(e);
		}
	}

	public void disableAttribute(ElementDocument whose, String toDisable, String valueToDisable) 
	throws UVOSFault
	{
		try
		{
			log.debug("DisableAttribute invoked");
			authenticator.authenticate();
			Element el = XMLBeansMapper.map2APIElement(whose);
			voMan.disableAttribute(el, toDisable, valueToDisable);
			if (log.isDebugEnabled())
				log.debug("Disabled " + el.prettyString() + " attribute named " + toDisable + 
					" value " + valueToDisable);			
		} catch (UVOSException e)
		{
			log.info("Error invoking disableAttribute: " + e);
			throw new UVOSFault(e);
		}
	}

	public void enableAttribute(ElementDocument whose, String toEnable, String valueToEnable) 
	throws UVOSFault
	{
		try
		{
			log.debug("EnableAttribute invoked");
			authenticator.authenticate();
			Element el = XMLBeansMapper.map2APIElement(whose);
			voMan.enableAttribute(el, toEnable, valueToEnable);
			if (log.isDebugEnabled())
				log.debug("Disabled " + el.prettyString() + " attribute named " + 
					toEnable + " value " + valueToEnable);
		} catch (UVOSException e)
		{
			log.info("Error invoking enableAttribute: " + e);
			throw new UVOSFault(e);
		}
	}

	public AttributesListDocument getDisabledAttributes(ElementDocument whose) 
	throws UVOSFault
	{
		try
		{
			log.debug("GetDisabledAttributes invoked");
			authenticator.authenticate();
			List<Attribute> rawRet;
			rawRet = voMan.getDisabledAttributes(
					XMLBeansMapper.map2APIElement(whose));
			return XMLBeansMapper.map2XMLAttributesList(rawRet);
		} catch (UVOSException e)
		{
			log.info("Error invoking getDisabledAttributes: " + e);
			throw new UVOSFault(e);
		}
	}

	public void purgeHistoricalData(Calendar arg0) 
		throws UVOSFault
	{
		try
		{
			log.debug("PurgeHistoricalData invoked");
			authenticator.authenticate();

			voMan.purgeHistoricalData(XMLBeansMapper.c2d(arg0));
			if (log.isDebugEnabled())
				log.debug("Historical data older than " + arg0.toString() + 
					" was deleted.");			
		} catch (UVOSException e)
		{
			log.info("Error invoking purgeHistoricalData: " + e);
			throw new UVOSFault(e);
		}
	}

	public void updateAttributeTypes(AttributesListDocument toUpdate, boolean clear) 
		throws UVOSFault
	{
		try
		{
			log.debug("UpadteAttributeTypes invoked");
			authenticator.authenticate();

			voMan.updateAttributeTypes(
					XMLBeansMapper.map2APIAttributeTypesList(toUpdate), 
					clear);
		} catch (UVOSException e)
		{
			log.info("Error invoking updateAttributeTypes: " + e);
			throw new UVOSFault(e);
		}
	}

	public void setIdentityStatus(IdentityDocument toUpdate, boolean how) 
		throws UVOSFault
	{
		try
		{
			log.debug("SetIdentityStatus invoked");
			authenticator.authenticate();

			Identity id = XMLBeansMapper.map2APIIdentity(toUpdate);
			voMan.setIdentityStatus(id, how);
			if (log.isDebugEnabled())
				log.debug("Identity " + id.prettyString() + " status is now " + 
					(how ? "ENABLED" : "DISABLED"));
		} catch (UVOSException e)
		{
			log.info("Error invoking setIdentityStatus: " + e);
			throw new UVOSFault(e);
		}
		
	}

	public void addNotification(String action, String receipents, String groupFilter) 
		throws UVOSFault
	{
		try
		{
			log.debug("AddNotification invoked");
			authenticator.authenticate();

			voMan.addNotification(action, receipents, groupFilter);
			if (log.isDebugEnabled())
				log.debug("Added notififcation for action " + action + 
					" with group filter " + groupFilter + 
					" which will be send to " + receipents);			
		} catch (UVOSException e)
		{
			log.info("Error invoking addNotification: " + e);
			throw new UVOSFault(e);
		}
	}

	public NotificationListDocument getNotifications(String action, String groupFilter) 
		throws UVOSFault
	{
		try
		{
			log.debug("GetNotifications invoked");
			authenticator.authenticate();

			return XMLBeansMapper.map2XMLNotificationsList(
				voMan.getNotifications(action, groupFilter));
		} catch (UVOSException e)
		{
			log.info("Error invoking getNotifications: " + e);
			throw new UVOSFault(e);
		}
	}

	public void removeNotification(int id) throws UVOSFault
	{
		try
		{
			log.debug("RemoveNotification invoked");
			authenticator.authenticate();
			if (log.isDebugEnabled())
				log.debug("Removed notification id " + id);			

			voMan.removeNotification(id);
		} catch (UVOSException e)
		{
			log.info("Error invoking removeNotification: " + e);
			throw new UVOSFault(e);
		}
	};
}
