/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Jan 13, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.server.saml;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

import eu.unicore.samly2.SAMLConstants;
import eu.unicore.samly2.assertion.Assertion;
import eu.unicore.samly2.elements.SAMLAttribute;
import eu.unicore.samly2.exceptions.SAMLRequestException;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.saml.SAMLVOConstants;

/**
 * Handles mapping of UVOS attributes and groups onto SAML attributes.
 * @author K. Benedyczak
 */
public class AttributeConverter
{
	//TODO move it somewhere from here?
	public static final AttributeType GROUP_AT = new AttributeType(
		SAMLVOConstants.GROUP_ATTR, "Group membership", 
		"Attribtue bearer is a memebr of the groups " +
		"listed as this attribute values.");
	
	//currently all attributes are given string XACML type. Can be made
	//more flexible in the future.
	public static final String XACML_DT = SAMLConstants.XACMLDT_STRING;
	
	//key is an attribute's name
	private HashMap<String, AttributesBag> preparedAttrs;
	
	public AttributeConverter()
	{
		preparedAttrs = new HashMap<String, AttributesBag>();
	}
	
	/**
	 * Adds list of groups to the set of SAML attributes. May be called several 
	 * times.
	 * @param groups
	 */
	public void preInjectGroups(List<Group> groups)
	{
		if (groups.size() == 0)
			return;
		
		AttributesBag curVal = preparedAttrs.get(
				SAMLVOConstants.GROUP_ATTR);
		if (curVal == null)
		{
			curVal = new AttributesBag(SAMLConstants.XACMLDT_STRING, 
					SAMLConstants.SCOPE_TYPE_NONE);
			preparedAttrs.put(SAMLVOConstants.GROUP_ATTR, curVal);
		}
		Attribute a = new Attribute(GROUP_AT, null, null);
		for (Group group: groups)
			a.addValue(group.toString());
		
		curVal.attributes.add(a);
	}
	
	/**
	 * Adds attributes list to the set of SAML attributes. 
	 * All attributes in the list must have the same name, what is not checked here. (<-??? is it true)
	 * May be called several times but always attributes added must have a different name
	 * than those already added.
	 * @param groups
	 * @throws SAMLRequestException 
	 */
	public void preInjectAttributes(List<Attribute> attr, String requestedScopeType) 
		throws SAMLRequestException
	{
		if (attr.size() == 0)
			return;
		HashMap<String, AttributesBag> tmpAttrs = new HashMap<String, AttributesBag>();
		for (Attribute a: attr)
		{
			String name = a.getName().toString();
			AttributesBag curVal = preparedAttrs.get(name);
			if (curVal != null)
			{
				//Return an error as it means that attribute with the same
				//name was queried multiple times. It is SAML protocol violation
				//and would cause implementation problems.
				throw new SAMLRequestException(SAMLConstants.STATUS2_INVALID_ATTR,
						"Invalid query: attribute must be specified only " +
						"once according to SAML specification");
			}
			
			curVal = tmpAttrs.get(name);
			String current = curVal == null ? null : curVal.scopeType;
			String scopingType = establishscopingType(requestedScopeType, a, current);
			
			String xacmlDT = scopingType.equals(SAMLConstants.SCOPE_TYPE_ATTRIBUTE) ?
					SAMLConstants.XACMLDT_SCOPED_STRING : 
					SAMLConstants.XACMLDT_STRING;
			if (curVal == null)
			{
				curVal = new AttributesBag(xacmlDT, scopingType);
				tmpAttrs.put(name, curVal);
			} else
			{
				curVal.scopeType = scopingType;
				curVal.xacmlDT = xacmlDT;
			}
			curVal.attributes.add(a);
		}
		preparedAttrs.putAll(tmpAttrs);
	}
	
	/**
	 * Inserts all attributes preInjected before to the assertion given.
	 * There is filtering here: all attributes that are scoped but scopeingType
	 * is NONE are not injected.
	 * @param assertion
	 */
	public void injectAttributes(Assertion assertion)
	{
		if (preparedAttrs.size() == 0)
			return;
		Iterator<Entry<String, AttributesBag>> entries = 
			preparedAttrs.entrySet().iterator();
		while(entries.hasNext())
		{
			Entry<String, AttributesBag> entry = entries.next();
			String xacmlType = entry.getValue().xacmlDT;
			String scopingType = entry.getValue().scopeType;
			
			SAMLAttribute samlA = new SAMLAttribute(entry.getKey(),
					SAMLConstants.AFORMAT_URI);
			String shortDesc = entry.getValue().
				attributes.get(0).getType().getShortDescription();
			if (shortDesc != null)
				samlA.setFriendlyName(shortDesc);
			samlA.setXACMLDataType(xacmlType);
			if (scopingType != SAMLConstants.SCOPE_TYPE_NONE)
				samlA.setScopingType(scopingType);
			
			for (Attribute a: entry.getValue().attributes)
				addAttribute(samlA, a, scopingType);
			assertion.addAttribute(samlA);
		}
	}	

	private void addAttribute(SAMLAttribute samlA, Attribute a, String scopingType)
	{
		List<String> vals = a.getValues();
		if (scopingType.equals(SAMLConstants.SCOPE_TYPE_SIMPLE))
		{
			String scope = a.getScope();
			scope = (scope == null) ? "@/" : "@" + scope; 
			for (String val: vals)
				samlA.addStringAttributeValue(val + scope);
			if (vals.size() == 0)
				samlA.addStringAttributeValue(scope);
		} else if (scopingType.equals(SAMLConstants.SCOPE_TYPE_ATTRIBUTE))
		{
			String scope = a.getScope();
			for (String val: vals)
				samlA.addScopedStringAttributeValue(val, 
					scope == null ? "/" : scope);
			if (vals.size() == 0)
				samlA.addScopedStringAttributeValue(null, 
					scope == null ? "/" : scope);
		} else
		{
			//filtering
			if (a.getScope() != null)
				return;
			for (String val: vals)
				samlA.addStringAttributeValue(val);
		}
	}
	
	/**
	 * Algorithm:
	 *  - when there is requested type by user then just use 
	 *  it after checking if it is supported
	 *  - check if attribute has scope defined. If true then return SCOPE_TYPE_ATTRIBUTE
	 *  - If there is no scope defined return current type (if set -> previous attrs 
	 *  could have scope defined) or SCOPE_TYPE_NONE otherwise. 
	 *  
	 * @param requested
	 * @param newAttributes
	 * @param current
	 * @return
	 * @throws SAMLRequestException
	 */
	private String establishscopingType(String requested, Attribute newAttribute, 
			String current) 
		throws SAMLRequestException
	{
		if (requested != null)
		{
			if (!requested.equals(SAMLConstants.SCOPE_TYPE_ATTRIBUTE) 
				&& !requested.equals(SAMLConstants.SCOPE_TYPE_SIMPLE)
				&& !requested.equals(SAMLConstants.SCOPE_TYPE_NONE))
				throw new SAMLRequestException(SAMLConstants.STATUS2_INVALID_ATTR,
						"Unsupported ");
			return requested;
		}
		boolean isScopeDefined = false;
		if (newAttribute.getScope() != null)
			isScopeDefined = true;
		if (isScopeDefined)
			return SAMLConstants.SCOPE_TYPE_ATTRIBUTE;
		else
			return current == null ? SAMLConstants.SCOPE_TYPE_NONE : current;
	}
	
	private static class AttributesBag
	{
		String xacmlDT;
		String scopeType;
		List<Attribute> attributes;
		
		public AttributesBag(String xacml, String scopeType)
		{
			xacmlDT = xacml;
			this.scopeType = scopeType;
			attributes = new ArrayList<Attribute>();
		}
	}
}
