package pl.edu.icm.unicore.uvos.server.saml;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.regex.PatternSyntaxException;
import org.apache.log4j.Logger;

import eu.unicore.util.Log;
import eu.unicore.util.configuration.FilePropertiesHelper;
import eu.unicore.util.configuration.PropertyMD;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.util.AttributeFilter;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;
import pl.edu.icm.unicore.uvos.util.FileWatcher;
import pl.edu.icm.unicore.uvos.util.ThreadScheduler;

/**
 * Configuraton-driven attribute filter
 * @author K. Stasiak
 */
public class AttributeFilterImpl implements AttributeFilter 
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, AttributeFilterImpl.class);
	private FilePropertiesHelper conf;
	private final String configFile;
	private List<AttributeFilterPattern> excluded = new ArrayList<AttributeFilterPattern>();
	private List<AttributeFilterPattern> included = new ArrayList<AttributeFilterPattern>();
	private ReadWriteLock lock = new ReentrantReadWriteLock();
	private int delay;

	public final static Map<String, PropertyMD> META=new HashMap<String, PropertyMD>();
	static
	{
		META.put(UVOSServerProperties.ATTRIBUTE_FILTER_EXPOSED_ATTRIBUTES, new PropertyMD().setCanHaveSubkeys());
		META.put(UVOSServerProperties.ATTRIBUTE_FILTER_EXCLUDED_ATTRIBUTES, new PropertyMD().setCanHaveSubkeys());
	}
	
	public AttributeFilterImpl(UVOSServerProperties baseConf, ThreadScheduler scheduler) throws IOException
	{
		this(baseConf.getValue(UVOSServerProperties.ATTRIBUTE_FILTER_FILE), 5000, scheduler);
	}

	public AttributeFilterImpl(String configFile, int recheckDelay, ThreadScheduler scheduler) throws IOException
	{
		this.configFile = configFile;
		this.delay = recheckDelay;
		readConfiguration();
		startConfigWatcher(scheduler.getScheduledExecutor());
	}

	private void readConfiguration() throws IOException
	{
		lock.writeLock().lock();
		try
		{
			conf = new FilePropertiesHelper("", configFile,	META, log);
			excluded.clear();
			included.clear();
			for (int x = 1; true; x++) 
			{
				if (!addRule(x, UVOSServerProperties.ATTRIBUTE_FILTER_EXCLUDED_ATTRIBUTES, 
						true))
					break;
			}
			for (int x = 1; true; x++) 
			{
				if (!addRule(x, UVOSServerProperties.ATTRIBUTE_FILTER_EXPOSED_ATTRIBUTES, 
						false))
					break;
			}
		} finally
		{
			lock.writeLock().unlock();
		}
	}
	
	private void startConfigWatcher(ScheduledExecutorService timer)
	{
		try
		{
			Runnable r = new Runnable()
			{
				public void run()
				{
					log.info("ATTRIBUTE FILTER CONFIGURATION MODIFIED, re-configuring.");
					try
					{
						readConfiguration();
					} catch (IOException e)
					{
						log.error("Failed to reload the attribute release policy", e);
					}
				}
			};
			FileWatcher fw = new FileWatcher(new File(configFile), r);
			timer.scheduleWithFixedDelay(fw, delay, delay, TimeUnit.MILLISECONDS);
		}catch (FileNotFoundException fex)
		{
			log.warn("Attribute filter configuration file <" + configFile + "> not found.");
		}
	}

	
	private boolean addRule(int x, String cfgPfx, boolean mode)
	{
		String key = cfgPfx + "." + x;
		String regex = conf.getValue(key);
		if (regex == null) 
			return false;
		AttributeFilterPattern afp;
		try 
		{
			afp = new AttributeFilterPattern(regex);
		} catch (PatternSyntaxException ex) {
			log.error("Invalid regular expression: '" + regex + "', will be ignored");
			return true;
		}
			
		log.debug((mode ? "Hiding" : "Exposing") + " attributes matching '"+regex+"'");
		for (int y = 1; true; y++) 
		{
			String lkey = key + UVOSServerProperties.ATTRIBUTE_FILTER_ALLOWED_SCOPE_INFIX + y;
			String regex2 = conf.getValue(lkey);
			if (regex2 == null) 
				break;
			try 
			{
				afp.addAllowedScope(regex2);
				log.debug(" included for '" + regex2 + "' scope");
			} catch (PatternSyntaxException ex) 
			{
				log.error("Invalid regular expression for property " + lkey + 
						": '" + regex2 + "', will be ignored");
			}
		}
		for (int y = 1; true; y++) 
		{
			String lkey = key + UVOSServerProperties.ATTRIBUTE_FILTER_IGNORED_SCOPE_INFIX + y;
			String regex2 = conf.getValue(lkey);
			if (regex2 == null) 
				break;
			try 
			{
				afp.addForbiddenScope(regex2);
				log.debug(" excluded for '" + regex2 + "' scope");
			} catch (PatternSyntaxException ex) 
			{
				log.error("Invalid regular expression for property " + lkey + 
						": '" + regex2 + "', will be ignored");
			}
		}
		if (mode)
			excluded.add(afp);
		else
			included.add(afp);
		return true;
	}
	
	@Override
	public List<Attribute> filter(List<Attribute> attribs) 
	{
		lock.readLock().lock();
		try
		{
			if (included.isEmpty() && excluded.isEmpty()) 
			{
				log.trace("No attribute filters defined, returning an original attribute list");
				return attribs;
			}
			List<Attribute> res = new ArrayList<Attribute>(attribs.size());

			big_for:
				for (Attribute a : attribs) 
				{
					String name = a.getName();
					String scope = a.getScope();
					for (AttributeFilterPattern excludingPattern : excluded) 
					{
						if (excludingPattern.doesApplyAndMatch(name, scope)) 
						{
							log.debug("Attribute " + a + " filtered by an excluding pattern " 
									+ excludingPattern);
							continue big_for;
						}
					}
					if (included.isEmpty()) 
					{
						log.trace("Attribute " + a + " accepted due to no defined inclusion rules");
						res.add(a);
					} else 
					{
						for (AttributeFilterPattern includingPattern : included) 
						{
							if (includingPattern.doesApplyAndMatch(name, scope)) 
							{
								log.debug("Attribute " + a + " accepted by an including pattern " 
										+ includingPattern);
								res.add(a);
								continue big_for;
							}
						}
						log.debug("Attribute " + a + " filtered as it is not matching any including pattern.");
					}
				}
			return res;
		} finally
		{
			lock.readLock().unlock();
		}
	}

	@Override
	public boolean allowsEverything()
	{
		return excluded.size() == 0 && included.size() == 0;
	}
}
