package pl.edu.icm.unicore.uvos.server.saml;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

/**
 * Utility class. Represents a single set of uvos.server.exposedAttributes.* or
 * uvos.server.excludedAttributes.* from cinfig file; Matches attributes using
 * their names and scopes.
 * 
 * @author K. Stasiak
 */
public class AttributeFilterPattern
{
	private Pattern namePattern;

	private List<Pattern> allowedScopes, ignoredScopes;

	/**
	 * Creates new filter pattern, given attribute name pattern in form of
	 * regular expression.
	 * 
	 * @param regex
	 *                regular expression for an attribute name
	 */
	public AttributeFilterPattern(String regex) throws PatternSyntaxException
	{
		namePattern = Pattern.compile("^" + regex + "$");
		allowedScopes = new ArrayList<Pattern>();
		ignoredScopes = new ArrayList<Pattern>();
	}

	/**
	 * Adds an allowed scope name pattern in form of regular expression. If
	 * there are allowed scope name patterns added, only attributes that
	 * have scopes with names matching this pattern will match this filter
	 * pattern. Forbidden patterns have priority over allowed patterns.
	 * 
	 * @param regex
	 *                regular expression for a scope name
	 */
	public void addAllowedScope(String regex) throws PatternSyntaxException
	{
		allowedScopes.add(Pattern.compile("^" + regex + "$"));
	}

	/**
	 * Adds a forbidden scope name pattern in form of regular expression.
	 * Attributes that have scopes with names matching this pattern will not
	 * match this filter pattern.
	 * 
	 * @param regex
	 *                regular expression for a scope name
	 */
	public void addForbiddenScope(String regex) throws PatternSyntaxException
	{
		ignoredScopes.add(Pattern.compile("^" + regex + "$"));
	}

	/**
	 * Checks if:
	 * <ul>
	 * <li>scope name matches no forbidden scope name patterns, and</li>
	 * <li>if there are any allowed scope name patterns, scope name matches
	 * at least one</li>
	 * </ul>
	 * 
	 * @param scope
	 *                attribute scope name
	 * @return true if both conditions satisfied, false otherwise
	 */
	public boolean doesApply(String scope)
	{
		if (scope == null)
		{
			scope = "";
		}
		for (Pattern pattern : ignoredScopes)
		{
			if (pattern.matcher(scope).matches())
			{
				return false;
			}
		}
		if (allowedScopes.isEmpty())
		{
			return true;
		}
		for (Pattern pattern : allowedScopes)
		{
			if (pattern.matcher(scope).matches())
			{
				return true;
			}
		}
		return false;
	}

	/**
	 * Checks whether attribute name matches pattern
	 * 
	 * @param name
	 *                attribute name
	 * @return true if matches, false otherwise
	 */
	public boolean doesNameMatch(String name)
	{
		return namePattern.matcher(name).matches();
	}

	/**
	 * Checks if:
	 * <ul>
	 * <li>attribute name matches attribute name pattern</li>
	 * <li>scope name matches no forbidden scope name patterns, and</li>
	 * <li>if there are any allowed scope name patterns, scope name matches
	 * at least one</li>
	 * </ul>
	 * 
	 * @param name
	 *                attribute name
	 * @param scope
	 *                attribute sope name
	 * @return true if all three conditions are true, false otherwise
	 */
	public boolean doesApplyAndMatch(String name, String scope)
	{
		return doesNameMatch(name) && doesApply(scope);
	}

	@Override
	public String toString()
	{
		return namePattern + ", allowed scopes:" + allowedScopes
				+ ", excluded scopes:" + ignoredScopes;
	}
}
