/*
 * Copyright (c) 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Jul 15, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.server.saml;

import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.List;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlOptions;

import eu.unicore.samly2.SAMLConstants;
import eu.unicore.samly2.assertion.Assertion;
import eu.unicore.samly2.elements.Subject;
import eu.unicore.samly2.exceptions.SAMLProtocolException;
import eu.unicore.samly2.exceptions.SAMLRequestException;
import eu.unicore.samly2.proto.AssertionResponse;
import eu.unicore.security.dsig.DSigException;
import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.exceptions.AuthenticationException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.UVOSException;
import pl.edu.icm.unicore.uvos.engine.api.UVOSAuthZInterface;
import pl.edu.icm.unicore.uvos.idhelpers.X509IdentityHelper;
import pl.edu.icm.unicore.uvos.saml.SAMLAuthnUtil;
import pl.edu.icm.unicore.uvos.saml.SAMLConfiguration;
import pl.edu.icm.unicore.uvos.saml.SAMLErrorException;
import pl.edu.icm.unicore.uvos.server.IssuerIdentityProvider;
import pl.edu.icm.unicore.uvos.server.idresolver.Authenticator;
import pl.edu.icm.unicore.uvos.server.idresolver.IdResolveException;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.SAMLAuthnInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.UVOSFault;
import xmlbeans.org.oasis.saml2.assertion.AuthnContextType;
import xmlbeans.org.oasis.saml2.assertion.SubjectConfirmationDataType;
import xmlbeans.org.oasis.saml2.assertion.SubjectConfirmationType;
import xmlbeans.org.oasis.saml2.protocol.AuthnRequestDocument;
import xmlbeans.org.oasis.saml2.protocol.AuthnRequestType;
import xmlbeans.org.oasis.saml2.protocol.ResponseDocument;

/**
 * This is implementation of SAML Authentication protocol. It can be used
 * with a SOAP binding (mostly for testing) or internally (e.g. by web application
 * supporting HTTP POST binding).
 * @author golbi
 */
public class SAMLAuthImpl implements SAMLAuthnInterface
{
	public static final String SAML_AC_UNSPEC = 
		"urn:oasis:names:tc:SAML:2.0:ac:classes:unspecified";
	private UVOSAuthZInterface voAuthz;
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, SAMLAuthImpl.class);
	private int reqValidity;
	private boolean allowX509AsDN;
	private UVOSServerProperties conf;
	private IssuerIdentityProvider cryptoProvider;
	private Authenticator authenticator;
	private ReplayAttackChecker repChecker;
	private SAMLConfiguration samlConfiguration;
	
	public SAMLAuthImpl(UVOSAuthZInterface voAuthz, UVOSServerProperties conf, IssuerIdentityProvider cryptoProvider,
			Authenticator authenticator, ReplayAttackChecker repChecker, SAMLConfiguration samlConfig)
	{
		this.samlConfiguration=samlConfig;
		this.voAuthz = voAuthz;
		this.conf = conf;
		this.cryptoProvider = cryptoProvider;
		this.authenticator = authenticator;
		this.repChecker = repChecker;
		reqValidity = conf.getIntValue(UVOSServerProperties.SAML_REQUEST_VALIDITY);
		allowX509AsDN = conf.getBooleanValue(UVOSServerProperties.MAP_CERT_TO_DN);
	}

	public ResponseDocument authnRequest(AuthnRequestDocument reqDoc) 
		throws UVOSFault
	{
		if (log.isTraceEnabled())
			log.trace("Received request: " + 
				reqDoc.xmlText(new XmlOptions().setSavePrettyPrint()));
		Calendar authTime = Calendar.getInstance();

		//CryptoDataProvider cryptoProvider = VOServer.getCryptoProvider();
		SAMLAuthnUtil samlUtil = new SAMLAuthnUtil(reqDoc, cryptoProvider, conf, samlConfiguration);
		try
		{
			authenticator.resolve();
		} catch (IdResolveException e)
		{
			log.debug("UVOS engine problem while processing " +
					"SAML request (authN stage )", e);
			throw new UVOSFault(e);
		} catch (UVOSException e)
		{
			log.debug("Authentication FAILED: " + e);
			String info = "Authentication failed: " + ((e.getMessage() == null) ? 
					e.getClass().getSimpleName() : e.getMessage());
			return samlUtil.getErrorResponse(new AuthenticationException(), 
					info).getDoc();
		}
		AuthnContextType authContext = setupAuthnContext();
		
		SAMLAuthnUtil samlCtx;
		try
		{
			samlCtx = preprocessAuthnRequest(reqDoc);
		} catch (SAMLErrorException e)
		{
			return e.getResponse();
		}
		
		return processAuthnRequest(samlCtx, authContext, authTime);
	}
	
	/**
	 * Only unspecified - it's too much work to implement it fully
	 * with minimal effect.
	 * @return
	 */
	public AuthnContextType setupAuthnContext()
	{
		AuthnContextType ret = AuthnContextType.Factory.newInstance();
		ret.setAuthnContextClassRef(SAML_AC_UNSPEC);
		return ret;
	}
	
	public SAMLAuthnUtil preprocessAuthnRequest(AuthnRequestDocument reqDoc) 
		throws SAMLErrorException
	{
		//CryptoDataProvider cryptoProvider = VOServer.getCryptoProvider();
		SAMLAuthnUtil samlUtil = new SAMLAuthnUtil(reqDoc, cryptoProvider, conf, samlConfiguration);
		
		try
		{
			samlUtil.parse();
			samlUtil.checkRequestValidity(reqValidity);
			if (!samlUtil.isRequestCorrectlySigned())
			{
				log.info("Got request with invalid signature: " +
						reqDoc.xmlText());
				throw new SAMLErrorException(
					samlUtil.getErrorResponse(new SAMLRequestException(null, 
					"Request signature is invalid")).getDoc());
			}
		} catch (SAMLProtocolException e)
		{
			log.debug("Parsing the SAML request FAILED: " + e);
			throw new SAMLErrorException(samlUtil.getErrorResponse(e).getDoc());
		} catch (DSigException e)
		{
			log.info("Signature verification problem: " + e);
			throw new SAMLErrorException(
				samlUtil.getErrorResponse(new SAMLRequestException(null, 
					"Signature can't be verified: " + 
					e.getMessage())).getDoc());
		}

		try
		{
			checkUnsupportedConstraints(reqDoc);
		} catch (SAMLProtocolException e)
		{
			log.debug("Ignoring unsupported authn request: " + e.getMessage());
			throw new SAMLErrorException(samlUtil.getErrorResponse(e).getDoc());
		}
		
		//ReplayAttackChecker repChecker = Manager.getInstance().getReplayChecker();
		try
		{
			repChecker.checkAndAdd(samlUtil.getReqId(), System.currentTimeMillis());
		} catch (Exception e2)
		{
			String msg = "Request is denied: " + e2.getMessage(); 
			log.info(msg);
			throw new SAMLErrorException(samlUtil.getErrorResponse(
				new SAMLRequestException(null, msg)).getDoc());
		}
		return samlUtil;
	}
	
	public ResponseDocument processAuthnRequest(SAMLAuthnUtil samlUtil, 
			AuthnContextType authContext, Calendar authTime) throws UVOSFault
	{
		//CryptoDataProvider cryptoProvider = VOServer.getCryptoProvider();
		String format = samlUtil.getRequestedFormat();
		List<Identity> equivs;
		try
		{
			equivs = voAuthz.getMyIdentities();
		} catch (UVOSException e1)
		{
			return samlUtil.getErrorResponse(e1).getDoc();
		}
		if (equivs.size() == 0)
		{
			String msg = "getAllEquivalents returned empty list";
			log.error(msg);
			throw new UVOSFault(new InternalException(msg));
		}
				
		Subject authenticatedOne;
		try
		{
			authenticatedOne = findUser(format, equivs);
		} catch (SAMLProtocolException e)
		{
			return samlUtil.getErrorResponse(e).getDoc();
		}
		
		SubjectConfirmationType subConf = SubjectConfirmationType.Factory.newInstance();
		subConf.setMethod(SAMLConstants.CONFIRMATION_BEARER);
		SubjectConfirmationDataType confData = subConf.addNewSubjectConfirmationData();
		confData.setInResponseTo(samlUtil.getReqId());
		confData.setRecipient(samlUtil.getConsumerURL());
		Calendar validity = Calendar.getInstance();
		validity.setTimeInMillis(authTime.getTimeInMillis());
		validity.add(Calendar.SECOND, reqValidity);
		confData.setNotOnOrAfter(validity);
		authenticatedOne.setSubjectConfirmation(
				new SubjectConfirmationType[] {subConf});
				
		AssertionResponse resp = samlUtil.getOKResponseDocument();
		Assertion assertion = new Assertion();
		assertion.setIssuer(samlUtil.getIssuer().getXBean());
		assertion.setSubject(authenticatedOne.getXBean());
		assertion.addAuthStatement(authTime, authContext);
		assertion.setAudienceRestriction(new String[] {samlUtil.getIssuerName()});

		try
		{
			assertion.sign(conf.getAuthAndTrust().getCredential().getKey(), 
					conf.getAuthAndTrust().getCredential().getCertificateChain());
		} catch (DSigException e)
		{
			log.error("Signing SAML response error", e);
			throw new UVOSFault(new InternalException(
				"Signing response problem", e));
		}

		resp.addAssertion(assertion);
		return resp.getDoc();
	}
	
	private Subject findUser(String format, List<Identity> equivs) 
		throws SAMLProtocolException
	{
		if (format.equals(SAMLConstants.NFORMAT_PERSISTENT) || 
				format.equals(SAMLConstants.NFORMAT_UNSPEC))
		{
			String persId = equivs.get(0).getPersistentEntityId();
			return new Subject(persId, SAMLConstants.NFORMAT_PERSISTENT);
		} else if (format.equals(SAMLConstants.NFORMAT_DN) ||
				format.equals(SAMLConstants.NFORMAT_EMAIL) ||
				format.equals(IdentityType.X509CERT.getName()))
		{
			for (Identity i: equivs)
				if (i.getType().getName().equals(format))
					return new Subject(i.getValue(), format);
			//there was DN-type login requested, there is no such DN id
			//and cert to dn mapping is enabled - so let's try to find cert. 
			if (allowX509AsDN && format.equals(IdentityType.DN))
				for (Identity i: equivs)
					if (i.getType().equals(IdentityType.X509CERT))
					{
						X509Certificate cert;
						try
						{
							cert = new X509IdentityHelper(i).getCertificate();
						} catch (InvalidValueException e)
						{
							log.error("BUG - there is an unparsable " +
									"certificate in DB: " + i.getValue());
							continue;
						}
						String cn = cert.getSubjectX500Principal().getName();
						return new Subject(cn, format);
					}
			throw new SAMLRequestException(
					SAMLConstants.STATUS2_UNKNOWN_PRINCIPIAL,
					"There is no identity in the requested " +
					"format for the authenticated principial.");			
		} else
		{
			throw new SAMLRequestException(
					SAMLConstants.STATUS2_INVALID_NAMEID_POLICY,
					format + " is not supported by this service.");
		}
	}
	
	/**
	 * Note that elements "Scoping", "ForceAuthn", "IsPassive" and "Conditions",
	 * "ProtocolBinding" are ignored. Either are processed at higher level 
	 * (e.g. by web servlet doing the actual authn) or not important. 
	 * @throws SAMLProtocolException 
	 *
	 */
	private void checkUnsupportedConstraints(AuthnRequestDocument reqDoc) 
		throws SAMLProtocolException
	{
		AuthnRequestType req = reqDoc.getAuthnRequest();
		//1 - presence of Subject element in request
		//WARNING - if this is implemented then filtering of inactive identities
		//must be also applied in this class.
		if (req.getSubject() != null)
			throw new SAMLProtocolException(SAMLConstants.STATUS_RESPONDER,
				null, "This implementation doesn't support authn " +
						"requests with Subject set.");
		//2 - requestedAuthnContext
		if (req.getRequestedAuthnContext() != null)
			throw new SAMLProtocolException(SAMLConstants.STATUS_RESPONDER,
				null, "This implementation doesn't support authn " +
						"requests with RequestedAuthnContext set.");
		//3 - assertionConsumerServiceIndex
		if (req.isSetAssertionConsumerServiceIndex())
			throw new SAMLProtocolException(SAMLConstants.STATUS_RESPONDER,
				null, "This implementation doesn't support authn " +
						"requests with AssertionConsumerServiceIndex set.");
		//4 - AttributeconsumingServiceIndex
		if (req.isSetAttributeConsumingServiceIndex())
			throw new SAMLProtocolException(SAMLConstants.STATUS_RESPONDER,
				null, "This implementation doesn't support authn " +
						"requests with AttributeConsumingServiceIndex set.");
	}
}
