/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Aug 7, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.server.saml;

import java.util.ArrayList;
import java.util.List;

import javax.jws.WebService;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.XmlCursor;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;

import eu.unicore.samly2.SAMLConstants;
import eu.unicore.samly2.assertion.Assertion;
import eu.unicore.samly2.exceptions.SAMLParseException;
import eu.unicore.samly2.exceptions.SAMLProtocolException;
import eu.unicore.samly2.exceptions.SAMLRequestException;
import eu.unicore.samly2.proto.AssertionResponse;
import eu.unicore.security.dsig.DSigException;
import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.exceptions.AuthenticationException;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementNotKnownException;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.api.exceptions.InvalidValueException;
import pl.edu.icm.unicore.uvos.api.exceptions.SecurityException;
import pl.edu.icm.unicore.uvos.api.exceptions.UVOSException;
import pl.edu.icm.unicore.uvos.engine.InternalAuthorization;
import pl.edu.icm.unicore.uvos.engine.api.UVOSQueryInterface;
import pl.edu.icm.unicore.uvos.saml.SAMLAttributeQueryUtil;
import pl.edu.icm.unicore.uvos.saml.SAMLConfiguration;
import pl.edu.icm.unicore.uvos.saml.SAMLVOConstants;
import pl.edu.icm.unicore.uvos.server.IssuerIdentityProvider;
import pl.edu.icm.unicore.uvos.server.idresolver.IdResolveException;
import pl.edu.icm.unicore.uvos.server.idresolver.Authenticator;
import pl.edu.icm.unicore.uvos.util.AttributeFilter;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.SAMLQueryInterface;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.SAMLXMLBeansMapper;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.UVOSFault;
import pl.edu.icm.unicore.uvos.wsapi.xmlbeans.UnsupportedSAMLException;
import xmlbeans.org.oasis.saml2.assertion.AttributeType;
import xmlbeans.org.oasis.saml2.protocol.AttributeQueryDocument;
import xmlbeans.org.oasis.saml2.protocol.ResponseDocument;

/**
 * @author K. Benedyczak
 */
@WebService(endpointInterface="pl.edu.icm.unicore.uvos.wsapi.SAMLQueryInterface")
public class SAMLQueryImpl implements SAMLQueryInterface
{
	private UVOSQueryInterface voQuery;
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, SAMLQueryImpl.class);
	private int reqValidity;
	private UVOSServerProperties conf;
	private IssuerIdentityProvider cryptoProvider;
	private Authenticator authenticator;
	private InternalAuthorization authz;
	private SAMLX509QueryProfile samlQueryProfile;
	private AttributeFilter attribFilter;
	private SAMLConfiguration samlConfiguration;
	
	public SAMLQueryImpl(UVOSQueryInterface voQuery, UVOSServerProperties conf, IssuerIdentityProvider cryptoProvider,
			Authenticator authenticator, InternalAuthorization authz, SAMLX509QueryProfile samlQueryProfile,
			AttributeFilter attribFilter, SAMLConfiguration samlConfig)
	{
		this.samlConfiguration=samlConfig;
		this.voQuery = voQuery;
		this.conf = conf;
		this.cryptoProvider = cryptoProvider;
		this.authenticator = authenticator;
		this.authz = authz;
		this.samlQueryProfile = samlQueryProfile;
		this.attribFilter = attribFilter;
		reqValidity = conf.getIntValue(UVOSServerProperties.SAML_REQUEST_VALIDITY);
	}
	
	/**
	 * Implementation of SAML mapping onto internal operations.
	 * <p>
	 * Basic rules are:
	 * <ul>
	 * <li> Subject of SAML query is mapped on UVOS Element: Identity or 
	 * Identity in a group scope (later case if Subject's NameQualifier is set 
	 * and contains group).
	 * <li> If there is no attributes in query and Element is id in group scope then the 
	 * request is equal to getAllAttributes of {@link UVOSQueryInterfacege}
	 * <li> If there is no attributes in query and Element is id without group scope 
	 * then the request is equal to 
	 * {@link UVOSQueryInterface#getAllAttributes(Element, boolean, SecurityContext)}
	 * and 
	 * {@link UVOSQueryInterface#getAllGroups(pl.edu.icm.unicore.uvos.api.Identity)}
	 * <li> If there is set of attributes specified then it is translated to multiple
	 * getAttribute invocations (plus extra results filtering if values are given in
	 * the query).
	 * <li> The exception to above rules is  asking for attribute with name 
	 * {@link SAMLConstants#GROUP_ATTR} which is translated into getAllGroups
	 * (with optional filtering of results if values are given in the query). 
	 * This make sense only for elements NOT in a group scope.
	 * </ul> 
	 * @throws IdResolveException 
	 * @throws InternalException 
	 */
	public ResponseDocument attributeQuery(AttributeQueryDocument reqDoc) 
		throws UVOSFault
	{
		if (log.isTraceEnabled())
			log.trace("Processing request:\n" + reqDoc.xmlText(
				new XmlOptions().setSavePrettyPrint()));
		else if (log.isDebugEnabled())
			log.debug("Processing SAML attribute query");

		SAMLAttributeQueryUtil samlUtil = new SAMLAttributeQueryUtil(reqDoc, cryptoProvider, conf, samlConfiguration);

		try
		{
			authenticator.resolve();
			samlUtil.parse();
			samlUtil.checkRequestValidity(reqValidity);
		} catch (IdResolveException e)
		{
			log.debug("UVOS engine problem while processing " +
					"SAML request (authN stage )", e);
			throw new UVOSFault(e);
		} catch (UVOSException e)
		{
			log.debug("Authentication FAILED: " + e);
			String info = "Authentication failed: " + ((e.getMessage() == null) ? 
					e.getClass().getSimpleName() : e.getMessage());
			return samlUtil.getErrorResponse(new AuthenticationException(), 
					info).getDoc();
		} catch (SAMLProtocolException e)
		{
			log.debug("Parsing the SAML request FAILED: " + e);
			return samlUtil.getErrorResponse(e).getDoc();
		} 
		if (log.isDebugEnabled())
			log.debug("Parsed SAML request. Issuer: " + 
			samlUtil.getRequestIssuerRaw().getStringValue() + 
			"\nSubject: " + samlUtil.getSubject().getXBean().getNameID().getStringValue());
		
		AssertionResponse resp = samlUtil.getOKResponseDocument();
		Assertion assertion = new Assertion();
		assertion.setIssuer(samlUtil.getIssuer().getXBean());
		assertion.setSubject(samlUtil.getSubject().getXBean());
		
		try
		{
			samlQueryProfile.handleX509Profile(samlUtil, 
					assertion);
		} catch (SAMLProtocolException e)
		{
			log.debug("Error while applying SAML query " +
					"profile for X.509 subjects: " + e);
			return samlUtil.getErrorResponse(e).getDoc();
		}
		
		
		Element subElement = samlUtil.getSubjectElement();
		
		try
		{
			Identity effectiveId = authz.mapIfNeededToX509(subElement.getIdentity());
			subElement.setIdentity(effectiveId);
		} catch (UVOSException e2)
		{
			log.debug("Query subject is unknown: " + e2);
			return samlUtil.getErrorResponse(e2).getDoc();
		}
		
		try
		{
			if (!authz.isActive(subElement.getIdentity()))
			{
				log.debug("Denying request about disabled entity " + 
					subElement.getIdentity());
				SAMLRequestException e = new SAMLRequestException(
						SAMLConstants.STATUS2_UNKNOWN_PRINCIPIAL,
						"Query subject is unknown");
				return samlUtil.getErrorResponse(e).getDoc();
			}
		} catch (UVOSException e1)
		{
			log.debug("Query subject is unknown: " + e1);
			return samlUtil.getErrorResponse(e1).getDoc();
		}
		
		
		AttributeType []queriedAttrs = samlUtil.getQueriedAttributes();

		AttributeConverter attrsConverter = new AttributeConverter();
		
		try
		{
			if (queriedAttrs == null || queriedAttrs.length == 0)
			{
				//-> get all subject's attributes
				getAllAttributes(subElement, attrsConverter);
			} else
			{
				for (AttributeType attrToGet: queriedAttrs)
					getAttribute(subElement, 
						attrToGet, attrsConverter);
			}
			attrsConverter.injectAttributes(assertion);
		} catch (UVOSException e)
		{
			log.debug("Error in processing query: " + e);
			return samlUtil.getErrorResponse(e).getDoc();
		} catch (SAMLRequestException e)
		{
			log.debug("SAML request processing failed while parsing: " +
					e);
			return samlUtil.getErrorResponse(e).getDoc();
		} catch (UnsupportedSAMLException e)
		{
			log.debug("SAML request processing failed as query is unsupported: " +
					e);
			SAMLRequestException ex = new SAMLRequestException(
					SAMLConstants.STATUS2_REQUEST_UNSUPP,
					e.getMessage());
			return samlUtil.getErrorResponse(ex).getDoc();
		} catch (SAMLParseException e)
		{
			log.debug("SAML request processing failed while parsing: " +
					e);
			SAMLRequestException ex = new SAMLRequestException(null,
					e.getMessage());
			return samlUtil.getErrorResponse(ex).getDoc();
		}
		
		boolean insertA = false;
		if (assertion.getAttributes().length != 0)
			insertA = true;

		if (insertA && samlUtil.doSignAssertion())
			try
			{
				assertion.sign(conf.getAuthAndTrust().getCredential().getKey(), 
						conf.getAuthAndTrust().getCredential().getCertificateChain());
			} catch (DSigException e)
			{
				log.error("Signing SAML response error", e);
				throw new UVOSFault(new InternalException(
					"Signing response problem", e));
			}
		if (insertA)
			resp.addAssertion(assertion);
		
		if (samlUtil.doSignResponse())
			try
			{
				resp.sign(conf.getAuthAndTrust().getCredential().getKey(), 
						conf.getAuthAndTrust().getCredential().getCertificateChain());
			} catch (DSigException e)
			{
				log.error("Signing SAML assertion error", e);
				throw new UVOSFault(new InternalException(
					"Signing response assertion problem", e));
			}
		if (log.isTraceEnabled())
			log.trace("Returning response: " + resp.getDoc().xmlText(
				new XmlOptions().setSavePrettyPrint()));
		return resp.getDoc();
	}

	
	/**
	 * Gets all attributes of element. If it isn't id@group also all groups are
	 * encoded as attributes with special name. If it is id@group then scoped attributes are added
	 * and those groups (as values of SAMLVOConstants.GROUP_ATTR attribute) which are subgroups
	 * of the scope (including it).  
	 * @param whose
	 * @param ctx
	 * @param as
	 * @throws UVOSException
	 * @throws SAMLRequestException 
	 */
	private void getAllAttributes(Element whose, AttributeConverter converter) 
		throws UVOSException, SAMLRequestException
	{
		List<Attribute> attrs = 
			voQuery.getAttributes(whose, null, true, true, true);
		
		attrs = attribFilter.filter(attrs);

		converter.preInjectAttributes(attrs, null);

		List<Group> groups = 
			voQuery.getAllGroups(whose.getIdentity(), true);
		if (whose.getGroup() == null)
		{
			if (attrs.size() == 0 && groups.size() == 0)
				return;
			converter.preInjectGroups(groups);
		} else
		{
			Group scope = whose.getGroup();
			String scopeStr = scope.toString();
			for (int i=groups.size()-1; i>=0; i--)
			{
				Group g = groups.get(i);
				if (!(g.isChild(scope) || g.toString().equals(scopeStr)))
					groups.remove(i);
			}
			if (attrs.size() == 0 && groups.size() == 0)
				return;
			converter.preInjectGroups(groups);
		}
	}
	
	
	/**
	 * Gets named attribute and injects it to the response assertion. 
	 * It recognizes special 'group' attribute, does filtering of results
	 * if values are given in query and encodes results to the correct XACML type 
	 * if needed. 
	 * @param whose
	 * @param attrQ
	 * @param ctx
	 * @param as
	 * @throws SAMLRequestException
	 * @throws UVOSException
	 * @throws SAMLParseException 
	 * @throws UnsupportedSAMLException 
	 */
	private void getAttribute(Element whose, AttributeType attrQ, AttributeConverter converter) 
		throws SAMLRequestException, UVOSException, UnsupportedSAMLException,
		SAMLParseException
	{
		String attrName = attrQ.getName();
		XmlCursor cur = attrQ.newCursor();
		String requestedAttributeWideScope = cur.getAttributeText(
				SAMLConstants.ATTRIBUTE_SCOPE_XMLATTRIBUTE);
		String requestedScopingType = cur.getAttributeText(
				SAMLConstants.SCOPE_TYPE_XMLATTRIBUTE);
		cur.dispose();
		
		if (attrName == null)
			throw new SAMLRequestException(null, 
					"Attribute name can't be empty");
		//get all with name (special handling: groups)
		List<Attribute> attributes = getNamedAttribute(whose, 
				attrName);
		
		//now filter results
		filterAttributes(attributes, attrQ, requestedScopingType, 
				requestedAttributeWideScope);
		
		//finally encode
		converter.preInjectAttributes(attributes, requestedScopingType);
	}
	
	private List<Attribute> getNamedAttribute(Element whose, String name) 
		throws SecurityException, InternalException, 
		InvalidValueException, ElementNotKnownException
	{
		if (name.equals(SAMLVOConstants.GROUP_ATTR))
		{
			List<Group> groups = voQuery.getAllGroups(whose.getIdentity(), 
					true); 
			ArrayList<String> groupNames = new ArrayList<String>();
			for (Group g: groups)
				groupNames.add(g.toString());
			ArrayList<Attribute> ret = new ArrayList<Attribute>();
			ret.add(new Attribute(AttributeConverter.GROUP_AT, groupNames, null));
			return ret;
		}
		return voQuery.getAttributes(whose, name, true, true, 
				true);
	}
	
	private void filterAttributes(List<Attribute> unfiltered, 
			AttributeType attrQ, String requestedScoping, String requestedScope)
		throws UnsupportedSAMLException, SAMLParseException
	{
		XmlObject[] valuesSpec = attrQ.getAttributeValueArray();
		if (valuesSpec == null || valuesSpec.length == 0 || unfiltered.size() == 0)
		{
			filterAttributesByScopeOnly(unfiltered, 
					attrQ, requestedScoping, requestedScope);
			return;
		}
		//here is filtering by scope and value
		List<Attribute> converted = SAMLXMLBeansMapper.map2APIAttributes(attrQ);
		for (int i=0; i<unfiltered.size(); i++)
		{
			Attribute checked = unfiltered.get(i);
			boolean found = false;
			for (Attribute okAttr: converted)
			{
				if (!areOfSameScope(okAttr, checked, requestedScoping))
					continue;
				//here checked and okAttr have the same name and scope
				//so the only difference can be in values. Let's filter those
				if (filterAttributeValues(checked, okAttr))
					found = true;
				break;
			}
			if (!found)
			{
				unfiltered.remove(i);
				i--;
			}
		}
	}
	
	private void filterAttributesByScopeOnly(List<Attribute> unfiltered, 
			AttributeType attrQ, String requestedScoping, String requestedScope)
		throws UnsupportedSAMLException, SAMLParseException
	{
		if (requestedScope == null)
			return;
		List<Attribute> converted = SAMLXMLBeansMapper.map2APIAttributes(attrQ);
		for (int i=0; i<unfiltered.size(); i++)
		{
			Attribute checked = unfiltered.get(i);
			boolean found = false;
			for (Attribute okAttr: converted)
			{
				if (!areOfSameScope(okAttr, checked, requestedScoping))
					continue;
				found = true;
				break;
			}
			if (!found)
			{
				unfiltered.remove(i);
				i--;
			}
		}
	}
	
	/**
	 * 
	 * @param a attribute to have its values filtered.
	 * @param filter attribute with values filter. If there are no values, then
	 * there is no filtering - all original values are returned.
	 * @return true if the attribute should be removed completely as
	 * filtering of values removed all.
	 */
	private boolean filterAttributeValues(Attribute a, Attribute filter)
	{
		List<String> aVals = a.getValues();
		List<String> filterVals = filter.getValues();
		//no values in filter - meaning we got either an empty attribute in query
		//(=> we should return all values) or some nil attributes which specified
		//the scope for a value, but doesn't mean that client is interested only in 
		//nil value (yup - this is SAML violation).
		if (filterVals.size() == 0)
			return true;
		
		//we got some vals in query so let's filter
		for (int i=0; i<aVals.size(); i++)
		{
			if (!filterVals.contains(aVals.get(i)))
			{
				aVals.remove(i);
				i--;
			}
		}
		//return false when all values were filtered out
		//so the attribute should be deleted completely.
		return !(aVals.size() == 0);
	}
	
	private boolean areOfSameScope(Attribute base, Attribute a2, String requestedScoping)
	{
		if (requestedScoping == null)
		{
			//here if base attribute has no scope every other scope shall match. 
			if (base.getScope() == null)
				return true;
		}
		String bSc = base.getScope();
		String a2Sc = a2.getScope();
		if (bSc == null)
			bSc = "/";
		if (a2Sc == null)
			a2Sc = "/";
		return bSc.equals(a2Sc);
	}
}









