/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Jan 12, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.server.saml;

import java.net.URI;
import java.net.URISyntaxException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Date;

import org.apache.log4j.Logger;

import eu.emi.security.authn.x509.impl.X500NameUtils;
import eu.unicore.samly2.SAMLConstants;
import eu.unicore.samly2.assertion.Assertion;
import eu.unicore.samly2.exceptions.SAMLProtocolException;
import eu.unicore.samly2.exceptions.SAMLRequestException;
import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.saml.SAMLAttributeQueryUtil;
import pl.edu.icm.unicore.uvos.server.idresolver.Authenticator;
import pl.edu.icm.unicore.uvos.util.UVOSServerProperties;
import xmlbeans.org.oasis.saml2.assertion.NameIDType;

/**
 * Implementation of SAML logic necessary to be conformant to 
 * "SAML V2.0 Deployment Profiles for X.509 Subjects"
 *
 * @author K. Benedyczak
 */
public class SAMLX509QueryProfile
{
	private enum Mode {NO_X509_PROFILE, X509_QUERY_PROFILE, X509_SELF_QUERY_PROFILE};
	
	private long defValidity;
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, SAMLX509QueryProfile.class);
	
	public SAMLX509QueryProfile(UVOSServerProperties conf)
	{
		defValidity = conf.getIntValue(UVOSServerProperties.DEF_ATTR_ASSERTION_VALIDITY);
		defValidity *= 1000;
	}
	
	public void handleX509Profile(SAMLAttributeQueryUtil samlUtil, Assertion assertion) 
		throws SAMLProtocolException
	{
		Mode mode = checkX509Profile(samlUtil);
		log.debug("Request will be processed in accordance to the profile: " + mode);
		setTimeConditions(mode, assertion);
		setAudienceRestriciton(mode, samlUtil, assertion);
		setSubjectConfirmation(mode, assertion);
		if (mode == Mode.X509_SELF_QUERY_PROFILE)
			samlUtil.setEnforceSignAssertion(true);
	}
	
	/**
	 * The rules are:
	 *  - for (normal) X509_PROFILE - issuer with format null|entity|unspec AND
	 *  issuer value is a proper URI AND subject is DN
	 *  - for X509_SELF_QUERY - issuer with format DN AND subject is DN AND issuer
	 *  DN == subject's DN AND authenticated user's DN == subject's DN. 
	 * @param samlUtil
	 * @return
	 * @throws SAMLRequestException 
	 */
	private Mode checkX509Profile(SAMLAttributeQueryUtil samlUtil) 
		throws SAMLRequestException
	{
		String subjectDN, issuerDN, authedDN;
		X509Certificate certFromAuthN = Authenticator.getAuthNCert();
		if (certFromAuthN == null)
			return Mode.NO_X509_PROFILE;
		authedDN = certFromAuthN.getSubjectX500Principal().getName();
		Identity subject = samlUtil.getSubjectIdentity();
		if (subject.getType().equals(IdentityType.DN))
		{
			subjectDN = subject.getValue();
		} else
			return Mode.NO_X509_PROFILE;
		
		//issuer identity must be URI of one of two formats if 
		//X509_PROFILE has to be used
		
		NameIDType issuer = samlUtil.getRequestIssuerRaw();
		
		String issuerFormat = issuer.getFormat();
		boolean isDNIssuer = false;
		boolean isSystemEntityIssuer = false;
		if (issuerFormat == null 
			|| issuerFormat.equals(SAMLConstants.NFORMAT_ENTITY) 
			|| issuerFormat.equals(SAMLConstants.NFORMAT_UNSPEC))
		{
			try
			{
				new URI(issuer.getStringValue());
				isSystemEntityIssuer = true;
			} catch (URISyntaxException e){}
		} else if (issuerFormat.equals(SAMLConstants.NFORMAT_DN))
		{
			isDNIssuer = true;
		}
		
		if (isSystemEntityIssuer)
			return Mode.X509_QUERY_PROFILE;
		if (!isDNIssuer)
			return Mode.NO_X509_PROFILE;
		
		issuerDN = issuer.getStringValue();

		if (safeEqualsDN(issuerDN, subjectDN) && safeEqualsDN(issuerDN, authedDN))
			return Mode.X509_SELF_QUERY_PROFILE;

		
		return Mode.NO_X509_PROFILE;
	}
	
	private boolean safeEqualsDN(String dn1, String dn2) throws SAMLRequestException
	{
		try
		{
			return X500NameUtils.equal(dn1, dn2);
		} catch (IllegalArgumentException e)
		{
			throw new SAMLRequestException(
				SAMLConstants.STATUS2_UNKNOWN_PRINCIPIAL, e.getMessage());
		}
	}
	
	private void setTimeConditions(Mode mode, Assertion assertion) 
		throws SAMLProtocolException
	{
		if (mode == Mode.NO_X509_PROFILE)
			return;
		Date start, end;
		X509Certificate certFromAuthN = Authenticator.getAuthNCert();
		start = new Date();
		try
		{
			certFromAuthN.checkValidity(start);
		} catch (CertificateException e)
		{
			throw new SAMLProtocolException(SAMLConstants.STATUS_REQUESTER, 
					SAMLConstants.STATUS2_REQUEST_DENIED, 
					"The issuer's certificate is expired");
		}
		Date endCert = certFromAuthN.getNotAfter();
		end = new Date(defValidity + System.currentTimeMillis());
		if (end.after(endCert))
			end = endCert;
		
		assertion.setTimeConditions(start, end);
	}
	
	private void setAudienceRestriciton(Mode mode, SAMLAttributeQueryUtil samlUtil, 
			Assertion assertion)
	{
		if (mode != Mode.X509_QUERY_PROFILE)
			return;
		NameIDType issuer = samlUtil.getRequestIssuerRaw();

		assertion.setAudienceRestriction(new String[] {issuer.getStringValue()});
	}
	
	private void setSubjectConfirmation(Mode mode, Assertion assertion) 
		throws SAMLProtocolException
	{
		if (mode == Mode.NO_X509_PROFILE || mode == Mode.X509_QUERY_PROFILE)
			return;
		X509Certificate certFromAuthN = Authenticator.getAuthNCert();
		
		try
		{
			assertion.setHolderOfKeyConfirmation(
					new X509Certificate[] {certFromAuthN});
		} catch (CertificateEncodingException e)
		{
			throw new SAMLProtocolException(SAMLConstants.STATUS_RESPONDER,
					null, "Problem occured when encoding " +
						"issuer's certificate back into assertion");
		}
	}
}
