/*
 * Copyright (c) 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Jul 29, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;

import eu.unicore.util.Log;

import pl.edu.icm.unicore.uvos.api.AttributeType;
import pl.edu.icm.unicore.uvos.api.exceptions.InternalException;
import pl.edu.icm.unicore.uvos.db.DB;
import pl.edu.icm.unicore.uvos.db.DBAttributeTypes;

/**
 * Loads attribute types from configured files and update DB accordingly.
 * @author golbi
 */
public class AttributeTypesLoader
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, 
			AttributeTypesLoader.class);
	private List<String> files;
	private boolean doUpdate;
	
	private List<AttributeType> parsedTypes;
	private DBAttributeTypes attributeTypes;
	private DB db;
	
	public AttributeTypesLoader(DB db, UVOSServerProperties conf, DBAttributeTypes attributeTypes)
	{
		this.db = db;
		this.attributeTypes = attributeTypes;
		doUpdate = conf.getBooleanValue(UVOSServerProperties.ATTRIBUTE_TYPES_UPDATE);
		String filesRaw = conf.getValue(UVOSServerProperties.ATTRIBUTE_TYPE_FILES);
		String []filesA = filesRaw.split(",( )+");
		StringBuffer filesS = new StringBuffer();
		boolean parse = false;
		files = new ArrayList<String>();
		for (String f: filesA)
		{
			if (filesS.length() > 0)
				filesS.append(", ");
			if (!f.equals(""))
			{
				files.add(f);
				filesS.append(f);
				parse = true;
			}
		}
		if (parse)
		{
			log.info("Will read attribute types from files: " + 
				filesS + ". Existing attribute" +
				" types will be updated: " + doUpdate);
		}
		parsedTypes = new ArrayList<AttributeType>();
	}
	
	public void load()
	{
		for (String file: files)
		{
			File f = new File(file);
			if (f.exists() && f.canRead())
				try
				{
					processFile(f);
				} catch (IOException e)
				{
					log.warn("IO error when processing " +
						"an attribute types file: " + 
						file + ": " + e);
				}
			else
				log.warn("The attribute type file " + 
					file + " doesn't exists or can't be read.");
		}
		addToDB();
	}
	
	private void processFile(File file) throws IOException
	{
		BufferedReader reader = new BufferedReader(new FileReader(file));
		String line = reader.readLine();
		while ((line = reader.readLine()) != null)
		{
			if (line.startsWith("#"))
				continue;
			if (line.equals("AT"))
				processBlock(reader);
			if (line.equals(""))
				continue;
			
		}
	}
	
	private void processBlock(BufferedReader reader) throws IOException
	{
		String atName = reader.readLine();
		String atSDesc = reader.readLine();
		String atDesc = reader.readLine();
		if (atName == null || atSDesc == null || atDesc == null)
			throw new IOException("Unexpected end of file: AT is incomplete.");
		atDesc = atDesc.replace("\\n", "\n");
		AttributeType at = new AttributeType(atName, atSDesc, atDesc);
		parsedTypes.add(at);
	}
	
	private void addToDB()
	{
		List<AttributeType> existing = attributeTypes.getAttributeTypes();
		Set<String> existingSet = new HashSet<String>();
		for (AttributeType at: existing)
			existingSet.add(at.getName());
		
		SqlSession session = db.getSqlSession();
		try
		{
			for (AttributeType at : parsedTypes)
			{
				boolean exists = existingSet.contains(at.getName());
				if (exists && !doUpdate)
					continue;
				try
				{
					if (exists)
						attributeTypes.updateAttrType(at, session);
					else
						attributeTypes.addAttrType(at, session);
				} catch (InternalException e)
				{
					log.warn("Can't add/update attribute type "
							+ at + " reason: " + e);
				}
			}
		} finally
		{
			session.commit();
			db.releaseSqlSession(session);
		}
	}
}













