/*
 * Copyright (c) 2012 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE.txt file for licensing information.
 */
package pl.edu.icm.unicore.uvos.util;

import java.io.File;
import java.io.IOException;

import org.apache.commons.io.FileUtils;
import org.apache.log4j.Logger;
import org.codehaus.jettison.json.JSONArray;
import org.codehaus.jettison.json.JSONException;
import org.codehaus.jettison.json.JSONObject;

import eu.emi.emir.client.ServiceBasicAttributeNames;
import eu.unicore.util.Log;

import xmlbeans.pl.edu.icm.uvos.glue2.EndpointHealthStateT;
import xmlbeans.pl.edu.icm.uvos.glue2.EndpointT;
import xmlbeans.pl.edu.icm.uvos.glue2.QualityLevelT;
import xmlbeans.pl.edu.icm.uvos.glue2.ServiceT;
import xmlbeans.pl.edu.icm.uvos.glue2.ServingStateT;

/**
 * Provides server information in Glue format. It is used by the glue info service and 
 * also as an input to EMIR.
 * @author K. Benedyczak
 */
public class GLUEInfoProvider
{
	private static final Logger log = Log.getLogger(UVOSServerProperties.LOG_PFX, GLUEInfoProvider.class);
	public static final String UVOS_HUMAN_NAME = "UNICORE Virtual Organisations Service (UVOS)";
	public static final String UVOS_SERVICE_TYPE = "unicore6.UVOSAssertionQueryService";
	public static final String EMI_VERSION_FILE = "/etc/emi-version";
	
	public static final String ASSERTION_QUERY_ENDPOINT = "/UVOSAssertionQueryService";

	private String baseURL;

	public GLUEInfoProvider(UVOSServerProperties conf)
	{
		boolean ssl = conf.getBooleanValue(UVOSServerProperties.HTTPS_ENABLE);
		if (ssl) 
		{
			baseURL = "https://" + conf.getValue(UVOSServerProperties.HTTPS_HOST) + ":" +
					conf.getValue(UVOSServerProperties.HTTPS_PORT);
		} else
		{
			baseURL = "http://" + conf.getValue(UVOSServerProperties.HTTP_HOST) + ":" +
					conf.getValue(UVOSServerProperties.HTTP_PORT);
		}
	}

	public GLUEInfoProvider(String baseUrl)
	{
		this.baseURL = baseUrl;
	}
	
	public ServiceT getServiceDesc()
	{
		ServiceT st = ServiceT.Factory.newInstance();
		st.setID(baseURL);
		st.setName(UVOS_HUMAN_NAME);
		st.setQualityLevel(QualityLevelT.PRODUCTION);
		st.setType(UVOS_SERVICE_TYPE);
		st.setEndpointArray(new EndpointT[] {getServiceendpointDesc()});
		return st;
	}

	public String getServiceId()
	{
		return baseURL+ASSERTION_QUERY_ENDPOINT;
	}
	
	
	//when changing also change conversion to JSON below
	private EndpointT getServiceendpointDesc()
	{
		EndpointT et = EndpointT.Factory.newInstance();
		et.setID(getServiceId());
		et.setURL(baseURL+ASSERTION_QUERY_ENDPOINT);
		et.setQualityLevel(QualityLevelT.PRODUCTION);
		et.setHealthState(EndpointHealthStateT.OK);
		et.setServingState(ServingStateT.PRODUCTION);
		
		String uVersion=getClass().getPackage().getSpecificationVersion();
		if (uVersion == null)
			uVersion = "DEVELOPMENT";
		et.addInterfaceVersion(uVersion);
		et.setImplementationName("UVOS");
		et.setImplementationVersion(uVersion);
		et.setInterfaceName("UVOS");

		et.setCapabilityArray(new String[] {"security.attributeauthority"});
		et.setTechnology("webservice");

		//add EMI Version, if present
		File emiFile = new File(EMI_VERSION_FILE);
		if (emiFile.exists() && emiFile.canRead())
		{
			try 
			{
				String emiVersion = FileUtils.readFileToString(emiFile);
				et.addOtherInfo("MiddlewareName=EMI");
				et.addOtherInfo("MiddlewareVersion=" + emiVersion);
			} catch (IOException ex)
			{
				log.warn("Can't read emi version.", ex);
			}
		}
		return et;
	}
	
	public JSONArray getEndpointDescription() throws JSONException 
	{
		ServiceT glue = getServiceDesc();
		JSONObject serdelek = new JSONObject();
		
		serdelek.put(ServiceBasicAttributeNames.SERVICE_ID.toString(), glue.getID());
		serdelek.put(ServiceBasicAttributeNames.SERVICE_NAME.toString(), glue.getName());
		serdelek.put(ServiceBasicAttributeNames.SERVICE_TYPE.toString(), glue.getType());
		serdelek.put(ServiceBasicAttributeNames.SERVICE_QUALITYLEVEL.toString(), 
				glue.getQualityLevel().toString());

		EndpointT[] endpoints = glue.getEndpointArray();
		for (EndpointT endpoint: endpoints)
		{
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_ID.toString(), endpoint.getID());
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_URL.toString(), endpoint.getURL());
			JSONArray endpCap = new JSONArray();
			for (String cap: endpoint.getCapabilityArray())
				endpCap.put(cap);
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_CAPABILITY.toString(), endpCap);

			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_TECHNOLOGY.toString(), 
					endpoint.getTechnology());
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_IFACENAME.toString(), 
					endpoint.getInterfaceName());
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_IFACE_VER.toString(),
					endpoint.getInterfaceVersionArray()[0]);
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_IMPL_NAME.toString(), 
					endpoint.getImplementationName());
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_IMPL_VERSION.toString(),
					endpoint.getImplementationVersion());
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_HEALTH_STATE.toString(),
					endpoint.getHealthState());
			serdelek.put(ServiceBasicAttributeNames.SERVICE_ENDPOINT_SERVING_STATE.toString(),
					endpoint.getServingState());
		}
		JSONArray array = new JSONArray();
		array.put(serdelek);
		return array;
	}

}
