/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Apr 6, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.stress;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import static junit.framework.Assert.*;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.db.InitDB;
import pl.edu.icm.unicore.uvos.engine.VirtualSecurityContextImpl;
import pl.edu.icm.unicore.uvos.engine.api.UVOSManagementInterface;
import pl.edu.icm.unicore.uvos.engine.api.UVOSQueryInterface;
import pl.edu.icm.unicore.uvos.server.ComponentContainer;
import pl.edu.icm.unicore.uvos.server.idresolver.Authenticator;

/**
 * @author K. Benedyczak
 */
public class PerformanceTest
{
	private UVOSManagementInterface man;
	private UVOSQueryInterface query;
	
	@Before
	public void setUp() throws Exception
	{
		try
		{
			InitDB.main(new String[]{"src/test/resources/uvosServer.conf", "-b"});
			ComponentContainer container = new ComponentContainer("src/test/resources/uvosServer.conf");
			man = container.getComponent(UVOSManagementInterface.class);
			query = container.getComponent(UVOSQueryInterface.class);
			Authenticator.injectFakeContext(new VirtualSecurityContextImpl(Permissions.ALL_P));
		} catch (Throwable e)
		{
			System.err.println("Severe error when starting UVOS Web Server:");
			e.printStackTrace();
			fail();
		}
	}

	private long startT;

	private void startTimer()
	{
		startT = System.currentTimeMillis();
	}
	
	private void stopTimer(int ops, String label, int acceptableOpsPS)
	{
		long endT = System.currentTimeMillis();
		long periodMs = endT - startT;
		double periodS = periodMs/1000.0;
		double opsPerS = (ops*1000/periodMs);
		System.out.println(label + " performed " + ops + 
				" in " + periodS + "s, " + 
				opsPerS + " ops/s");
		assertTrue("Too slow!, should be at least " + acceptableOpsPS + 
				" was " + opsPerS, opsPerS >= acceptableOpsPS);
	}
	
	private static int LIMIT_LOW = 10;
	private static int LIMIT_NORMAL = 20;
	private static int LIMIT_HIGH = 40;
	
	@Test
	public void test1()
	{
		try
		{
			Identity toAdd = new Identity(IdentityType.DN, 
					"cn=ala,c=pl,ou=icm,o=WarsawUniv,l=loc0");
			System.out.println("\n\nStarting performance metering...\n\n");

			//******CREATION*************
			
			int N = 500;
			startTimer();
			for (int i=0; i<N; i++)
			{
				toAdd = new Identity(IdentityType.DN, 
					"cn=ala,c=pl,ou=icm,o=WarsawUniv,l=loc" + i);
				man.addIdentity(toAdd);
				//if ((i+1)%100 == 0)
				//	stopTimer(i, "IDENTITY CREATION snapshot");
			}
			stopTimer(N, "IDENTITY CREATION", LIMIT_NORMAL);
			
			N=10;
			startTimer();
			for (int i=0; i<N; i++)
			{
				Group g = man.addGroup(null, "G1_" + i);
				for (int j=0; j<N; j++)
				{
					Group g2 = man.addGroup(g, "G2_" + j);
					for (int k=0; k<N; k++)
						man.addGroup(g2, "G3_" + k);
				}
				//if ((i+1)%100 == 0)
				//	stopTimer(i, "IDENTITY CREATION snapshot");
			}
			stopTimer(N*N*N, "GROUP CREATION", LIMIT_NORMAL);
			
			N=100;
			startTimer();
			Group g = null;
			for (int i=0; i<N; i++)
				g = man.addGroup(g, "GNested_" + i);
			stopTimer(N, "NESTED GROUP CREATION", LIMIT_LOW);
			
			N=10;
			startTimer();
			for (int k=0; k<N; k++)
			{
				Identity idd = new Identity(IdentityType.DN, 
					"cn=ala,c=pl,ou=icm,o=WarsawUniv,l=loc0");

				for (int i=0; i<N; i++)
					for (int j=0; j<N; j++)
						man.addToGroup(idd, new Group("/G1_" + i + "/G2_" + j));
			}
			stopTimer(N*N*N, "ADD TO GROUP", LIMIT_NORMAL);
			
			N=100;
			List<String> vals = new ArrayList<String>();
			vals.add("some example value");
			Element idEl = new Element(null, toAdd);
			startTimer();
			for (int i=0; i<N; i++)
				man.setAttribute(idEl, 
					new Attribute("urn:unicore:someAttr" + i, vals, ""), 
					false);			
			stopTimer(N, "ADD GLOBAL ATTR", LIMIT_NORMAL);

			idEl = new Element(new Group("/G1_0"), null);
			startTimer();
			for (int i=0; i<N; i++)
				man.setAttribute(idEl, 
					new Attribute("urn:unicore:someAttr" + i, vals, ""), 
					false);			
			stopTimer(N, "ADD GROUP ATTR", LIMIT_NORMAL);
			
			idEl = new Element(new Group("/G1_0"), toAdd);
			startTimer();
			for (int i=0; i<N; i++)
				man.setAttribute(idEl, 
					new Attribute("urn:unicore:someAttr" + i, vals, ""), 
					false);			
			stopTimer(N, "ADD ID@GROUP ATTR", LIMIT_NORMAL);
			
			
			//********UPDATE************
			vals.add("someexamplevalue 2");
			N=100;
			idEl = new Element(null, toAdd);
			startTimer();
			for (int i=0; i<N; i++)
				man.setAttribute(idEl, 
					new Attribute("urn:unicore:someAttr" + i, vals, ""), 
					true);			
			stopTimer(N, "UPDATE GLOBAL ATTR", LIMIT_NORMAL);

			idEl = new Element(new Group("/G1_0"), null);
			startTimer();
			for (int i=0; i<N; i++)
				man.setAttribute(idEl, 
					new Attribute("urn:unicore:someAttr" + i, vals, ""), 
					true);			
			stopTimer(N, "UPDATE GROUP ATTR", LIMIT_NORMAL);
			
			idEl = new Element(new Group("/G1_0"), toAdd);
			startTimer();
			for (int i=0; i<N; i++)
				man.setAttribute(idEl, 
					new Attribute("urn:unicore:someAttr" + i, vals, ""), 
					true);			
			stopTimer(N, "UPDATE ID@GROUP ATTR", LIMIT_NORMAL);
			
			
			
			
			N=100;
			List<Identity> allIds = null;
			startTimer();
			for (int i=0; i<N; i++)
				allIds = query.getAllIdentities();
			stopTimer(N, "GET ALL IDENTITIES", LIMIT_NORMAL);
			
			N=10;
			startTimer();
			for (int j=0; j<N; j++)
				for (int i=0; i<N; i++)
				{
					Identity toCheck = new Identity(IdentityType.DN, 
						"cn=ala,c=pl,ou=icm,o=WarsawUniv,l=loc"+i);
					query.getAllGroups(toCheck, true);
				}
			stopTimer(N*N, "GET ALL IDENTITY'S IMPLIED GROUPS", LIMIT_HIGH);
			startTimer();
			for (int j=0; j<N; j++)
				for (int i=0; i<N; i++)
				{
					Identity toCheck = new Identity(IdentityType.DN, 
						"cn=ala,c=pl,ou=icm,o=WarsawUniv,l=loc"+i);
					query.getAllGroups(toCheck, false);
				}
			stopTimer(N*N, "GET ALL IDENTITY'S DIRECT GROUPS", LIMIT_HIGH);
			
			N=100;
			idEl = new Element(null, toAdd);
			startTimer();
			for (int i=0; i<N; i++)
				query.getAttributes(idEl, null, true, true, true);
			stopTimer(N, "GET ALL IDENTITY'S ATTRIBUTES (true, true, true)", LIMIT_HIGH);

			startTimer();
			for (int i=0; i<N; i++)
				query.getAttributes(idEl, null, true, true, false);
			stopTimer(N, "GET ALL IDENTITY'S ATTRIBUTES (true, true, false)", LIMIT_HIGH);

			startTimer();
			for (int i=0; i<N; i++)
				query.getAttributes(idEl, null, true, false, false);
			stopTimer(N, "GET ALL IDENTITY'S ATTRIBUTES (true, false, false)", LIMIT_HIGH);

			startTimer();
			for (int i=0; i<N; i++)
				query.getAttributes(idEl, null, false, false, false);
			stopTimer(N, "GET IDENTITY'S GLOBAL ATTRIBUTES", LIMIT_HIGH);

			idEl = new Element(new Group("/G1_0"), toAdd);
			startTimer();
			for (int i=0; i<N; i++)
				query.getAttributes(idEl, null, true, true, true);
			stopTimer(N, "GET IDENTITY'S ATTRIBUTES@GROUP (true, true, true)", LIMIT_HIGH);

			
			N=10;
			startTimer();
			for (int j=0; j<N; j++)
				for (int i=0; i<N; i++)
					query.getGroupContent(new Group("/G1_"+i));
			stopTimer(N*N, "GET GROUP CONTENTS", LIMIT_HIGH);
			
			
			
			//******DELETION************
			
			N=10;
			startTimer();
			for (int i=0; i<N; i++)
				for (int j=0; j<N; j++)
					man.removeFromGroup(toAdd, 
						new Group("/G1_" + i + "/G2_" + j));
			stopTimer(N, "REMOVE FROM GROUP", LIMIT_NORMAL);

			N=10;
			startTimer();
			man.removeGroup(new Group("/GNested_0"), true);
			for (int i=0; i<N; i++)
				man.removeGroup(new Group("/G1_" + i), true);
			stopTimer(N+1, "NESTED GROUP REMOVAL", LIMIT_NORMAL/10);
			
			N=500;
			startTimer();
			for (int i=0; i<N; i++)
				man.removeIdentity(allIds.get(i));
			stopTimer(N+1, "IDENTITIY REMOVAL", LIMIT_NORMAL);
			
			man.purgeHistoricalData(new Date());
			
			System.out.println("PERFORMANCE TEST PASSED");
		} catch (Exception e)
		{
			e.printStackTrace();
			fail();
		}
	}
}







