/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Apr 6, 2008
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.stress;

import java.util.ArrayList;
import java.util.Date;

import org.junit.Before;
import org.junit.Test;

import static junit.framework.Assert.*;

import pl.edu.icm.unicore.uvos.api.Attribute;
import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.Identity;
import pl.edu.icm.unicore.uvos.api.IdentityType;
import pl.edu.icm.unicore.uvos.api.Permissions;
import pl.edu.icm.unicore.uvos.db.InitDB;
import pl.edu.icm.unicore.uvos.engine.VirtualSecurityContextImpl;
import pl.edu.icm.unicore.uvos.engine.api.UVOSManagementInterface;
import pl.edu.icm.unicore.uvos.engine.api.UVOSQueryInterface;
import pl.edu.icm.unicore.uvos.server.ComponentContainer;
import pl.edu.icm.unicore.uvos.server.idresolver.Authenticator;

/**
 * @author K. Benedyczak
 */
public class StressTest
{
	private boolean work = true;
	private long lastMem = -1;
	
	private UVOSManagementInterface man;
	private UVOSQueryInterface query;
	private Identity u;
	private Group a, b, c, d;
	private ComponentContainer container;
	
	@Before
	public void setUp() throws Exception
	{
		try
		{
			InitDB.main(new String[]{"src/test/resources/uvosServer.conf", "-b"});
			container = new ComponentContainer("src/test/resources/uvosServer.conf");
		} catch (Throwable e)
		{
			System.err.println("Severe error when starting UVOS Web Server:");
			e.printStackTrace();
			fail();
		}
	}

	private TestThread[] createThreads()
	{
		TestThread threads[] = new TestThread[13];
		threads[0] = new TIsMember();
		threads[1] = new TGetAllGroups();
		threads[2] = new TGetAttributes();
		threads[3] = new TGetAllEquivs();
		threads[4] = new TGetGroupCont();
		threads[5] = new TGetAllIds();
		threads[6] = new TMembersMess();
		threads[7] = new TGroupMess();
		threads[8] = new TAttrsMess();
		threads[9] = new TEquivMess();
		threads[10] = new TGroup2Mess();
		threads[11] = new TIdsMess();
		threads[12] = new TIdsSimpleMess();
		return threads;
	}
	
	@Test
	public void test1()
	{
		Authenticator.injectFakeContext(new VirtualSecurityContextImpl(Permissions.ALL_P));
		try
		{
			man = container.getComponent(UVOSManagementInterface.class);
			query = container.getComponent(UVOSQueryInterface.class);
			
			u = new Identity(IdentityType.EMAIL, "ex1@ex.com", "label");
			
			man.addIdentity(u);
			a = man.addGroup(null, "A");
			b = man.addGroup(a, "B");
			
			TestThread threads[] = createThreads();
			int N = threads.length;

			/*System.out.println("One-by-one part");
			memoryCheck();
			for (int i=0; i<N; i++)
			{
				work = true;
				threads[i].start();
				Thread.sleep(2000);
				work = false;
				threads[i].join();
			}
			assertTrue("Memory leak??", memoryCheck()<10000000);			
			*/
			
			System.out.println("Concurrent part");
			threads = createThreads();
			work = true;
			for (int i=0; i<N; i++)
				if (threads[i] != null)
					threads[i].start();
			Thread.sleep(1200);//for now
			work = false;
			for (int i=0; i<N; i++)
				if (threads[i] != null)
					threads[i].join();
			man.purgeHistoricalData(new Date());
			
			assertTrue("Memory leak??", memoryCheck()<10000000);
		} catch (Exception e)
		{
			e.printStackTrace();
			fail();
		}
	}
	
	private long memoryCheck()
	{
		Runtime r = Runtime.getRuntime();
		System.gc();
		long usedMemory = r.totalMemory() - r.freeMemory();
		long ret = 0;
		
		System.out.println("Used memory: " + usedMemory);
		if (lastMem != -1)
		{
			ret = (usedMemory - lastMem);
			System.out.println("Change: " + ret);
			
		}
		lastMem = usedMemory;
		return ret;
	}
	
	private abstract class TestThread extends Thread
	{
		public TestThread(String name)
		{
			super(name);
		}
		
		public void run()
		{
			int n = 0;
			Authenticator.injectFakeContext(new VirtualSecurityContextImpl(Permissions.ALL_P));
			while (work)
			{
				try
				{
					workWork();
					n++;
				} catch (Exception e)
				{
					System.out.println("Error in thread " + 
							getName());
					e.printStackTrace();
					System.out.println("Thread " + getName() + " did " + n + " ops.");
					fail();
				}
			}
			System.out.println("Thread " + getName() + " did " + n + " ops.");
		}
		
		protected abstract void workWork() throws Exception;
	}
	
	private class TIsMember extends TestThread
	{
		public TIsMember()
		{
			super("TIsMember");
		}

		public void workWork() throws Exception
		{
			query.isMember(u, a, true);
		}
	}
	
	private class TGetAllGroups extends TestThread
	{
		public TGetAllGroups()
		{
			super("TGetAllGroups");
		}

		public void workWork() throws Exception
		{
			/*List<Group> ret = */query.getAllGroups(u, true);
			/*if (ret.contains(b) && !ret.contains(a))
				throw new Exception("Consistency: is in /A/B not in /A");
			if (ret.contains(d) && !ret.contains(c))
				throw new Exception("Consistency: is in /C/D not in /C");
				*/
		}
	}
	
	private class TGetAttributes extends TestThread
	{
		private Element e = new Element(null, u);

		public TGetAttributes()
		{
			super("TGetAttributes");
		}
		
		public void workWork() throws Exception
		{
			query.getAttributes(e, null, true, true, true);
		}
	}
	
	private class TGetAllEquivs extends TestThread
	{
		public TGetAllEquivs()
		{
			super("TGetAllEquivs");
		}

		public void workWork() throws Exception
		{
			query.getAllEquivalents(u);
		}
	}
	
	private class TGetGroupCont extends TestThread
	{
		public TGetGroupCont()
		{
			super("TGetGroupCont");
		}

		public void workWork() throws Exception
		{
			query.getGroupContent(a);
		}
	}
	
	private class TGetAllIds extends TestThread
	{
		public TGetAllIds()
		{
			super("TGetAllIds");
		}

		public void workWork() throws Exception
		{
			query.getAllIdentities();
		}
	}

	
	private class TMembersMess extends TestThread
	{
		public TMembersMess()
		{
			super("TMembersMess");
		}

		public void workWork() throws Exception
		{
			man.addToGroup(u, b);
			man.removeFromGroup(u, b);
		}
	}

	private class TGroupMess extends TestThread
	{
		public TGroupMess()
		{
			super("TGroupMess");
		}

		public void workWork() throws Exception
		{
			c = man.addGroup(null, "C");
			d = man.addGroup(c, "D");
			man.addToGroup(u, d);
			man.removeGroup(c, true);
		}
	}

	private class TAttrsMess extends TestThread
	{
		public TAttrsMess()
		{
			super("TAttrsMess");
		}

		private Element e = new Element(null, u);
		private Element e2 = new Element(a, u);
		private ArrayList<String> vals = new ArrayList<String>();
		private ArrayList<String> vals2 = new ArrayList<String>();
		{
			vals.add("1");
			vals.add("2");
			vals2.add("3");
			vals2.add("4");
		}

		private Attribute a1 = new Attribute("attr1", vals, null);
		private Attribute a2 = new Attribute("attr2", vals, null);
		
		public void workWork() throws Exception
		{
			man.setAttribute(e, a1, false);
			man.setAttribute(e2, a2, false);
			man.setAttribute(e, a1, true);
			man.removeAttribute(e2, a2.getName());
			man.removeAttribute(e, a1.getName());
		}
	}
	
	private class TEquivMess extends TestThread
	{
		private Identity u1 = new Identity(IdentityType.DN, "cn=ola");
		private Identity u2 = new Identity(IdentityType.DN, "cn=tola");
		private Identity u3 = new Identity(IdentityType.DN, "cn=fasola");
		
		public TEquivMess()
		{
			super("TEquivMess");
		}

		public void workWork() throws Exception
		{
			man.addIdentity(u1, u);
			man.addIdentity(u2, u);
			man.addIdentity(u3, u1);
			man.removeIdentity(u1);
			man.removeIdentity(u2);
			man.removeIdentity(u3);
		}
	}
	
	
	private class TGroup2Mess extends TestThread
	{
		public TGroup2Mess()
		{
			super("TGroup2Mess");
		}

		public void workWork() throws Exception
		{
			Group x = man.addGroup(a, "X");
			Group y = man.addGroup(a, "Y");
			man.removeGroup(x, false);
			man.removeGroup(y, false);
		}
	}

	private class TIdsMess extends TestThread
	{
		public TIdsMess()
		{
			super("TIdsMess");
		}

		private final int N = 10;
		
		public void workWork() throws Exception
		{
			Identity ids[] = new Identity[N];
			for(int i=0; i<N; i++)
			{
				ids[i] = new Identity(IdentityType.EMAIL, 
						"aa@example." + i);
				man.addIdentity(ids[i]);
				man.addToGroup(ids[i], a);
			}
			for(int i=0; i<N; i++)
			{
				man.removeIdentity(ids[i]);
			}
		}
	}
	
	private class TIdsSimpleMess extends TestThread
	{
		public TIdsSimpleMess()
		{
			super("TIdsSimpleMess");
		}

		public void workWork() throws Exception
		{
			Identity id = new Identity(IdentityType.EMAIL, 
						"aa@example");
			man.addIdentity(id);
			man.removeIdentity(id);
		}
	}
	
	
}
