/*
 * Copyright (c) 2007, 2008 ICM Uniwersytet Warszawski All rights reserved.
 * See LICENCE file for licencing information.
 *
 * Created on Sep 20, 2007
 * Author: K. Benedyczak <golbi@mat.umk.pl>
 */

package pl.edu.icm.unicore.uvos.tests.functional;

import java.util.Date;
import java.util.List;

import org.junit.Test;

import eu.unicore.bugsreporter.annotation.FunctionalTest;
import eu.unicore.bugsreporter.annotation.FunctionalTests;

import static junit.framework.Assert.*;

import pl.edu.icm.unicore.uvos.api.Element;
import pl.edu.icm.unicore.uvos.api.Group;
import pl.edu.icm.unicore.uvos.api.exceptions.ElementAlreadyExistsException;
import pl.edu.icm.unicore.uvos.api.exceptions.GroupNotEmptyException;

/**
 * Performs functional tests. Assumptions:
 * Server is up and running and DB is empty
 * @author K. Benedyczak
 */
public class GroupFunctionalTest extends AbstractTestBase
{
	@Test
	public void testAll()
	{
		try
		{
			tstGroups();
		} catch (Exception e)
		{
			e.printStackTrace();
			fail("Exception: " + e.getMessage());
		}
	}
	
	@FunctionalTests({
		@FunctionalTest(id="uvos_grp_add"),
		@FunctionalTest(id="uvos_grp_copy"),
		@FunctionalTest(id="uvos_grp_del")
	})
	private void tstGroups() throws Exception
	{
		//create
		Group gA = voMan.addGroup(null, "A");
		Group gB = voMan.addGroup(gA, "B");
		Group gC = voMan.addGroup(gB, "C");
		voMan.addGroup(gA, "Z");
		Group gY = voMan.addGroup(gA, "Y");
		Group gX = voMan.addGroup(gA, "X");
		//list + verify creation
		List<Element> cont1 = voQuery.getGroupContent(null);
		List<Element> contA = voQuery.getGroupContent(gA);
		if (cont1.size() != 1)
			fail("Added one root group, got " + dumpList(cont1));
		if (contA.size() != 4)
			fail("Added 4 subgroups, got " + dumpList(contA));
		for (Element e: contA)
			if (e.getIdentity() != null || !"BXYZ".contains(
					e.getGroup().getName()))
				fail("Wrong content: " + dumpList(contA));
		//create duplicate
		try
		{
			voMan.addGroup(null, "A");
			fail("Succees adding duplicate to root (A)");
		} catch (ElementAlreadyExistsException ee){}
		try
		{
			voMan.addGroup(gA, "X");
			fail("Succees adding duplicate to A (X)");
		} catch (ElementAlreadyExistsException ee){}
		
		//copy
		voMan.copyGroup(gY, gC, "COPIEDY", false);
		cont1 = voQuery.getGroupContent(gC);
		if (cont1.size() != 1)
			fail("No result of copy....");
		//copy (rename)
		Group gRoot = new Group("/");
		Group gAA = voMan.copyGroup(gA, gRoot, "AA", true);
		cont1 = voQuery.getGroupContent(null);
		if (cont1.size() != 1 || 
			!cont1.get(0).getGroup().getName().equals(gAA.getName())
			|| !gAA.getName().equals("AA"))
			fail("Error renaming: " + gA + " to " + gAA + 
					" got: " + dumpList(cont1));
		voMan.copyGroup(gAA, gRoot, "A", true);
		try
		{
			voMan.copyGroup(gY, gA, "X", true);
			fail("renamed group to the existing name X");
		} catch(ElementAlreadyExistsException e){}
		
		//copy (move)
		Group movedG = voMan.copyGroup(gB, gY, null, true);
		contA = voQuery.getGroupContent(gA);
		cont1 = voQuery.getGroupContent(gY);
		if (contA.size() != 3 || cont1.size() != 1)
			fail("Move succeded but groups not moved: " + dumpList(contA) +
					" --and-- " + dumpList(cont1));
		if (!movedG.getParent().getName().equals("Y"))
			fail("Move returned group has wrong parent: " + 
					movedG.getParent().getName());
		voMan.copyGroup(movedG, gA, null, true);
		
		//delete
		try
		{
			voMan.removeGroup(gB, false);
			fail("non recursive delete removed nonempty group B");
		} catch(GroupNotEmptyException e){}
		voMan.removeGroup(gY, false);
		voMan.removeGroup(gX, true);
		if (voQuery.getGroupContent(gA).size() != 2)
			fail("Removed two groups and got: " + dumpList(voQuery.getGroupContent(gA)));
		voMan.removeGroup(gA, true);
		cont1 = voQuery.getGroupContent(null);
		if (cont1.size() != 0)
			fail("After recursive removal root is non empty: " + dumpList(cont1));
		
		//purge
		voMan.purgeHistoricalData(new Date());
		
	}
}














